/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*    This file is part of "mipsymmetries" a collection of                   */
/*    symmetry handling methods.                                             */
/*                                                                           */
/*    Copyright (C) 2010-2015  Marc Pfetsch, Thomas Rehn                     */
/*                                                                           */
/*                                                                           */
/*    Based on SCIP  --- Solving Constraint Integer Programs                 */
/*       SCIP is distributed under the terms of the SCIP Academic Licence,   */
/*       see file COPYING in the SCIP distribution.                          */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
#pragma ident "@(#) $Id: branch_first.c 377 2012-08-06 11:08:59Z pfetsch $"

/**@file   branch_first.c
 * @ingroup BRANCHINGRULES
 * @brief  index branching rule
 * @author Marc Pfetsch
 *
 * Branch on the first variable with fractional LP value.
 */

/*---+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8----+----9----+----0----+----1----+----2*/

#include <assert.h>
#include <string.h>

#include "branch_first.h"


#define BRANCHRULE_NAME            "first"
#define BRANCHRULE_DESC            "first variable branching rule"
#define BRANCHRULE_PRIORITY        2000000
#define BRANCHRULE_MAXDEPTH        -1
#define BRANCHRULE_MAXBOUNDDIST    1.0

/*
 * Callback methods of branching rule
 */

/** copy method for branchrule plugins (called when SCIP copies plugins) */
static
SCIP_DECL_BRANCHCOPY(branchCopyFirst)
{  /*lint --e{715}*/
   assert(scip != NULL);
   assert(branchrule != NULL);
   assert(strcmp(SCIPbranchruleGetName(branchrule), BRANCHRULE_NAME) == 0);

   /* call inclusion method of branchrule */
   SCIP_CALL( SCIPincludeBranchruleFirst(scip) );

   return SCIP_OKAY;
}


/** branching execution method for fractional LP solutions */
static
SCIP_DECL_BRANCHEXECLP(branchExeclpFirst)
{  /*lint --e{715}*/
   SCIP_VAR** vars;
   int nvars;

   assert( branchrule != NULL );
   assert( strcmp(SCIPbranchruleGetName(branchrule), BRANCHRULE_NAME) == 0 );
   assert( scip != NULL );
   assert( result != NULL );

   /* get branching candidates */
#if ( SCIP_VERSION > 302 || ( SCIP_VERSION == 302 && SCIP_SUBVERSION > 0 ) )
   SCIP_CALL( SCIPgetLPBranchCands(scip, &vars, NULL, NULL, &nvars, NULL, NULL) );
#else
   SCIP_CALL( SCIPgetLPBranchCands(scip, &vars, NULL, NULL, &nvars, NULL) );
#endif

   if ( nvars > 0 )
   {
      SCIP_VAR* var = vars[0];
      assert( SCIPvarGetStatus(var) == SCIP_VARSTATUS_COLUMN || SCIPvarGetStatus(var) == SCIP_VARSTATUS_LOOSE );
      assert( SCIPvarGetType(var) == SCIP_VARTYPE_BINARY || SCIPvarGetType(var) == SCIP_VARTYPE_INTEGER );

      /* branch on first variable with fractional value */
      SCIPdebugMessage("branch on var <%s> [%d]\n", SCIPvarGetName(var), SCIPvarGetProbindex(var));
      SCIP_CALL( SCIPbranchVar(scip, var, NULL, NULL, NULL) );
      *result = SCIP_BRANCHED;
      return SCIP_OKAY;
   }
   *result = SCIP_DIDNOTRUN;

   return SCIP_OKAY;
}


/** branching execution method for not completely fixed pseudo solutions */
static
SCIP_DECL_BRANCHEXECPS(branchExecpsFirst)
{  /*lint --e{715}*/
   SCIP_VAR** vars;
   int nvars;

   assert( branchrule != NULL );
   assert( strcmp(SCIPbranchruleGetName(branchrule), BRANCHRULE_NAME) == 0 );
   assert( scip != NULL );
   assert( result != NULL );

   /* get branching candidates */
   SCIP_CALL( SCIPgetPseudoBranchCands(scip, &vars, NULL, &nvars) );

   if ( nvars > 0 )
   {
      SCIP_VAR* var = vars[0];
      assert( SCIPvarGetStatus(var) == SCIP_VARSTATUS_COLUMN || SCIPvarGetStatus(var) == SCIP_VARSTATUS_LOOSE );
      assert( SCIPvarGetType(var) == SCIP_VARTYPE_BINARY || SCIPvarGetType(var) == SCIP_VARTYPE_INTEGER );

      /* branch on first variable */
      SCIPdebugMessage("branch on var <%s> [%d]\n", SCIPvarGetName(var), SCIPvarGetProbindex(var));
      SCIP_CALL( SCIPbranchVar(scip, var, NULL, NULL, NULL) );
      *result = SCIP_BRANCHED;
      return SCIP_OKAY;
   }
   *result = SCIP_DIDNOTRUN;

   return SCIP_OKAY;
}



/*
 * branching rule specific interface methods
 */

/** creates the first branching rule and includes it in SCIP */
SCIP_RETCODE SCIPincludeBranchruleFirst(
   SCIP*                 scip                /**< SCIP data structure */
   )
{
   SCIP_BRANCHRULE* branchrule = NULL;

   /* include branching rule */
   SCIP_CALL( SCIPincludeBranchruleBasic(scip, &branchrule, BRANCHRULE_NAME, BRANCHRULE_DESC, BRANCHRULE_PRIORITY, BRANCHRULE_MAXDEPTH, BRANCHRULE_MAXBOUNDDIST, NULL) );

   /* set additional callbacks */
   SCIP_CALL( SCIPsetBranchruleCopy(scip, branchrule, branchCopyFirst) );
   SCIP_CALL( SCIPsetBranchruleExecLp(scip, branchrule, branchExeclpFirst) );
   SCIP_CALL( SCIPsetBranchruleExecPs(scip, branchrule, branchExecpsFirst) );

   return SCIP_OKAY;
}
