/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*    This file is part of "mipsymmetries" a collection of                   */
/*    symmetry handling methods.                                             */
/*                                                                           */
/*    Copyright (C) 2010-2015  Marc Pfetsch, Thomas Rehn                     */
/*                                                                           */
/*                                                                           */
/*    Based on SCIP  --- Solving Constraint Integer Programs                 */
/*       SCIP is distributed under the terms of the SCIP Academic Licence,   */
/*       see file COPYING in the SCIP distribution.                          */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
#pragma ident "@(#) $Id: branch_firstindex.c 784 2015-03-29 11:18:44Z pfetsch $"

/**@file   branch_firstindex.c
 * @ingroup BRANCHINGRULES
 * @brief  first index branching rule
 * @author Marc Pfetsch
 *
 * We branch on the first variable that is not yet fixed, independent of its value in the LP solution. This branching
 * rule was necessary for the first version of isomorphism pruning.
 */

/*---+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8----+----9----+----0----+----1----+----2*/

#include <assert.h>
#include <string.h>

#include "branch_firstindex.h"


#define BRANCHRULE_NAME            "firstindex"
#define BRANCHRULE_DESC            "first index branching rule"
#define BRANCHRULE_PRIORITY        2000000
#define BRANCHRULE_MAXDEPTH        -1
#define BRANCHRULE_MAXBOUNDDIST    1.0




/*
 * Data structures
 */

/** branching rule data */
struct SCIP_BranchruleData
{
};



/*
 * Callback methods of branching rule
 */

/** copy method for branchrule plugins (called when SCIP copies plugins) */
static
SCIP_DECL_BRANCHCOPY(branchCopyFirstindex)
{  /*lint --e{715}*/
   assert(scip != NULL);
   assert(branchrule != NULL);
   assert(strcmp(SCIPbranchruleGetName(branchrule), BRANCHRULE_NAME) == 0);

   /* call inclusion method of branchrule */
   SCIP_CALL( SCIPincludeBranchruleFirstindex(scip) );

   return SCIP_OKAY;
}


/** branching execution method for fractional LP solutions */
static
SCIP_DECL_BRANCHEXECLP(branchExeclpFirstindex)
{  /*lint --e{715}*/
   SCIP_VAR** vars;
   int nvars;
   int j;

   assert( branchrule != NULL );
   assert( strcmp(SCIPbranchruleGetName(branchrule), BRANCHRULE_NAME) == 0 );
   assert( scip != NULL );
   assert( result != NULL );

   nvars = SCIPgetNVars(scip);
   vars = SCIPgetVars(scip);

   *result = SCIP_DIDNOTRUN;
   for (j = 0; j < nvars; ++j)
   {
      SCIP_VAR* var = vars[j];
      assert( SCIPvarGetStatus(var) == SCIP_VARSTATUS_COLUMN || SCIPvarGetStatus(var) == SCIP_VARSTATUS_LOOSE );

      if ( SCIPvarGetType(var) != SCIP_VARTYPE_BINARY && SCIPvarGetType(var) != SCIP_VARTYPE_INTEGER )
         continue;

      /* branch on first variable that is not fixed (independent on whether the LP value of the variable is integral) */
      if ( ! SCIPisEQ(scip, SCIPvarGetLbLocal(var), SCIPvarGetUbLocal(var)) )
      {
         SCIPdebugMessage("branch on var <%s> [%d]\n", SCIPvarGetName(var), SCIPvarGetProbindex(var));
         SCIP_CALL( SCIPbranchVar(scip, var, NULL, NULL, NULL) );
         *result = SCIP_BRANCHED;
         return SCIP_OKAY;
      }
      else
      {
         SCIPdebugMessage("ignore fixed var <%s> [%f, %f] for branching.\n",
            SCIPvarGetName(var), SCIPvarGetLbLocal(var), SCIPvarGetUbLocal(var));
      }
   }
   *result = SCIP_DIDNOTFIND;

   return SCIP_OKAY;
}


/** branching execution method for not completely fixed pseudo solutions */
static
SCIP_DECL_BRANCHEXECPS(branchExecpsFirstindex)
{  /*lint --e{715}*/
   SCIP_VAR** vars;
   int nvars;
   int j;

   assert( branchrule != NULL );
   assert( strcmp(SCIPbranchruleGetName(branchrule), BRANCHRULE_NAME) == 0 );
   assert( scip != NULL );
   assert( result != NULL );

   nvars = SCIPgetNVars(scip);
   vars = SCIPgetVars(scip);

   *result = SCIP_DIDNOTRUN;
   for (j = 0; j < nvars; ++j)
   {
      SCIP_VAR* var = vars[j];
      assert( SCIPvarGetStatus(var) == SCIP_VARSTATUS_COLUMN || SCIPvarGetStatus(var) == SCIP_VARSTATUS_LOOSE );

      if ( SCIPvarGetType(var) != SCIP_VARTYPE_BINARY && SCIPvarGetType(var) != SCIP_VARTYPE_INTEGER )
         continue;

      /* branch on first variable that is not fixed */
      if ( ! SCIPisEQ(scip, SCIPvarGetLbLocal(var), SCIPvarGetUbLocal(var)) )
      {
         SCIPdebugMessage("pseudo branch on var <%s> [%d]\n", SCIPvarGetName(var), SCIPvarGetIndex(var));
         SCIP_CALL( SCIPbranchVar(scip, var, NULL, NULL, NULL) );
         *result = SCIP_BRANCHED;
         return SCIP_OKAY;
      }
      else
      {
         SCIPdebugMessage("ignore fixed var <%s> [%f, %f] for pseudo branching.\n",
            SCIPvarGetName(var), SCIPvarGetLbLocal(var), SCIPvarGetUbLocal(var));
      }
   }
   *result = SCIP_DIDNOTFIND;

   return SCIP_OKAY;
}



/*
 * branching rule specific interface methods
 */

/** creates the first index branching rule and includes it in SCIP */
SCIP_RETCODE SCIPincludeBranchruleFirstindex(
   SCIP*                 scip                /**< SCIP data structure */
   )
{
   SCIP_BRANCHRULE* branchrule = NULL;

   /* include branching rule */
   SCIP_CALL( SCIPincludeBranchruleBasic(scip, &branchrule, BRANCHRULE_NAME, BRANCHRULE_DESC, BRANCHRULE_PRIORITY, BRANCHRULE_MAXDEPTH, BRANCHRULE_MAXBOUNDDIST, NULL) );

   /* set additional callbacks */
   SCIP_CALL( SCIPsetBranchruleCopy(scip, branchrule, branchCopyFirstindex) );
   SCIP_CALL( SCIPsetBranchruleExecLp(scip, branchrule, branchExeclpFirstindex) );
   SCIP_CALL( SCIPsetBranchruleExecPs(scip, branchrule, branchExecpsFirstindex) );

   return SCIP_OKAY;
}
