/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*    This file is part of "mipsymmetries" a collection of                   */
/*    symmetry handling methods.                                             */
/*                                                                           */
/*    Copyright (C) 2010-2015  Marc Pfetsch, Thomas Rehn                     */
/*                                                                           */
/*                                                                           */
/*    Based on SCIP  --- Solving Constraint Integer Programs                 */
/*       SCIP is distributed under the terms of the SCIP Academic Licence,   */
/*       see file COPYING in the SCIP distribution.                          */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
#pragma ident "@(#) $Id: branch_orbital.cpp 10 2011-06-16 15:27:47Z pfetsch $"

/**@file   branch_orbital.cpp
 * @brief  Implementation of orbital branching
 * @author Marc Pfetsch
 * @author Thomas Rehn
 *
 * This file provides an implementation of orbital branching as described in
 * J. Ostrowski, J. Linderoth, F. Rossi, and S. Smriglio: Orbital branching,@n
 * Mathematical Programming 126, No 1, pp. 147-178 (2011)
 *
 * Orbital branching has only been implemented for symmetries acting on binary variables. Consider an orbit \f$O\f$ of
 * binary variables. Then the disjunction
 * \f[
 *    \sum_{j \in O} x_j \geq 0 \qquad\mbox{or}\qquad \sum_{j \in O} x_j \leq 0
 * \f]
 * is valid. The second is equivalent to \f$x_j = 0\f$ for all \f$j \in O\f$. For the first we can choose on element
 * \f$i \in O\f$ of the orbit and fix this variable to 1, because of the symmetry. We arrive at the following
 * \f[
 *    x_i = 1 \qquad\mbox{or}\qquad x_j = 0 \qquad \forall\, j \in O.
 * \f]
 * The implementation includes orbital fixing, i.e., we fix all variables in an orbit to 0, if one variable has already
 * been fixed to 0 and similarly for 1 fixings.
 *
 * Continuous variables are also allowed, since they do not disturb the process. Integer variables can be present, but
 * we make sure that they are not moved by the symmetries considered for orbital branching.
 *
 * The parameter avoidsubtreenodes avoids the use of orbital branching on sub-nodes in the tree, if the symmetry group
 * becomes trivial in a certain node. Due to branching new symmetry might occur, but is ignored in this case.
 */

#include <cstring>
#include "branch_orbital.h"
#include <vector>
#include <list>
#include <set>

#include "cons_subtreebool.h"
#include "cons_symmetries.h"
#include "presol_orbitprobing.h"

/* branching rule properties */
#define BRANCHRULE_NAME            "Orbital"
#define BRANCHRULE_DESC            "orbital branching"
#define BRANCHRULE_PRIORITY        10000000
#define BRANCHRULE_MAXDEPTH        -1
#define BRANCHRULE_MAXBOUNDDIST    1.0

/* default parameters */
#define SCIP_DEFAULT_ORBITCHOICE             0    /**< choice of orbit to branch on (0: largest, 1: break symmetry, 2: keep symmetry, 3: max product) */
#define SCIP_DEFAULT_MINIMALORBITSIZE        2    /**< require a minimum orbit size of to branch on an orbit */
#define SCIP_DEFAULT_AVOIDSUBTREENODES   FALSE    /**< If symmetry group becomes trivial avoid orbital branching on sub-nodes? */
#define SCIP_DEFAULT_LOCALSYMMETRY       FALSE    /**< Compute local symmetry group in each node? */


/* useful typedefs */
using namespace permlib;

typedef std::list<unsigned int> uintlist;

typedef AbstractPermutationGroup::OrbitList OrbitList;
typedef boost::scoped_ptr<OrbitList> OrbitListPtr;
typedef std::set<permlib::dom_int>::const_iterator OrbitSetIt;
typedef OrbitList::const_iterator OrbIt;



/*
 * Data structures
 */

/** branching rule data */
struct SCIP_BranchruleData
{
   PermutationGroupPtr   group;              /**< symmetry group */
   unsigned int          nvars;              /**< number of variables */
   SCIP_VAR**            vars;               /**< array with variables on which orbital branching should be performed */
   SCIP_HASHMAP*         varmap;             /**< map of variables to indices in vars array */
   int                   orbitchoice;        /**< choice of orbit to branch on (0: largest, 1: keep symmetry, 2: max product) */
   int                   minimalorbitsize;   /**< require a minimum orbit size of to branch on an orbit */
   int                   nfixedzero;         /**< number of variables fixed to 0 */
   int                   nfixedone;          /**< number of variables fixed to 1 */
   uintvecPtr            binaryvars;         /**< indices of binary variables to perform orbital branching on */
   SCIP_Bool             avoidsubtreenodes;  /**< if symmetry group becomes trivial, avoid orbital branching on sub-nodes? */
   SCIP_CONSHDLR*        subtreeboolconshdlr;/**< pointer to cons_subtreebool constraint handler (needed for avoidsubtreenodes) */
   SCIP_CONSHDLR*        symmetriesconshdlr; /**< pointer to cons_symmetries constraint handler */
   SCIP_Bool             localsymmetry;      /**< Compute local symmetry group in each node? */
   SCIP_Bool             enabled;            /**< run orbital branching? */
};



/*
 * Local methods
 */

/** Get binary branching variables on the path to root */
static
SCIP_RETCODE computeBranchingVariables(
   SCIP*                 scip,               /**< SCIP pointer */
   unsigned int          nvars,              /**< number of variables */
   SCIP_HASHMAP*         varmap,             /**< map of variables to indices in vars array */
   uintvec&              B1                  /**< indices of all variables with branched value 1 */
   )
{
   /* get curent node */
   SCIP_NODE* node = SCIPgetCurrentNode(scip);

#ifdef SCIP_OUTPUT
   SCIP_CALL( SCIPprintNodeRootPath(scip, node, NULL) );
#endif

   /* prepare data */
   B1.clear();

   /* follow path to the root (in the root no domains were changed due to branching) */
   while ( SCIPnodeGetDepth(node) != 0 )
   {
      SCIP_BOUNDCHG* boundchg;
      SCIP_DOMCHG* domchg;
      int nboundchgs;

      /* get domain changes of current node */
      domchg = SCIPnodeGetDomchg(node);
      assert( domchg != 0 );

      /* loop through all bound changes */
      nboundchgs = SCIPdomchgGetNBoundchgs(domchg);
      for (int i = 0; i < nboundchgs; ++i)
      {
         /* get bound change info */
         boundchg = SCIPdomchgGetBoundchg(domchg, i);
         assert( boundchg != 0 );

         /* branching decisions have to be in the beginning of the bound change array */
         if ( SCIPboundchgGetBoundchgtype(boundchg) != SCIP_BOUNDCHGTYPE_BRANCHING )
            break;

         /* get corresponding branching variable */
         SCIP_VAR* branchvar = SCIPboundchgGetVar(boundchg);

         /* we only consider binary variables */
         if ( SCIPvarIsBinary(branchvar) )
         {
            /* make sure that branching variable is known */
            assert( SCIPhashmapExists(varmap, (void*) branchvar) );

            if ( SCIPvarGetLbLocal(branchvar) > 0.5 )
            {
               unsigned int branchvaridx = (unsigned int) (size_t) SCIPhashmapGetImage(varmap, (void*) branchvar);
               assert( branchvaridx < nvars );
               B1.push_back(branchvaridx);
            }
         }
      }

      node = SCIPnodeGetParent(node);
   }

   return SCIP_OKAY;
}


/** largest orbit branching rule */
static
OrbIt branchLargestOrbit(
   SCIP*                 scip,               /**< SCIP pointer */
   const SCIP_BRANCHRULEDATA* branchruledata,/**< data of branching rule */
   const OrbitListPtr&   orbits,             /**< list of orbits among which we choose zero or one to branch on */
   bool                  branchLP,           /**< whether we branch on an LP solution (vs. pseudo solution) */
   uintlist&             fixzero,            /**< list of variables that should be fixed to 0 in both branches */
   uintlist&             fixone,             /**< list of variables that should be fixed to 1 in both branches */
   bool&                 infeasible          /**< whether the current node is infeasible */
   )
{
   OrbIt orbitToBranch = orbits->end();
   std::size_t largestSize = (std::size_t) branchruledata->minimalorbitsize - 1;

   fixzero.clear();
   fixone.clear();

   for (OrbIt it = orbits->begin(); it != orbits->end(); ++it)
   {
      /* check whether orbit is large enough */
      std::size_t orbitsize = it->size();
      if ( (int) orbitsize < branchruledata->minimalorbitsize )
         continue;

      /* check all variables in the orbit */
      bool havefixedone = false;
      bool havefixedzero = false;
      bool havefractional = false;
      OrbitSetIt oit, oend = it->end();
      for (oit = it->begin(); oit != oend; ++oit)
      {
         const unsigned int idx = *oit;
         assert( idx < branchruledata->nvars );
         SCIP_VAR* var = branchruledata->vars[idx];
         assert( SCIPvarIsBinary(var) );

         /* if variable is fixed to 1 -> can fix all variables to 1, see paper for explanation */
         if ( SCIPvarGetLbLocal(var) > 0.5 )
            havefixedone = true;

         /* check for zero-fixed variables */
         if ( SCIPvarGetUbLocal(var) < 0.5 )
            havefixedzero = true;

         if ( branchLP && ! SCIPisFeasIntegral(scip, SCIPgetSolVal(scip, NULL, var)) )
            havefractional = true;
      }

      /* check consistency */
      if ( havefixedone && havefixedzero )
      {
         infeasible = true;
         break;
      }

      /* fix all variables to 0 if there is one variable fixed to 0 */
      if ( havefixedzero )
      {
         assert( ! havefixedone );
         assert( oend == it->end() );
         for (oit = it->begin(); oit != oend; ++oit)
         {
            const unsigned int idx = *oit;
            assert( idx < branchruledata->nvars );
            /* only variables that are not yet fixed to 0 */
            if ( SCIPvarGetUbLocal(branchruledata->vars[idx]) > 0.5 )
               fixzero.push_back(idx);
         }
      }
      else
      {
         /* fix all variables to 1 if there is one variable fixed to 1 */
         if ( havefixedone )
         {
            assert( ! havefixedzero );
            assert( oend == it->end() );
            for (oit = it->begin(); oit != oend; ++oit)
            {
               const unsigned int idx = *oit;
               assert( idx < branchruledata->nvars );
               /* only variables that are not yet fixed to 1 */
               if ( SCIPvarGetLbLocal(branchruledata->vars[idx]) < 0.5 )
                  fixone.push_back(idx);
            }
         }
         else
         {
            /* do not chose orbit without variables with fractional values */
            if ( ! branchLP || havefractional )
            {
               /* check size */
               if ( orbitsize > largestSize )
               {
                  orbitToBranch = it;
                  largestSize = orbitsize;
               }
            }
         }
      }
   }

#ifndef NDEBUG
   if ( (int) largestSize >= branchruledata->minimalorbitsize )
   {
      SCIPdebugMessage("Found orbit of size %lu to branch on.\n", largestSize);
   }
#endif

   return orbitToBranch;
}


/** keep symmetry / break symmetry branching rule */
template<typename InputIterator>
OrbIt branchSymmetry(
   SCIP*                 scip,               /**< SCIP pointer */
   const SCIP_BRANCHRULEDATA* branchruledata,/**< data of branching rule */
   const OrbitListPtr&   orbits,             /**< list of orbits among which we choose zero or one to branch on */
   bool                  branchLP,           /**< whether we branch on an LP solution (vs. pseudo solution) */
   bool                  keepSymmetry,       /**< true iff symmetry of children is to be maximized; false for group order minimization */
   const InputIterator&  B1_begin,           /**< indices of variables which are branched to one */
   const InputIterator&  B1_end,             /**< indices of variables which are branched to one */
   uintlist&             fixzero,            /**< list of variables that should be fixed to 0 in both branches */
   uintlist&             fixone,             /**< list of variables that should be fixed to 1 in both branches */
   bool&                 infeasible          /**< whether the current node is infeasible */
   )
{
   OrbIt orbitToBranch = orbits->end();

   /** @todo use an GMP type here to compare group sizes? Or are we happy to put log(GROUP ORDER) into a double and compare those? */
   boost::uint64_t extremalGroupSize = 0xffffffffffffffff;
   if ( keepSymmetry )
      extremalGroupSize = 1;

   /* use for storing variables that have been branched on, plus one additional variable below */
   std::vector<unsigned int> B1(B1_begin, B1_end);

   /* placeholder for the candidate orbit representative - is replaced below */
   B1.push_back(0);

   fixzero.clear();
   fixone.clear();

   /* loop through orbits */
   for (OrbIt it = orbits->begin(); it != orbits->end(); ++it)
   {
      /* check whether orbit is large enough */
      const std::size_t orbitsize = it->size();
      if ( (int) orbitsize < branchruledata->minimalorbitsize )
         continue;

      /* check all variables in the orbit */
      bool havefixedone = false;
      bool havefixedzero = false;
      bool havefractional = false;
      OrbitSetIt oit, oend = it->end();
      for (oit = it->begin(); oit != oend; ++oit)
      {
         const unsigned int idx = *oit;
         assert( idx < branchruledata->nvars );
         SCIP_VAR* var = branchruledata->vars[idx];

         /* if variable is fixed to 1 -> can fix all variables to 1, see paper for explanation */
         if ( SCIPvarGetLbLocal(var) > 0.5 )
            havefixedone = true;

         /* check for zero-fixed variables */
         if ( SCIPvarGetUbLocal(var) < 0.5 )
            havefixedzero = true;

         if ( branchLP && ! SCIPisFeasIntegral(scip, SCIPgetSolVal(scip, NULL, var)) )
            havefractional = true;
      }

      /* check consistency */
      if ( havefixedone && havefixedzero )
      {
         infeasible = true;
         break;
      }

      /* fix all variables to zero if there is one variable fixed to 0 */
      if ( havefixedzero )
      {
         assert( ! havefixedone );
         assert( oend == it->end() );
         for (oit = it->begin(); oit != oend; ++oit)
         {
            const unsigned int idx = *oit;
            assert( idx < branchruledata->nvars );
            /* only variables that are not yet fixed to 0 */
            if ( SCIPvarGetUbLocal(branchruledata->vars[idx]) > 0.5 )
               fixzero.push_back(idx);
         }
      }
      else
      {
         /* fix all variables to 1 if there is one variable fixed to 1 */
         if ( havefixedone )
         {
            assert( ! havefixedzero );
            assert( oend == it->end() );
            for (oit = it->begin(); oit != oend; ++oit)
            {
               const unsigned int idx = *oit;
               assert( idx < branchruledata->nvars );
               /* only variables that are not yet fixed to 1 */
               if ( SCIPvarGetLbLocal(branchruledata->vars[idx]) < 0.5 )
                  fixone.push_back(idx);
            }
         }
         else
         {
            /* do not chose orbit without variables with fractional values */
            if ( ! branchLP || havefractional )
            {
               /* now check the size of the stabilizer after branching on the first variable */

               /* replace placeholder by new index */
               unsigned int firstidx = *(it->begin());
               B1.back() = firstidx;

               try
               {
                  const PermutationGroupPtr stab(branchruledata->group->setStabilizer(B1));

                  /* compute stabilizer of enlarged set of variables branched to 1 */
                  const boost::uint64_t stabOrder = stab->order();

                  if ( (keepSymmetry && (stabOrder > extremalGroupSize) ) || (! keepSymmetry && (stabOrder < extremalGroupSize) ) )
                  {
                     extremalGroupSize = stabOrder;
                     orbitToBranch = it;
                     SCIPdebugMessage("Found branching variable <%s> with symmetry group of order %lu.\n",
                        SCIPvarGetName(branchruledata->vars[firstidx]), stabOrder);

                     /* check for early termination */
                     if ( stabOrder == 1 )
                     {
                        assert( ! keepSymmetry );
                        /* Because of above if-condition, we have to be in the !keepSymmetry branch. We can break if we find a trivial group. */
                        break;
                     }
                  }
               }
               catch (std::bad_alloc& ba)
               {
                  SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Aborted orbital branching because we went out of memory.\n");
                  break;
               }
            }
         }
      }
   }

   return orbitToBranch;
}


/** max product branching rule */
template<typename InputIterator>
OrbIt branchMaxProduct(
   SCIP*                 scip,               /**< SCIP pointer */
   const SCIP_BRANCHRULEDATA* branchruledata,/**< data of branching rule */
   const OrbitListPtr&   orbits,             /**< list of orbits among which we choose zero or one to branch on */
   bool                  branchLP,           /**< whether we branch on an LP solution (vs. pseudo solution) */
   const InputIterator&  B1_begin,           /**< indices of variables which are branched to one */
   const InputIterator&  B1_end,             /**< indices of variables which are branched to one */
   uintlist&             fixzero,            /**< list of variables that should be fixed to 0 in both branches */
   uintlist&             fixone,             /**< list of variables that should be fixed to 1 in both branches */
   bool&                 infeasible          /**< whether the current node is infeasible */
   )
{
   OrbIt orbitToBranch = orbits->end();

   /* initialize with lower bound for orbit size */
   std::size_t maxProduct = (branchruledata->minimalorbitsize - 1) * (branchruledata->minimalorbitsize - 1);

   /* use for storing variables that have been branched on, plus one additional variable below */
   std::vector<unsigned int> B1(B1_begin, B1_end);

   /* placeholder for the candidate orbit representative - is replaced below */
   B1.push_back(0);

   fixzero.clear();
   fixone.clear();

   /* loop through orbits */
   for (OrbIt it = orbits->begin(); it != orbits->end(); ++it)
   {
      /* check whether orbit is large enough */
      const std::size_t orbitsize = it->size();
      if ( (int) orbitsize < branchruledata->minimalorbitsize )
         continue;

      /* check all variables in the orbit */
      bool havefixedone = false;
      bool havefixedzero = false;
      bool havefractional = false;
      OrbitSetIt oit, oend = it->end();
      for (oit = it->begin(); oit != oend; ++oit)
      {
         const unsigned int idx = *oit;
         assert( idx < branchruledata->nvars );
         SCIP_VAR* var = branchruledata->vars[idx];

         /* if variable is fixed to 1 -> can fix all variables to 1, see paper for explanation */
         if ( SCIPvarGetLbLocal(var) > 0.5 )
            havefixedone = true;

         /* check for zero-fixed variables */
         if ( SCIPvarGetUbLocal(var) < 0.5 )
            havefixedzero = true;

         if ( branchLP && ! SCIPisFeasIntegral(scip, SCIPgetSolVal(scip, NULL, var)) )
            havefractional = true;
      }

      /* check consistency */
      if ( havefixedone && havefixedzero )
      {
         infeasible = true;
         break;
      }

      /* fix all variables to zero if there is one variable fixed to 0 */
      if ( havefixedzero )
      {
         assert( ! havefixedone );
         assert( oend == it->end() );
         for (oit = it->begin(); oit != oend; ++oit)
         {
            const unsigned int idx = *oit;
            assert( idx < branchruledata->nvars );
            if ( SCIPvarGetUbLocal(branchruledata->vars[idx]) > 0.5 )
               fixzero.push_back(idx);
         }
      }
      else
      {
         /* fix all variables to 1 if there is one variable fixed to 1 */
         if ( havefixedone )
         {
            assert( ! havefixedzero );
            assert( oend == it->end() );
            for (oit = it->begin(); oit != oend; ++oit)
            {
               const unsigned int idx = *oit;
               assert( idx < branchruledata->nvars );
               /* only variables that are not yet fixed to 1 */
               if ( SCIPvarGetLbLocal(branchruledata->vars[idx]) < 0.5 )
                  fixone.push_back(idx);
            }
         }
         else
         {
            if ( ! branchLP || havefractional )
            {
               /* now check the size of the stabilizer after branching on the first variable */

               /* replace placeholder by new index */
               unsigned int firstidx = *(it->begin());
               B1.back() = firstidx;

               try
               {
                  /* compute stabilizer of enlarged set of variables branched to 1 */
                  const PermutationGroupPtr stab(branchruledata->group->setStabilizer(B1));

                  /* trival groups imply trivial orbits, so we can ignore them */
                  if (stab->order() == 1)
                     continue;

                  /* compute orbits that would result from branching on the first variable */
                  boost::scoped_ptr<OrbitList> childOrbits(stab->orbits(*branchruledata->binaryvars));

                  /* loop through resulting orbits */
                  std::size_t maxChildOrbitSize = (size_t) branchruledata->minimalorbitsize - 1;
                  for (OrbitList::const_iterator cit = childOrbits->begin(); cit != childOrbits->end(); ++cit)
                  {
                     const unsigned int idx2 = *(*cit).begin();
                     assert( idx2 < branchruledata->nvars );
                     SCIP_VAR* var2 = branchruledata->vars[idx2];

                     /* take only non-fixed orbits - it suffices to check only one variable in the orbit */
                     if ( SCIPvarGetLbLocal(var2) > 0.5 || SCIPvarGetUbLocal(var2) < 0.5 )
                        continue;

                     /* store largest orbit size */
                     const std::size_t childorbitsize = (*cit).size();
                     if ( childorbitsize > maxChildOrbitSize)
                        maxChildOrbitSize = childorbitsize;
                  }

                  /* compute product of the size of the original orbit with the maximal size of a child orbit */
                  const std::size_t product = orbitsize * maxChildOrbitSize;
                  if ( product > maxProduct )
                  {
                     maxProduct = product;
                     orbitToBranch = it;
                     SCIPdebugMessage("Found branching variable with orbit size %lu and child orbit size %lu (product %lu).\n",
                        (unsigned long) orbitsize, (unsigned long) maxChildOrbitSize, (unsigned long) product);
                  }
               }
               catch (std::bad_alloc& ba)
               {
                  SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Aborted orbital branching because we went out of memory.\n");
                  return orbits->end();
               }
            }
         }
      }
   }

   return orbitToBranch;
}


/** perform branching */
static
SCIP_RETCODE branch(
   SCIP*                 scip,               /**< SCIP pointer */
   SCIP_BRANCHRULEDATA*  branchruledata,     /**< data of branching rule */
   bool                  branchLP,           /**< whether we branch on an LP solution (vs. pseudo solution) */
   bool&                 branched,           /**< whether we performed a branching */
   bool&                 infeasible          /**< whether the node was determined to be infeasible */
   )
{
   assert( scip != NULL );
   assert( branchruledata != 0 );
   assert( branchruledata->vars != 0 );
   assert( branchruledata->varmap != 0 );

   /* init */
   branched = false;
   infeasible = false;

   /* check whether we want to avoid using orbital branching */
   if ( branchruledata->avoidsubtreenodes )
   {
      assert( branchruledata->subtreeboolconshdlr != 0 );
      if ( SUBTREEgetBool(branchruledata->subtreeboolconshdlr) )
      {
         SCIPdebugMessage("Avoid applying orbital branching, because we encountered a trivial stabilizer or did not find an orbit to branch on earlier.\n");
         return SCIP_OKAY;
      }
   }

   /* check whether we want to compute stabilizers */
   PermutationGroupPtr group;
   uintvec B1;
   try
   {
      if ( branchruledata->localsymmetry )
      {
         /* compute local symmetry information */
         SCIP_CALL( SYMgetLocalSymmetry(scip, branchruledata->symmetriesconshdlr, group) );
      }
      else
      {
         /* get branching variables */
         SCIP_CALL( computeBranchingVariables(scip, branchruledata->nvars, branchruledata->varmap, B1) );

         SCIPdebugMessage("Number of variables branched to 1: %lu.\n", (unsigned long) B1.size());
#ifdef SCIP_OUTPUT
         for (uintvec::const_iterator it = B1.begin(); it != B1.end(); ++it)
            std::cout << *it << " ";
         std::cout << std::endl;
#endif

         /* prepare stabilizer w.r.t. all variables branched to 1 */
         if ( B1.size() )
            group = PermutationGroupPtr(branchruledata->group->setStabilizer(B1));
         else
            group = branchruledata->group;
      }
   }
   catch (std::bad_alloc& ba)
   {
      SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Aborted orbital branching because we went out of memory.\n");
      return SCIP_OKAY;
   }

   if ( group == 0 || group->order() == 1 )
   {
      SCIPdebugMessage("Symmetry group is trivial. Skip orbital branching.\n");

      if ( branchruledata->avoidsubtreenodes )
      {
         /* store that we want to avoid orbital branching */
         assert( branchruledata->subtreeboolconshdlr != 0 );
         SCIP_CALL( SUBTREEsetTrue(scip, branchruledata->subtreeboolconshdlr) );
      }

      return SCIP_OKAY;
   }

   /* get all orbits */
   boost::scoped_ptr<OrbitList> O(group->orbits(*branchruledata->binaryvars));
   SCIPdebugMessage("Number of orbits: %lu.\n", (unsigned long) O->size());

   if ( SCIPisStopped(scip) )
      return SCIP_OKAY;

   /* choose orbit to branch on */
   OrbIt orbitToBranch;
   uintlist fixzero;
   uintlist fixone;
   switch ( branchruledata->orbitchoice )
   {
   case 0:
      SCIPdebugMessage("Choose largest orbit.\n");
      orbitToBranch = branchLargestOrbit(scip, branchruledata, O, branchLP, fixzero, fixone, infeasible);
      break;
   case 1:
      SCIPdebugMessage("Choose orbit that breaks symmetry.\n");
      orbitToBranch = branchSymmetry(scip, branchruledata, O, branchLP, false, B1.begin(), B1.begin() + B1.size(), fixzero, fixone, infeasible);
      break;
   case 2:
      SCIPdebugMessage("Choose orbit that keeps symmetry.\n");
      orbitToBranch = branchSymmetry(scip, branchruledata, O, branchLP, true, B1.begin(), B1.begin() + B1.size(), fixzero, fixone, infeasible);
      break;
   case 3:
      SCIPdebugMessage("Choose orbit that maximizes the product of its size and of its child.\n");
      orbitToBranch = branchMaxProduct(scip, branchruledata, O, branchLP, B1.begin(), B1.begin() + B1.size(), fixzero, fixone, infeasible);
      break;
   default:
      SCIPerrorMessage("Wrong orbit choice parameter: %d (allowed 0, 1, 2, 3).\n", branchruledata->orbitchoice);
      return SCIP_PARAMETERWRONGVAL;
   }

   /* exit if we detected infeasibility */
   if ( infeasible )
   {
      SCIPdebugMessage("Current node detected to be infeasible.\n");
      return SCIP_OKAY;
   }

   /* exit if no branching orbit could be found */
   if ( orbitToBranch == O->end() )
   {
      SCIPdebugMessage("Could not find orbit to branch on.\n");

      if ( branchruledata->avoidsubtreenodes )
      {
         /* store that we want to avoid orbital branching */
         assert( branchruledata->subtreeboolconshdlr != 0 );
         SCIP_CALL( SUBTREEsetTrue(scip, branchruledata->subtreeboolconshdlr) );
      }

      return SCIP_OKAY;
   }

#ifndef NDEBUG
   /* prepare output of orbit */
   char name[SCIP_MAXSTRLEN];
   char tmp[SCIP_MAXSTRLEN];
   name[0] = '\0';
#endif

   /* compute estimate for 0-branch: */
   SCIP_Real estimate = SCIPgetLocalLowerbound(scip);
   for (OrbitSetIt it = (*orbitToBranch).begin(); it != (*orbitToBranch).end(); ++it)
   {
      SCIP_VAR* var = branchruledata->vars[*it];
      estimate = MAX(estimate, SCIPcalcChildEstimate(scip, var, 0.0) );

#ifndef NDEBUG
      strncpy(tmp, name, SCIP_MAXSTRLEN-1);
      if ( branchLP )
      {
         (void) SCIPsnprintf(name, SCIP_MAXSTRLEN-1, "%s <%s> %u [%1f,%1f,%f]", tmp, SCIPvarGetName(var), *it,
            SCIPvarGetLbLocal(var), SCIPvarGetUbLocal(var), SCIPgetSolVal(scip, NULL, var));
      }
      else
         (void) SCIPsnprintf(name, SCIP_MAXSTRLEN-1, "%s <%s> %u [%1f,%1f]", tmp, SCIPvarGetName(var), *it, SCIPvarGetLbLocal(var), SCIPvarGetUbLocal(var));
#endif
   }
#ifndef NDEBUG
   SCIPdebugMessage("Branching on orbit:%s\n", name);
#endif

   /* branch on 0 */
   SCIP_Node* node;
   SCIP_CALL( SCIPcreateChild(scip, &node, 0.0, estimate) );
   for (OrbitSetIt it = (*orbitToBranch).begin(); it != (*orbitToBranch).end(); ++it)
   {
      SCIP_VAR* var = branchruledata->vars[*it];
      assert( SCIPvarGetUbLocal(var) > 0.5 );
      SCIP_CALL( SCIPchgVarUbNode(scip, node, var, 0.0) );
   }

   /* fix variables to 0 */
   for (uintlist::const_iterator fit = fixzero.begin(); fit != fixzero.end(); ++fit)
   {
      SCIP_VAR* var = branchruledata->vars[*fit];

      assert( SCIPvarGetLbLocal(var) < 0.5 );
      if ( SCIPvarGetUbLocal(var) > 0.5 )
      {
         SCIP_CALL( SCIPchgVarUbNode(scip, node, var, 0.0) );
         ++branchruledata->nfixedzero;
      }
   }

   /* fix variables to 1 */
   for (uintlist::const_iterator fit = fixone.begin(); fit != fixone.end(); ++fit)
   {
      SCIP_VAR* var = branchruledata->vars[*fit];

      assert( SCIPvarGetUbLocal(var) > 0.5 );
      if ( SCIPvarGetLbLocal(var) < 0.5 )
      {
         SCIP_CALL( SCIPchgVarLbNode(scip, node, var, 1.0) );
         ++branchruledata->nfixedone;
      }
   }

   /* take first unfixed variable in orbit as variable to branch to 1 */
   unsigned int idx = *((*orbitToBranch).begin());
   assert( idx < branchruledata->nvars );
   SCIP_VAR* branchvar = branchruledata->vars[idx];
   assert( SCIPvarGetLbLocal(branchvar) < 0.5 );    /* variable should not be fixed, by the choice above */

   SCIP_CALL( SCIPcreateChild(scip, &node, 0.0, SCIPcalcChildEstimate(scip, branchvar, 1.0)) );
   SCIP_CALL( SCIPchgVarLbNode(scip, node, branchvar, 1.0) );

   /* fix variables to 0 */
   for (uintlist::const_iterator fit = fixzero.begin(); fit != fixzero.end(); ++fit)
   {
      SCIP_VAR* var = branchruledata->vars[*fit];

      assert( SCIPvarGetLbLocal(var) < 0.5 );
      if ( SCIPvarGetUbLocal(var) > 0.5 )
      {
         SCIP_CALL( SCIPchgVarUbNode(scip, node, var, 0.0) );
      }
   }

   /* fix variables to 1 */
   for (uintlist::const_iterator fit = fixone.begin(); fit != fixone.end(); ++fit)
   {
      SCIP_VAR* var = branchruledata->vars[*fit];

      assert( SCIPvarGetUbLocal(var) > 0.5 );
      if ( SCIPvarGetLbLocal(var) < 0.5 )
      {
         SCIP_CALL( SCIPchgVarLbNode(scip, node, var, 1.0) );
      }
   }

   SCIPdebugMessage("Variables fixed to 0/1 in both branches: [%lu, %lu].\n", (unsigned long) fixzero.size(), (unsigned long) fixone.size());

   branched = true;

   return SCIP_OKAY;
}





/*
 * Callback methods of branching rule
 */


/** destructor of branching rule to free user data (called when SCIP is exiting) */
static
SCIP_DECL_BRANCHFREE(branchFreeOrbital)
{  /*lint --e{715,818}*/
   SCIP_BRANCHRULEDATA* branchruledata;

   assert( scip != 0 );
   assert( branchrule != 0 );

   branchruledata = SCIPbranchruleGetData(branchrule);
   assert( branchruledata != 0 );

   branchruledata->group.reset();
   branchruledata->binaryvars.reset();

   SCIPfreeMemory(scip, &branchruledata);

   return SCIP_OKAY;
}


/** Initialization method of branching rule (called after problem was transformed) */
SCIP_DECL_BRANCHINIT(branchInitOrbital)
{
   SCIP_BRANCHRULEDATA* branchruledata = 0;

   assert( branchrule != 0 );
   assert( strcmp(SCIPbranchruleGetName(branchrule), BRANCHRULE_NAME) == 0 );
   assert( scip != 0 );

   SCIPdebugMessage("Init of orbital branching ...\n");

   branchruledata = SCIPbranchruleGetData(branchrule);
   assert( branchruledata != 0 );
   SCIP_CALL( SCIPgetBoolParam(scip, "useorbital", &branchruledata->enabled) );

   if ( branchruledata->enabled )
   {
      /* work on binary variables while fixing integer variables */
      SYMsetProbSpecRequirement(branchruledata->symmetriesconshdlr, SCIP_PROBSPEC_BINARY);
      SYMsetProbSpecRequirementFixed(branchruledata->symmetriesconshdlr, SCIP_PROBSPEC_INTEGER);
   }

   return SCIP_OKAY;
}


/** deinitialization method of branching rule (called before transformed problem is freed) */
static
SCIP_DECL_BRANCHEXIT(branchExitOrbital)
{  /*lint --e{715}*/
   SCIP_BRANCHRULEDATA* branchruledata;

   assert( scip != 0 );
   assert( branchrule != 0 );

   branchruledata = SCIPbranchruleGetData(branchrule);
   assert( branchruledata != 0 );

   if ( branchruledata->enabled )
   {
      SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "\nNumber of variables fixed to 0 by orbital branching:\t%6d\n", branchruledata->nfixedzero);
      SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Number of variables fixed to 1 by orbital branching:\t%6d\n\n", branchruledata->nfixedone);
   }

   return SCIP_OKAY;
}


/** solving process initialization method of branching rule (called when branch and bound process is about to begin) */
SCIP_DECL_BRANCHINITSOL(branchInitsolOrbital)
{
   SCIP_BRANCHRULEDATA* branchruledata;

   assert( branchrule != 0 );
   assert( strcmp(SCIPbranchruleGetName(branchrule), BRANCHRULE_NAME) == 0 );
   assert( scip != 0 );

   branchruledata = SCIPbranchruleGetData(branchrule);
   assert( branchruledata != 0 );

   /* possibly skip orbital branching */
   if ( ! branchruledata->enabled )
      return SCIP_OKAY;

   SCIPdebugMessage("Initsol of orbital branching (obtain symmetry information) ...\n");

   /* get symmetry information */
   SCIP_CALL( SYMgetSymmetryData(scip, branchruledata->symmetriesconshdlr, branchruledata->group, branchruledata->nvars, branchruledata->vars,
         branchruledata->binaryvars, branchruledata->varmap) );

   /* possibly turn off */
   if ( ! branchruledata->group )
   {
      SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Skip orbital branching, since no symmetries were found.\n");
      branchruledata->enabled = FALSE;
   }

   return SCIP_OKAY;
}


/** branching execution method for fractional LP solutions */
static
SCIP_DECL_BRANCHEXECLP(branchExeclpOrbital)
{  /*lint --e{715}*/
   SCIP_BRANCHRULEDATA* branchruledata;

   assert( branchrule != 0 );
   assert( strcmp(SCIPbranchruleGetName(branchrule), BRANCHRULE_NAME) == 0 );
   assert( scip != 0 );
   assert( result != 0 );

   *result = SCIP_DIDNOTRUN;

   branchruledata = SCIPbranchruleGetData(branchrule);
   assert( branchruledata != 0 );

   /* exit if no or only trivial symmetry group is available */
   if ( branchruledata->group == 0 || branchruledata->group->order() == 1)
      return SCIP_OKAY;

   SCIPdebugMessage("Orbital branching (node: %lld, depth: %d) ...\n", SCIPnodeGetNumber(SCIPgetCurrentNode(scip)), SCIPgetDepth(scip));

   bool branched = false;
   bool infeasible = false;
   SCIP_CALL( branch(scip, branchruledata, true, branched, infeasible) );

   if ( infeasible )
      *result = SCIP_CUTOFF;
   else
   {
      if ( branched )
         *result = SCIP_BRANCHED;
   }

   return SCIP_OKAY;
}


/** branching execution method for not completely fixed pseudo solutions */
static
SCIP_DECL_BRANCHEXECPS(branchExecpsOrbital)
{  /*lint --e{715}*/
   SCIP_BRANCHRULEDATA* branchruledata;

   assert( branchrule != 0 );
   assert( strcmp(SCIPbranchruleGetName(branchrule), BRANCHRULE_NAME) == 0 );
   assert( scip != 0 );
   assert( result != 0 );

   SCIPdebugMessage("Orbital branching (node: %lld, depth: %d) ...\n", SCIPnodeGetNumber(SCIPgetCurrentNode(scip)), SCIPgetDepth(scip));

   *result = SCIP_DIDNOTRUN;

   branchruledata = SCIPbranchruleGetData(branchrule);
   assert( branchruledata != 0 );

   /* exit if no or only trivial symmetry group is available */
   if ( branchruledata->group == 0 || branchruledata->group->order() == 1)
      return SCIP_OKAY;

   bool branched = false;
   bool infeasible = false;
   SCIP_CALL( branch(scip, branchruledata, false, branched, infeasible) );

   if ( infeasible )
      *result = SCIP_CUTOFF;
   else
   {
      if ( branched )
         *result = SCIP_BRANCHED;
   }

   return SCIP_OKAY;
}


/*
 * branching rule specific interface methods
 */

/** creates the xyz branching rule and includes it in SCIP */
SCIP_RETCODE SCIPincludeBranchruleOrbital(
   SCIP*                 scip                /**< SCIP data structure */
   )
{
   SCIP_BRANCHRULEDATA* branchruledata = 0;

   /* create branching rule data */
   SCIP_CALL( SCIPallocMemory(scip, &branchruledata) );

   /* we must call the constructor explictly, because memory was m'alloced and not new'ed */
   new (&branchruledata->group) PermutationGroupPtr();
   new (&branchruledata->binaryvars) uintvecPtr();
   branchruledata->vars = 0;
   branchruledata->nvars = 0;
   branchruledata->varmap = 0;
   branchruledata->orbitchoice = -1;
   branchruledata->nfixedzero = 0;
   branchruledata->nfixedone = 0;
   branchruledata->avoidsubtreenodes = FALSE;
   branchruledata->enabled = TRUE;

   /* determine cons_subtreebool constraint handler */
   SCIP_CONSHDLR* conshdlr = SCIPfindConshdlr(scip, "subtreebool");
   if ( conshdlr == 0 )
   {
      SCIPerrorMessage("Could not find constraint handler <subtreebool>.\n");
      return SCIP_ERROR;
   }
   branchruledata->subtreeboolconshdlr = conshdlr;

   /* determine cons_symmetries constraint handler */
   conshdlr = SCIPfindConshdlr(scip, "symmetries");
   if ( conshdlr == 0 )
   {
      SCIPerrorMessage("Could not find constraint handler <symmetries>.\n");
      return SCIP_ERROR;
   }
   branchruledata->symmetriesconshdlr = conshdlr;

   /* include branching rule */
   SCIP_BRANCHRULE* branchrule = 0;
   SCIP_CALL( SCIPincludeBranchruleBasic(scip, &branchrule, BRANCHRULE_NAME, BRANCHRULE_DESC, BRANCHRULE_PRIORITY, BRANCHRULE_MAXDEPTH, BRANCHRULE_MAXBOUNDDIST, branchruledata) );

   /* set callbacks */
   SCIP_CALL( SCIPsetBranchruleFree(scip, branchrule, branchFreeOrbital) );
   SCIP_CALL( SCIPsetBranchruleInit(scip, branchrule, branchInitOrbital) );
   SCIP_CALL( SCIPsetBranchruleExit(scip, branchrule, branchExitOrbital) );
   SCIP_CALL( SCIPsetBranchruleInitsol(scip, branchrule, branchInitsolOrbital) );
   SCIP_CALL( SCIPsetBranchruleExecLp(scip, branchrule, branchExeclpOrbital) );
   SCIP_CALL( SCIPsetBranchruleExecPs(scip, branchrule, branchExecpsOrbital) );

   /* add orbital branching rule parameters */
   SCIP_CALL( SCIPaddIntParam(scip, "branching/orbital/orbitchoice",
         "choice of orbit to branch on (0: largest, 1: break symmetry, 2: keep symmetry, 3: max product)",
         &branchruledata->orbitchoice, TRUE, SCIP_DEFAULT_ORBITCHOICE, 0, 3, NULL, NULL) );

   SCIP_CALL( SCIPaddIntParam(scip, "branching/orbital/minimalorbitsize",
         "require a minimum orbit size of to branch on an orbit",
         &branchruledata->minimalorbitsize, TRUE, SCIP_DEFAULT_MINIMALORBITSIZE, 2, INT_MAX, NULL, NULL) );

   SCIP_CALL( SCIPaddBoolParam(scip, "branching/orbital/avoidsubtreenodes",
         "if symmetry group becomes trivial avoid orbital branching on sub-nodes?",
         &branchruledata->avoidsubtreenodes, TRUE, SCIP_DEFAULT_AVOIDSUBTREENODES, NULL, NULL) );

   SCIP_CALL( SCIPaddBoolParam(scip, "branching/orbital/localsymmetry",
         "Compute local symmetry group in each node?",
         &branchruledata->localsymmetry, TRUE, SCIP_DEFAULT_LOCALSYMMETRY, NULL, NULL) );

   return SCIP_OKAY;
}

// kate: space-indent on; indent-width 3; mixedindent off; indent-mode cstyle;
