/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*    This file is part of "mipsymmetries" a collection of                   */
/*    symmetry handling methods.                                             */
/*                                                                           */
/*    Copyright (C) 2010-2015  Marc Pfetsch, Thomas Rehn                     */
/*                                                                           */
/*                                                                           */
/*    Based on SCIP  --- Solving Constraint Integer Programs                 */
/*       SCIP is distributed under the terms of the SCIP Academic Licence,   */
/*       see file COPYING in the SCIP distribution.                          */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
#pragma ident "@(#) $Id: cons_subtreebool.c 267 2012-07-21 16:45:56Z pfetsch $"

/**@file   cons_subtreebool.c
 * @brief  constraint handler for storing a bool for subtrees
 * @author Marc Pfetsch
 *
 * We simply add a local constraint at a subtree to store a bool. This has the advantage that no additional bookkeeping
 * is needed and it is not important to add constraints at every node.
 */

/*---+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8----+----9----+----0----+----1----+----2*/

#include "cons_subtreebool.h"
#include <string.h>
#include <assert.h>

/* constraint handler properties */
#define CONSHDLR_NAME          "subtreebool"
#define CONSHDLR_DESC          "constraint handler for storing a bool for subtrees"
#define CONSHDLR_ENFOPRIORITY         0 /**< priority of the constraint handler for constraint enforcing */
#define CONSHDLR_CHECKPRIORITY        0 /**< priority of the constraint handler for checking feasibility */
#define CONSHDLR_EAGERFREQ          100 /**< frequency for using all instead of only the useful constraints in separation,
                                         *   propagation and enforcement, -1 for no eager evaluations, 0 for first only */
#define CONSHDLR_NEEDSCONS         TRUE /**< should the constraint handler be skipped, if no constraints are available? */



/*
 * Callback methods of constraint handler
 */

/** constraint enforcing method of constraint handler for LP solutions */
static
SCIP_DECL_CONSENFOLP(consEnfolpSubtreebool)
{  /*lint --e{715,818}*/
   assert( scip != NULL );
   assert( conshdlr != NULL );
   assert( strcmp(SCIPconshdlrGetName(conshdlr), CONSHDLR_NAME) == 0 );
   assert( result != NULL );

   /* do nothing */
   *result = SCIP_FEASIBLE;

   return SCIP_OKAY;
}


/** constraint enforcing method of constraint handler for pseudo solutions */
static
SCIP_DECL_CONSENFOPS(consEnfopsSubtreebool)
{  /*lint --e{715,818}*/
   assert( scip != NULL );
   assert( conshdlr != NULL );
   assert( strcmp(SCIPconshdlrGetName(conshdlr), CONSHDLR_NAME) == 0 );
   assert( result != NULL );

   /* do nothing */
   *result = SCIP_FEASIBLE;

   return SCIP_OKAY;
}


/** feasibility check method of constraint handler for primal solutions */
static
SCIP_DECL_CONSCHECK(consCheckSubtreebool)
{  /*lint --e{715,818}*/
   assert( scip != NULL );
   assert( conshdlr != NULL );
   assert( strcmp(SCIPconshdlrGetName(conshdlr), CONSHDLR_NAME) == 0 );
   assert( result != NULL );

   /* do nothing */
   *result = SCIP_FEASIBLE;

   return SCIP_OKAY;
}


/** variable rounding lock method of constraint handler */
static
SCIP_DECL_CONSLOCK(consLockSubtreebool)
{  /*lint --e{715,818}*/
   assert( scip != NULL );
   assert( conshdlr != NULL );
   assert( strcmp(SCIPconshdlrGetName(conshdlr), CONSHDLR_NAME) == 0 );

   /* do nothing */
   return SCIP_OKAY;
}


/** copy method for constraint handler plugins (called when SCIP copies plugins) */
static
SCIP_DECL_CONSHDLRCOPY(conshdlrCopySubtreebool)
{  /*lint --e{715,818}*/
   assert( scip != NULL );
   assert( conshdlr != NULL );
   assert( strcmp(SCIPconshdlrGetName(conshdlr), CONSHDLR_NAME) == 0 );
   assert( valid != NULL );

   /* call inclusion method of constraint handler */
   SCIP_CALL( SCIPincludeConshdlrSubtreebool(scip) );

   *valid = TRUE;

   return SCIP_OKAY;
}


/** constraint copying method of constraint handler */
static
SCIP_DECL_CONSCOPY(consCopySubtreebool)
{  /*lint --e{715,818}*/

   /* do not do anything: no subtreebool constraint should be present in the copy */
   return SCIP_OKAY;
}


/** include subtree bool constraint handler */
SCIP_RETCODE SCIPincludeConshdlrSubtreebool(
   SCIP*                 scip                /**< SCIP data structure */
   )
{
   SCIP_CONSHDLR* conshdlr = NULL;

   /* include constraint handler */
   SCIP_CALL( SCIPincludeConshdlrBasic(scip, &conshdlr, CONSHDLR_NAME, CONSHDLR_DESC,
         CONSHDLR_ENFOPRIORITY, CONSHDLR_CHECKPRIORITY, CONSHDLR_EAGERFREQ, CONSHDLR_NEEDSCONS,
         consEnfolpSubtreebool, consEnfopsSubtreebool, consCheckSubtreebool, consLockSubtreebool,
         NULL) );
   assert( conshdlr != NULL );

   SCIP_CALL( SCIPsetConshdlrCopy(scip, conshdlr, conshdlrCopySubtreebool, consCopySubtreebool) );

   return SCIP_OKAY;
}



/*
 * External functions
 */

/** determine whether bool is set for current subtree */
SCIP_Bool SUBTREEgetBool(
   SCIP_CONSHDLR*        conshdlr            /**< subtree bool constraint handler */
   )
{
   assert( conshdlr != NULL );
   assert( strcmp(SCIPconshdlrGetName(conshdlr), CONSHDLR_NAME) == 0 );

   /* It seems that if repropagation is active, the following assert is violated. */
   /* assert( SCIPconshdlrGetNActiveConss(conshdlr) >= 0 && SCIPconshdlrGetNActiveConss(conshdlr) <= 1 ); */
   assert( SCIPconshdlrGetNActiveConss(conshdlr) >= 0 );

   return ( SCIPconshdlrGetNActiveConss(conshdlr) > 0 );
}

/** set value for current subtree to true */
SCIP_RETCODE SUBTREEsetTrue(
   SCIP*                 scip,               /**< SCIP data structure */
   SCIP_CONSHDLR*        conshdlr            /**< subtree bool constraint handler */
   )
{
   assert( conshdlr != NULL );
   assert( strcmp(SCIPconshdlrGetName(conshdlr), CONSHDLR_NAME) == 0 );

   assert( SCIPconshdlrGetNActiveConss(conshdlr) >= 0 && SCIPconshdlrGetNActiveConss(conshdlr) <= 1 );

   /* only activate if we no already have a constraint */
   if ( SCIPconshdlrGetNActiveConss(conshdlr) == 0 )
   {
      SCIP_CONS* cons;
      char name[SCIP_MAXSTRLEN];

      (void) SCIPsnprintf(name, SCIP_MAXSTRLEN, "subtree"SCIP_LONGINT_FORMAT, SCIPnodeGetNumber(SCIPgetCurrentNode(scip)));

      /* create local constraint */
      SCIP_CALL( SCIPcreateCons(scip, &cons, name, conshdlr, NULL, FALSE, FALSE, FALSE, FALSE, FALSE, TRUE, FALSE, TRUE, FALSE, TRUE));
      SCIP_CALL( SCIPaddCons(scip, cons) );
      SCIP_CALL( SCIPreleaseCons(scip, &cons) );
   }

   return SCIP_OKAY;
}
