/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*    This file is part of "mipsymmetries" a collection of                   */
/*    symmetry handling methods.                                             */
/*                                                                           */
/*    Copyright (C) 2010-2015  Marc Pfetsch, Thomas Rehn                     */
/*                                                                           */
/*                                                                           */
/*    Based on SCIP  --- Solving Constraint Integer Programs                 */
/*       SCIP is distributed under the terms of the SCIP Academic Licence,   */
/*       see file COPYING in the SCIP distribution.                          */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
#pragma ident "@(#) $Id: cons_symmetries.c 267 2012-07-21 16:45:56Z pfetsch $"

/**@file   cons_symmetries.cpp
 * @brief  constraint handler for storing symmetry information about current problem
 * @author Marc Pfetsch
 * @author Thomas Rehn
 *
 * This constraint handler computes symmetries of the problem and stores this information in adequate form. It does not
 * perform additional actions.
 *
 * @note We treat implict integer variables as if they were continuous/real variables. The reason is that there is
 * currently no distinction between imilicit integer and imilicit binary and currently implicit integer variables hurt
 * our code more than continuous/real variables (we basically do not handle integral variables at all).
 */

/*---+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8----+----9----+----0----+----1----+----2*/

#include "cons_symmetries.h"

#include <permlib/abstract_symmetric_product.h>
#include <permlib/construct/known_bsgs_construction.h>
#include <permlib/test/type_recognition.h>

#include <handle_symmetries.h>

#include <unistd.h>
#include <sys/times.h>

#ifdef WITH_GMP
#include <gmpxx.h>
#include <sstream>
#endif


/* constraint handler properties */
#define CONSHDLR_NAME          "symmetries"
#define CONSHDLR_DESC          "constraint handler for storing symmetry information about current problem"
#define CONSHDLR_SEPAPRIORITY         0 /**< priority of the constraint handler for separation */
#define CONSHDLR_ENFOPRIORITY         0 /**< priority of the constraint handler for constraint enforcing */
#define CONSHDLR_CHECKPRIORITY -9000000 /**< priority of the constraint handler for checking feasibility */
#define CONSHDLR_SEPAFREQ            -1 /**< frequency for separating cuts; zero means to separate only in the root node */
#define CONSHDLR_PROPFREQ      -9000000 /**< frequency for propagating domains; zero means only preprocessing propagation */
#define CONSHDLR_EAGERFREQ          100 /**< frequency for using all instead of only the useful constraints in separation,
                                         *   propagation and enforcement, -1 for no eager evaluations, 0 for first only */
#define CONSHDLR_MAXPREROUNDS        -1 /**< maximal number of presolving rounds the constraint handler participates in (-1: no limit) */
#define CONSHDLR_DELAYSEPA        FALSE /**< should separation method be delayed, if other separators found cuts? */
#define CONSHDLR_DELAYPROP        FALSE /**< should propagation method be delayed, if other propagators found reductions? */
#define CONSHDLR_DELAYPRESOL      FALSE /**< should presolving method be delayed, if other presolvers found reductions? */
#define CONSHDLR_NEEDSCONS         TRUE /**< should the constraint handler be skipped, if no constraints are available? */

#define CONSHDLR_PROP_TIMING       SCIP_PROPTIMING_BEFORELP


/* default parameter values */
#define DEFAULT_USESYMHEUR        FALSE      /**< use heuristic symmetric subgroup detection (only symmetric groups) */
#define DEFAULT_USELAYEREDGRAPH   FALSE      /**< use layered graph to compute symmetries */
#define DEFAULT_MAXGENERATORS      1500      /**< limit on the number of generators that should be produced within symmetry detection (0 = no limit) */
#define DEFAULT_DETECTSYMPRESOL    TRUE      /**< Should the symmetry be detected within presolving (otherwise before presol)? */
#define DEFAULT_ANALYZECOMPONENTS FALSE      /**< Should the components of the symmetry group be analyzed (purely for informational purpose)? */
#define DEFAULT_PRODUCEGAPOUTPUT  FALSE      /**< Should the generators be output in GAP format? */

/* other defines */
#define MAXGENNUMERATOR        64000000      /**< determine maximal number of generators by dividing this number by the number of variables */


using namespace permlib;

/** constraint handler data */
struct SCIP_ConshdlrData
{
   SCIP_Bool             usesymheur;         /**< use heuristic symmetric subgroup detection (only symmetric groups) */
   SCIP_Bool             uselayeredgraph;    /**< use layered graph to compute symmetries */
   SCIP_Bool             detectsympresol;    /**< Should the symmetry be detected within presolving (otherwise before presol)? */
   SCIP_Bool             analyzecomponents;  /**< Should the components of the symmetry group be analyzed (purely for informational purpose)? */
   SCIP_Bool             producegapoutput;   /**< Should the generators be output in GAP format? */
   int                   maxgenerators;      /**< limit on the number of generators that should be produced within symmetry detection (0 = no limit) */
   int                   probspecrequire;    /**< problem specification for which we need to compute symmetries */
   int                   probspecrequirefixed; /**< problem specification of variables which must be fixed by symmetries */
};


/** constraint data */
struct SCIP_ConsData
{
   PermutationGroupPtr   group;              /**< symmetry group */
   unsigned int          nvars;              /**< number of variables taken into account */
   SCIP_VAR**            vars;               /**< array with variables taken into account */
   uintvecPtr            binaryvars;         /**< indices of binary variables taken into account (use vector for later use) */
   SCIP_HASHMAP*         varmap;             /**< map of variables to indices in vars array */
   SCIP_Bool             computedsym;        /**< have already tried to compute symmetries */
};


/*
 * Local methods
 */

/** determine symmetry group from permutations */
static
SCIP_RETCODE determineSymmetryGroupFromPerms(
   SCIP*                 scip,               /**< SCIP data structure */
   SCIP_CONSDATA*        consdata,           /**< symmetry constraint data */
   unsigned int          nvars,              /**< number of variables taken into account */
   SCIP_VAR**            vars,               /**< array with variables taken into account */
   unsigned int          nperms,             /**< number of permutations */
   unsigned int**        perms,              /**< generators of symmetry group: permutations as array of arrays (nperms x nvars) */
   unsigned int          nbasevars,          /**< number of base variables */
   unsigned int*         basevars,           /**< indices of variables that form a base for the strong generating set of permutations */
   SCIP_Bool             analyzeComponents   /**< whether components of the symmetry group should be analyzed for informational purpose */
   )
{
   assert( scip != 0 );
   assert( consdata != 0 );
   assert( consdata->vars == 0 );
   assert( consdata->varmap == 0 );

   /* check for empty information */
   if ( vars == 0 || nvars == 0 || nperms == 0 )
   {
      consdata->vars = 0;
      consdata->nvars = 0;
      return SCIP_OKAY;
   }

   /* copy variables */
   SCIP_CALL( SCIPduplicateBlockMemoryArray(scip, &consdata->vars, vars, nvars) );    /*lint !e866*/
   consdata->nvars = nvars;

   /* create symmetry group (if perms is available) */
   if ( perms )
   {
      boost::dynamic_bitset<> moved(nvars);
      std::list<PERMUTATION::ptr> generators;
      for (unsigned int i = 0; i < nperms; ++i)
      {
         PERMUTATION* p = new PERMUTATION(Permutation::perm(perms[i], perms[i] + nvars));
         for (unsigned int j = 0; j < nvars; ++j)
         {
            assert( perms[i][j] < nvars );
            if ( perms[i][j] != j )
               moved.set(j,1);
         }

#ifdef SCIP_OUTPUT
         std::cout << *p << "," << std::endl;
#endif
         generators.push_back(PERMUTATION::ptr(p));
      }
      assert( nbasevars > 0 );

      unsigned int nvarsmoved = moved.count();

      SCIPdebugMessage("begin BSGS construction ...\n");

      KnownBSGSConstruction<PERMUTATION, TRANSVERSAL> bsgsSetup(nvars);
      boost::shared_ptr<permlib::PermutationGroup> bsgs(new PermutationGroup(bsgsSetup.construct(generators.begin(), generators.end(), basevars, basevars + nbasevars) ));
      consdata->group = PermutationGroupPtr(new AbstractBSGS<TRANSVERSAL>(bsgs));

      SCIPdebugMessage("finished BSGS construction.\n");
      assert( consdata->group );

#ifdef SCIP_OUTPUT
      if ( consdata->group->order() != 1 )
         std::cout << *bsgs << std::endl;
      else
         std::cout << " found only trivial symmetry group :(" << std::endl;
#endif

#ifdef WITH_GMP
      mpz_class o = consdata->group->order<mpz_class>();
      size_t l = mpz_sizeinbase(o.get_mpz_t(), 10);

      /* output size if it is small enough */
      if ( l <= 6 )
      {
         std::stringstream ss;
         ss << o;
         SCIPdebugMessage("Order of symmetry group:\t\t%s\n\n", ss.str().c_str());
         SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Order of symmetry group:\t\t%s\n", ss.str().c_str());
      }

      /* compute log 10 of size: first exponentiate with 10 to get one digit after the decimal point */
      mpz_t s;
      mpz_init(s);
      mpz_pow_ui(s, o.get_mpz_t(), 10);
      size_t ll = mpz_sizeinbase(s, 10);
      SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Log10 of symmetry group order:\t\t%.1f\n", (double) ll/10.0);
      mpz_clear(s);
#else
      double s = consdata->group->order<double>();
      SCIPdebugMessage("Order of symmetry group:\t\t%.0f\n\n", s);
      if ( log10(s) <= 6.0 )
         SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Order of symmetry group:\t\t%.0f\n", s);
      SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Log10 of symmetry group order:\t\t%.1f\n", log10(s));
#endif
      SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Number of vars moved by some symmetry:\t%u\n", nvarsmoved);

      if ( analyzeComponents )
      {
         std::list<ComponentPtr> components;
         SCIP_CALL( determineComponents(bsgs, components) );

         BOOST_FOREACH(const ComponentPtr& component, components)
         {
            unsigned int componentDegree = 0;
            for (unsigned int i = 0; i < component->degree; ++i)
            {
               if ( component->bitset[i] )
                  ++componentDegree;
            }

            SCIPinfoMessage(scip, 0, "Component acting on %u points:\n", componentDegree);
            permlib::GroupType* componentType;
            std::vector<unsigned int> componentIndices;

            SCIP_CALL( analyzeComponent(component, componentType, componentIndices) );

            if ( componentType )
            {
               componentType->writeToStream(std::cout);
               delete componentType;
            }
            SCIPinfoMessage(scip, 0, "\n", componentDegree);
         }
      }
   }

   /* possibly get transformed variables */
   if ( SCIPisTransformed(scip) && nvars > 0 )
   {
      SCIP_CALL( SCIPgetTransformedVars(scip, nvars, consdata->vars, consdata->vars) );

      /* create hashmap for storing the indices of variables */
      SCIP_CALL( SCIPhashmapCreate(&consdata->varmap, SCIPblkmem(scip), SCIPcalcHashtableSize(10 * nvars)) );

      /* insert variables */
      for (unsigned int j = 0; j < nvars; ++j)
      {
         SCIP_CALL( SCIPhashmapInsert(consdata->varmap, consdata->vars[j], (void*) (size_t) j) );
      }

      /* get indices of binary variables */
      consdata->binaryvars->clear();
      for (unsigned int i = 0; i < nvars; ++i)
      {
         if ( SCIPvarGetType(vars[i]) == SCIP_VARTYPE_BINARY )
            consdata->binaryvars->push_back(i);
      }
   }

   return SCIP_OKAY;
}

/** determine local symmetry group from permutations */
static
SCIP_RETCODE determineLocalSymmetryGroupFromPerms(
   SCIP*                 scip,               /**< SCIP data structure */
   unsigned int          nvars,              /**< number of variables taken into account */
   unsigned int          nperms,             /**< number of permutations */
   unsigned int**        perms,              /**< generators of symmetry group: permutations as array of arrays (nperms x nvars) */
   unsigned int          nbasevars,          /**< number of base variables */
   unsigned int*         basevars,           /**< indices of variables that form a base for the strong generating set of permutations */
   PermutationGroupPtr&  group               /**< local symmetry group */
   )
{
   assert( scip != 0 );
   assert( perms != 0 );

   group.reset();

   if ( nperms == 0 )
      return SCIP_OKAY;

   boost::dynamic_bitset<> moved(nvars);
   std::list<PERMUTATION::ptr> generators;
   for (unsigned int i = 0; i < nperms; ++i)
   {
      PERMUTATION* p = new PERMUTATION(Permutation::perm(perms[i], perms[i] + nvars));
      for (unsigned int j = 0; j < nvars; ++j)
      {
         assert( perms[i][j] < nvars );
         if ( perms[i][j] != j )
            moved.set(j,1);
      }
      generators.push_back(PERMUTATION::ptr(p));
   }
   assert( nbasevars > 0 );

   SCIPdebugMessage("begin BSGS construction ...\n");

   KnownBSGSConstruction<PERMUTATION, TRANSVERSAL> bsgsSetup(nvars);
   boost::shared_ptr<permlib::PermutationGroup> bsgs(new PermutationGroup(bsgsSetup.construct(generators.begin(), generators.end(), basevars, basevars + nbasevars) ));
   group = PermutationGroupPtr(new AbstractBSGS<TRANSVERSAL>(bsgs));

   SCIPdebugMessage("finished BSGS construction.\n");
   assert( group );

#ifdef SCIP_OUTPUT
   if ( group->order() != 1 )
      std::cout << *bsgs << std::endl;
   else
      std::cout << " found only trivial symmetry group :(" << std::endl;
#endif

   return SCIP_OKAY;
}


/** compute heuristic symmetry group from orbits */
static
SCIP_RETCODE determineSymmetryGroupFromPermsHeur(
   SCIP*                 scip,               /**< SCIP data structure */
   SCIP_CONSDATA*        consdata,           /**< symmetry constraint data */
   unsigned int          nvars,              /**< number of variables taken into account */
   SCIP_VAR**            vars,               /**< array with variables taken into account */
   unsigned int*         orbits,             /**< indices of variables moved by a symmetric group; variables of the same orbit are grouped together in this list */
   unsigned int          norbits,            /**< number of entries in array orbits */
   unsigned int*         orbitsizes,         /**< list of orbit sizes into which the entries in the array orbits are grouped */
   unsigned int          norbitsizes,        /**< number of entries in array orbitsizes */
   SCIP_Bool             analyzeComponents   /**< whether components of the symmetry group should be analyzed for informational purpose */
   )
{
   assert( scip != 0 );
   assert( consdata != 0 );
   assert( consdata->vars == 0 );
   assert( consdata->varmap == 0 );

   /* check for empty information */
   if ( vars == 0 || nvars == 0 || norbitsizes == 0 )
   {
      consdata->vars = 0;
      consdata->nvars = 0;
      return SCIP_OKAY;
   }

   /* copy variables */
   SCIP_CALL( SCIPduplicateBlockMemoryArray(scip, &consdata->vars, vars, nvars) );    /*lint !e866*/
   consdata->nvars = nvars;

   unsigned int nvarsmoved = 0;
   std::vector<std::vector<unsigned int> > orbitsVector(norbitsizes);
   for (unsigned int i = 0; i < norbitsizes; ++i)
   {
      orbitsVector[i].resize( orbitsizes[i] );
      nvarsmoved += orbitsizes[i];
   }

   unsigned int c = 0;
   unsigned int orbIndex = 0;
   for (unsigned int i = 0; i < norbits; ++i)
   {
      assert( orbIndex < orbitsVector.size() );
      assert( c < orbitsVector[orbIndex].size() );

      orbitsVector[orbIndex][c++] = orbits[i];

      if (c == orbitsizes[orbIndex])
      {
         c = 0;
         ++orbIndex;
      }

      if ( orbIndex >= orbitsVector.size() )
         break;
   }

   /* create group */
   consdata->group = PermutationGroupPtr(new AbstractSymmetricProduct(orbitsVector.begin(), orbitsVector.end()));

   if ( consdata->group )
   {
#ifdef WITH_GMP
      mpz_class o = consdata->group->order<mpz_class>();
      size_t l = mpz_sizeinbase(o.get_mpz_t(), 10);

      /* output size if it is small enough */
      if ( l <= 6 )
      {
         std::stringstream ss;
         ss << o;
         SCIPdebugMessage("Order of symmetry group:\t\t%s\n\n", ss.str().c_str());
         SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Order of symmetry group:\t\t%s\n", ss.str().c_str());
      }

      /* compute log 10 of size: first exponentiate with 10 to get one digit after the decimal point */
      mpz_t s;
      mpz_init(s);
      mpz_pow_ui(s, o.get_mpz_t(), 10);
      size_t ll = mpz_sizeinbase(s, 10);
      SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Log10 of symmetry group order:\t\t%.1f\n", (double) ll/10.0);
      mpz_clear(s);
#else
      double s = consdata->group->order<double>();
      SCIPdebugMessage("Order of symmetry group:\t\t%.0f\n\n", s);
      if ( log10(s) <= 6.0 )
         SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Order of symmetry group:\t\t%.0f\n", s);
      SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Log10 of symmetry group order:\t\t%.1f\n", log10(s));
#endif
      SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Number of vars moved by some symmetry:\t%u\n", nvarsmoved);
   }

   /* possibly analyse components */
   if ( analyzeComponents )
   {
      BOOST_FOREACH(const std::vector<unsigned int>& orbit, orbitsVector)
      {
         SCIPinfoMessage(scip, 0, "Component acting on %u points:\nS_%u\n", orbit.size(), orbit.size());
      }
   }

   /* possibly get transformed variables */
   if ( SCIPisTransformed(scip) && nvars > 0 )
   {
      SCIP_CALL( SCIPgetTransformedVars(scip, nvars, consdata->vars, consdata->vars) );

      /* create hashmap for storing the indices of variables */
      SCIP_CALL( SCIPhashmapCreate(&consdata->varmap, SCIPblkmem(scip), SCIPcalcHashtableSize(10 * nvars)) );

      /* insert variables */
      for (unsigned int j = 0; j < nvars; ++j)
      {
         SCIP_CALL( SCIPhashmapInsert(consdata->varmap, consdata->vars[j], (void*) (size_t) j) );
      }

      /* get indices of binary variables */
      consdata->binaryvars->clear();
      for (unsigned int i = 0; i < nvars; ++i)
      {
         if ( SCIPvarGetType(vars[i]) == SCIP_VARTYPE_BINARY )
            consdata->binaryvars->push_back(i);
      }
   }

   return SCIP_OKAY;
}

/** compute heuristic local symmetry group from orbits */
static
SCIP_RETCODE determineLocalSymmetryGroupFromPermsHeur(
   SCIP*                 scip,               /**< SCIP data structure */
   unsigned int          nvars,              /**< number of variables taken into account */
   unsigned int*         orbits,             /**< indices of variables moved by a symmetric group; variables of the same orbit are grouped together in this list */
   unsigned int          norbits,            /**< number of entries in array orbits */
   unsigned int*         orbitsizes,         /**< list of orbit sizes into which the entries in the array orbits are grouped */
   unsigned int          norbitsizes,        /**< number of entries in array orbitsizes */
   PermutationGroupPtr&  group               /**< local symmetry group */
   )
{
   assert( scip != 0 );

   /* check for empty information */
   if ( nvars == 0 || norbitsizes == 0 )
      return SCIP_OKAY;

   unsigned int nvarsmoved = 0;
   std::vector<std::vector<unsigned int> > orbitsVector(norbitsizes);
   for (unsigned int i = 0; i < norbitsizes; ++i)
   {
      orbitsVector[i].resize( orbitsizes[i] );
      nvarsmoved += orbitsizes[i];
   }

   unsigned int c = 0;
   unsigned int orbIndex = 0;
   for (unsigned int i = 0; i < norbits; ++i)
   {
      assert( orbIndex < orbitsVector.size() );
      assert( c < orbitsVector[orbIndex].size() );

      orbitsVector[orbIndex][c++] = orbits[i];

      if (c == orbitsizes[orbIndex])
      {
         c = 0;
         ++orbIndex;
      }

      if ( orbIndex >= orbitsVector.size() )
         break;
   }

   /* create group */
   group = PermutationGroupPtr(new AbstractSymmetricProduct(orbitsVector.begin(), orbitsVector.end()));

   return SCIP_OKAY;
}

/** determine symmetry */
static
SCIP_RETCODE determineSymmetry(
   SCIP*                 scip,               /**< SCIP instance */
   SCIP_CONSHDLRDATA*    conshdlrdata,       /**< symmetries constraint handler data */
   SCIP_CONSDATA*        consdata            /**< symmetries constraint data */
   )
{
   assert( scip != 0 );
   assert( consdata != 0 );

   consdata->computedsym = TRUE;

   /* avoid trivial cases */
   if ( SCIPgetNVars(scip) <= 0 )
      return SCIP_OKAY;

   /* exit if group has been computed already */
   if ( consdata->group )
      return SCIP_OKAY;

   /* determine problem specification */
   int type = 0;
   if ( SCIPgetNBinVars(scip) > 0 )
      type |= (int) SCIP_PROBSPEC_BINARY;
   if ( SCIPgetNIntVars(scip) > 0 )
      type |= (int) SCIP_PROBSPEC_INTEGER;
   /* count implicit integer variables as real variables, since we cannot currently handle integral variables well */
   if ( SCIPgetNContVars(scip) > 0 || SCIPgetNImplVars(scip) > 0 )
      type |= (int) SCIP_PROBSPEC_REAL;

   /* skip symmetry computation if required variables are not present */
   if ( ! (type & conshdlrdata->probspecrequire) )
   {
      SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Skip symmetry computation, since type does not match requirements (%d bin, %d int, %d cont); required: (%d bin, %d int, %d cont).\n",
         SCIPgetNBinVars(scip), SCIPgetNIntVars(scip), SCIPgetNContVars(scip) + SCIPgetNImplVars(scip),
         (conshdlrdata->probspecrequire & SCIP_PROBSPEC_BINARY) != 0,
         (conshdlrdata->probspecrequire & SCIP_PROBSPEC_INTEGER) != 0,
         (conshdlrdata->probspecrequire & SCIP_PROBSPEC_REAL) != 0);
      return SCIP_OKAY;
   }

   SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Required symmetry:\t\t\t(%d bin, %d int, %d cont); (fixed: %d bin, %d int, %d cont)\n",
      (conshdlrdata->probspecrequire & SCIP_PROBSPEC_BINARY) != 0,
      (conshdlrdata->probspecrequire & SCIP_PROBSPEC_INTEGER) != 0,
      (conshdlrdata->probspecrequire & SCIP_PROBSPEC_REAL) != 0,
      (conshdlrdata->probspecrequirefixed & SCIP_PROBSPEC_BINARY) != 0,
      (conshdlrdata->probspecrequirefixed & SCIP_PROBSPEC_INTEGER) != 0,
      (conshdlrdata->probspecrequirefixed & SCIP_PROBSPEC_REAL) != 0);

   if ( conshdlrdata->probspecrequire & conshdlrdata->probspecrequirefixed )
      SCIPwarningMessage(scip, "Warning: some required symmetries must be fixed.\n");

   /* start to measure time */
   struct tms timer_beg;
   struct tms timer_end;
   times(&timer_beg);

   int maxgenerators = conshdlrdata->maxgenerators;

   /* reduce maximal number of generators depending on the number of variables */
   maxgenerators = MIN(maxgenerators, MAXGENNUMERATOR / SCIPgetNVars(scip));

   if ( conshdlrdata->usesymheur )
   {
      unsigned int* orbitsizes;
      unsigned int* orbits;
      unsigned int norbitsizes;
      unsigned int norbits;
      unsigned int nvars;
      SCIP_VAR** permvars;

      try
      {
         /* compute symmetry */
         SCIP_CALL( computeSymmetryByHeuristic(scip, conshdlrdata->uselayeredgraph, conshdlrdata->probspecrequirefixed, FALSE,
               &permvars, &nvars, &orbits, &norbits, &orbitsizes, &norbitsizes) );

         /* store symmetry group */
         SCIP_CALL( determineSymmetryGroupFromPermsHeur(scip, consdata, nvars, permvars, orbits, norbits, orbitsizes, norbitsizes, conshdlrdata->analyzecomponents) );
      }
      catch (std::bad_alloc& ba)
      {
         SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Aborted symmetry computation because we went out of memory.\n");
         /* try to catch bad memory allocation */
         consdata->group.reset();
      }

      /* free storage */
      SCIPfreeBlockMemoryArrayNull(scip, &orbitsizes, norbitsizes);
      SCIPfreeBlockMemoryArrayNull(scip, &orbits, norbits);
   }
   else
   {
      unsigned int** perms;
      unsigned int* baseindices;
      unsigned int nbaseindices;
      unsigned int npermvars;
      unsigned int nmaxperms;
      unsigned int nperms;
      unsigned int nbases;
      SCIP_VAR** permvars;

      try
      {
         /* compute symmetry */
         SCIP_CALL( computeSymmetry(scip, (unsigned int) maxgenerators, conshdlrdata->uselayeredgraph, conshdlrdata->probspecrequirefixed, FALSE,
               &permvars, &npermvars, &perms, &nperms, &nmaxperms, &baseindices, &nbaseindices, &nbases) );

         if ( conshdlrdata->producegapoutput && nperms > 0 )
         {
            char filename[SCIP_MAXSTRLEN];
            FILE* file;
            SCIP_Bool* covered;
            int* cycledec;

            if ( strlen(SCIPgetProbName(scip)) <= 0 )
               (int) SCIPsnprintf(filename, SCIP_MAXSTRLEN, "symmetries.g");
            else
               (int) SCIPsnprintf(filename, SCIP_MAXSTRLEN, "%s_symmetries.g", SCIPgetProbName(scip));

            file = fopen(filename, "w");
            if ( file == NULL )
            {
               SCIPerrorMessage("cannot create file <%s> for writing\n", filename);
               SCIPprintSysError(filename);
               return SCIP_FILECREATEERROR;
            }

            SCIP_CALL( SCIPallocBufferArray(scip, &covered, (int) npermvars) );
            SCIP_CALL( SCIPallocBufferArray(scip, &cycledec, (int) (2 * npermvars)) );

            SCIPinfoMessage(scip, file, "s := Group( ");
            for (unsigned int i = 0; i < nperms; ++i)
            {
               /* init cycle decomposition */
               for (unsigned int j = 0; j < npermvars; ++j)
               {
                  covered[j] = FALSE;
                  cycledec[j] = -2;
               }

               /* generate cycle decomposition */
               unsigned int k = 0;
               for (unsigned int j = 0; j < npermvars; ++j)
               {
                  if ( ! covered[j] )
                  {
                     /* avoid cycles of size 1 */
                     if ( j != perms[i][j] )
                     {
                        unsigned int p = j;
                        do
                        {
                           assert( p < npermvars );
                           assert( ! covered[p] );
                           cycledec[k++] = (int) p;
                           covered[p] = TRUE;
                           p = perms[i][p];
                        }
                        while ( p != j );
                        cycledec[k++] = -1;
                     }
                     else
                        covered[j] = TRUE;
                  }
               }

               if ( i > 0 )
                  SCIPinfoMessage(scip, file, ", ");

               k = 0;
               int size = 0;
               int ncycles = 0;
               while ( cycledec[k] != -2 )
               {
                  if ( cycledec[k] == -1 )
                  {
                     size = 0;
                     SCIPinfoMessage(scip, file, ")");
                     ++ncycles;
                  }
                  else
                  {
                     assert( 0 <= cycledec[k] );
                     if ( size == 0 )
                     {
                        SCIPinfoMessage(scip, file, "(");
                        SCIPinfoMessage(scip, file, "%d", cycledec[k] + 1);  /* gap starts counting at 1! */
                     }
                     else
                        SCIPinfoMessage(scip, file, ", %d", cycledec[k] + 1);  /* gap starts counting at 1! */
                     ++size;
                  }
                  ++k;
               }
            }
            SCIPinfoMessage(scip, file, ");\n\n");
            fclose(file);
            SCIPinfoMessage(scip, NULL, "\nWrote generators in GAP format to file <%s>.\n\n", filename);

            SCIPfreeBufferArray(scip, &cycledec);
            SCIPfreeBufferArray(scip, &covered);
         }

         /* store symmetry group */
         SCIP_CALL( determineSymmetryGroupFromPerms(scip, consdata, npermvars, permvars, nperms, perms, nbases, baseindices, conshdlrdata->analyzecomponents) );
      }
      catch (std::bad_alloc& ba)
      {
         SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Aborted symmetry computation because we went out of memory.\n");
         /* try to catch bad memory allocation */
         consdata->group.reset();
      }

      /* free storage */
      SCIPfreeBlockMemoryArrayNull(scip, &baseindices, nbaseindices);
      for (unsigned int i = 0; i < nperms; ++i)
         SCIPfreeBlockMemoryArrayNull(scip, &(perms[i]), npermvars);
      SCIPfreeBlockMemoryArrayNull(scip, &perms, nmaxperms);
   }

   if ( ! consdata->group )
   {
      SCIPdebugMessage("No symmetry found.\n");
      SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "No symmetry found.\n");
   }

   /* output time */
   times(&timer_end);
   double t = (timer_end.tms_utime - timer_beg.tms_utime) / (double)sysconf(_SC_CLK_TCK);
   SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Total time for symmetry detection:\t%4.2f\n\n", t);

   /* turn off some other presolving methods in order to be sure that they do not destroy symmetry afterwards */
   if ( consdata->group )
   {
      /* domcol avoids S_2-symmetries and may not be compatible with other symmetry handling methods. */
      SCIP_CALL( SCIPsetIntParam(scip, "presolving/domcol/maxrounds", 0) );
      SCIPinfoMessage(scip, 0, "Turned off presolver <domcol>.\n");

      /* components creates sub-SCIPs on which no symmetry handling is installed, thus turn this off. */
      SCIP_CALL( SCIPsetIntParam(scip, "presolving/components/maxrounds", 0) );
      SCIPinfoMessage(scip, 0, "Turned off presolver <components>.\n\n");
   }

   return SCIP_OKAY;
}

/** determine local symmetry */
static
SCIP_RETCODE determineLocalSymmetry(
   SCIP*                 scip,               /**< SCIP instance */
   SCIP_CONSHDLRDATA*    conshdlrdata,       /**< symmetries constraint handler data */
   PermutationGroupPtr&  group               /**< local symmetry group */
   )
{
   assert( scip != 0 );

   /* avoid trivial cases */
   if ( SCIPgetNVars(scip) <= 0 )
      return SCIP_OKAY;

   group.reset();

   /* determine problem specification */
   int type = 0;
   if ( SCIPgetNBinVars(scip) > 0 )
      type |= (int) SCIP_PROBSPEC_BINARY;
   if ( SCIPgetNIntVars(scip) > 0 )
      type |= (int) SCIP_PROBSPEC_INTEGER;
   /* count implicit integer variables as real variables, since we cannot currently handle integral variables well */
   if ( SCIPgetNContVars(scip) > 0 || SCIPgetNImplVars(scip) > 0 )
      type |= (int) SCIP_PROBSPEC_REAL;

   /* skip symmetry computation if required variables are not present */
   if ( ! (type & conshdlrdata->probspecrequire) )
   {
      SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Skip symmetry computation, since type does not match requirements (%d bin, %d int, %d cont); required: (%d bin, %d int, %d cont).\n",
         SCIPgetNBinVars(scip), SCIPgetNIntVars(scip), SCIPgetNContVars(scip) + SCIPgetNImplVars(scip),
         (conshdlrdata->probspecrequire & SCIP_PROBSPEC_BINARY) != 0,
         (conshdlrdata->probspecrequire & SCIP_PROBSPEC_INTEGER) != 0,
         (conshdlrdata->probspecrequire & SCIP_PROBSPEC_REAL) != 0);
      return SCIP_OKAY;
   }

   if ( conshdlrdata->probspecrequire & conshdlrdata->probspecrequirefixed )
      SCIPwarningMessage(scip, "Warning: some required symmetries must be fixed.\n");

   int maxgenerators = conshdlrdata->maxgenerators;

   /* reduce maximal number of generators depending on the number of variables */
   maxgenerators = MIN(maxgenerators, MAXGENNUMERATOR / SCIPgetNVars(scip));

   if ( conshdlrdata->usesymheur )
   {
      unsigned int* orbitsizes;
      unsigned int* orbits;
      unsigned int norbitsizes;
      unsigned int norbits;
      unsigned int nvars;
      SCIP_VAR** permvars;

      try
      {
         /* compute symmetry */
         SCIP_CALL( computeSymmetryByHeuristic(scip, conshdlrdata->uselayeredgraph, conshdlrdata->probspecrequirefixed, TRUE,
               &permvars, &nvars, &orbits, &norbits, &orbitsizes, &norbitsizes) );

         /* store symmetry group */
         SCIP_CALL( determineLocalSymmetryGroupFromPermsHeur(scip, nvars, orbits, norbits, orbitsizes, norbitsizes, group) );
      }
      catch (std::bad_alloc& ba)
      {
         SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Aborted symmetry computation because we went out of memory.\n");
         /* try to catch bad memory allocation */
         group.reset();
      }

      /* free storage */
      SCIPfreeBlockMemoryArrayNull(scip, &orbitsizes, norbitsizes);
      SCIPfreeBlockMemoryArrayNull(scip, &orbits, norbits);
   }
   else
   {
      unsigned int** perms;
      unsigned int* baseindices;
      unsigned int nbaseindices;
      unsigned int npermvars;
      unsigned int nmaxperms;
      unsigned int nperms;
      unsigned int nbases;
      SCIP_VAR** permvars;

      try
      {
         /* compute symmetry */
         SCIP_CALL( computeSymmetry(scip, (unsigned int) maxgenerators, conshdlrdata->uselayeredgraph, conshdlrdata->probspecrequirefixed, TRUE,
               &permvars, &npermvars, &perms, &nperms, &nmaxperms, &baseindices, &nbaseindices, &nbases) );

         SCIP_CALL( determineLocalSymmetryGroupFromPerms(scip, npermvars, nperms, perms, nbases, baseindices, group) );
      }
      catch (std::bad_alloc& ba)
      {
         SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Aborted symmetry computation because we went out of memory.\n");
         /* try to catch bad memory allocation */
         group.reset();
      }

      /* free storage */
      SCIPfreeBlockMemoryArrayNull(scip, &baseindices, nbaseindices);
      for (unsigned int i = 0; i < nperms; ++i)
         SCIPfreeBlockMemoryArrayNull(scip, &(perms[i]), npermvars);
      SCIPfreeBlockMemoryArrayNull(scip, &perms, nmaxperms);
   }

   if ( ! group )
   {
      SCIPdebugMessage("No symmetry found.\n");
   }

   return SCIP_OKAY;
}


/* creates a TypeRecognition object for a given component */
static
TypeRecognition<PERMUTATION, SchreierTreeTransversal<PERMUTATION> >
typeRecognition(
   const ComponentPtr& comp,
   std::vector<unsigned int>& componentIndices
)
{
   for (unsigned int i = 0; i < comp->degree; ++i)
   {
      if ( comp->bitset[i] )
         componentIndices.push_back(i);
   }

   const unsigned int n_comp = componentIndices.size();
   std::list<PERMUTATION::ptr> confinedOrbitPermutations;

   BOOST_FOREACH(const PERMUTATION::ptr& p, comp->generators)
   {
      PERMUTATION::ptr proj( p->project(n_comp, componentIndices.begin(), componentIndices.end()) );
      confinedOrbitPermutations.push_back( proj );
   }

   return
      TypeRecognition<PERMUTATION, SchreierTreeTransversal<PERMUTATION> >(n_comp, confinedOrbitPermutations.begin(), confinedOrbitPermutations.end());

}


/*
 * Callback methods of constraint handler
 */

/** destructor of constraint handler to free constraint handler data (called when SCIP is exiting) */
static
SCIP_DECL_CONSFREE(consFreeSymmetries)
{  /*lint --e{715}*/
   assert( scip != 0 );
   assert( conshdlr != 0 );
   assert( strcmp(SCIPconshdlrGetName(conshdlr), CONSHDLR_NAME) == 0 );

   SCIPdebugMessage("Deleting symmetries constraint handler.\n");

   SCIP_CONSHDLRDATA* conshdlrdata = SCIPconshdlrGetData(conshdlr);
   assert( conshdlrdata != 0 );
   SCIPfreeMemory(scip, &conshdlrdata);

   return SCIP_OKAY;
}


/** presolving initialization method of constraint handler (called when presolving is about to begin) */
static
SCIP_DECL_CONSINITPRE(consInitpreSymmetries)
{
   assert( scip != 0 );
   assert( conshdlr != 0 );
   assert( strcmp(SCIPconshdlrGetName(conshdlr), CONSHDLR_NAME) == 0 );

   SCIP_CONSHDLRDATA* conshdlrdata = SCIPconshdlrGetData(conshdlr);
   assert( conshdlrdata != 0 );

   /* compute symmetries if not requested during presolving */
   if ( ! conshdlrdata->detectsympresol )
   {
      for (int c = 0; c < nconss; ++c)
      {
         assert( conss[c] != 0 );
         SCIP_CONSDATA* consdata = SCIPconsGetData(conss[c]);
         assert( consdata != 0 );

         if ( consdata->computedsym )
            continue;

         SCIPdebugMessage("Initialization of symmetries constraint <%s>.\n", SCIPconsGetName(conss[c]));

         /* determine symmetry here in initpre, since other components specify their problem type in init() */
         SCIP_CALL( determineSymmetry(scip, conshdlrdata, consdata) );
      }
   }

   return SCIP_OKAY;
}


/** frees specific constraint data */
static
SCIP_DECL_CONSDELETE(consDeleteSymmetries)
{  /*lint --e{715}*/
   assert( scip != 0 );
   assert( conshdlr != 0 );
   assert( strcmp(SCIPconshdlrGetName(conshdlr), CONSHDLR_NAME) == 0 );
   assert( consdata != 0 );
   assert( cons != 0 );

   SCIPdebugMessage("Deleting symmetries constraint <%s>.\n", SCIPconsGetName(cons));

   SCIPfreeBlockMemoryArrayNull(scip, &(*consdata)->vars, (*consdata)->nvars);
   if ( (*consdata)->varmap != 0 )
   {
      SCIPhashmapFree(&(*consdata)->varmap);
   }

   (*consdata)->group.reset();
   (*consdata)->binaryvars.reset();

   SCIPfreeBlockMemory(scip, consdata);

   return SCIP_OKAY;
}


/** transforms constraint data into data belonging to the transformed problem */
static
SCIP_DECL_CONSTRANS(consTransSymmetries)
{
   assert( scip != 0 );
   assert( conshdlr != 0 );
   assert( strcmp(SCIPconshdlrGetName(conshdlr), CONSHDLR_NAME) == 0 );
   assert( sourcecons != 0 );

   SCIPdebugMessage("Transforming symmetries constraint <%s> ...\n", SCIPconsGetName(sourcecons));

   SCIP_CONSDATA* sourceconsdata = SCIPconsGetData(sourcecons);
   assert( sourceconsdata != 0 );
   SCIP_CONSDATA* targetconsdata;
   SCIP_CALL( SCIPallocBlockMemory(scip, &targetconsdata) );

   /* copy pointers to group */
   new (&targetconsdata->group) PermutationGroupPtr(sourceconsdata->group);
   new (&targetconsdata->binaryvars) uintvecPtr(sourceconsdata->binaryvars);
   targetconsdata->nvars = sourceconsdata->nvars;
   targetconsdata->varmap = 0;
   targetconsdata->vars = 0;
   targetconsdata->computedsym = sourceconsdata->computedsym;

   /* copy variables and set up variable map */
   if ( sourceconsdata->nvars > 0 )
   {
      assert( sourceconsdata->vars != 0 );
      SCIP_CALL( SCIPduplicateBlockMemoryArray(scip, &(targetconsdata->vars), sourceconsdata->vars, sourceconsdata->nvars) );
      SCIP_CALL( SCIPgetTransformedVars(scip, sourceconsdata->nvars, sourceconsdata->vars, targetconsdata->vars) );

      /* create hashmap for storing the indices of variables */
      SCIP_CALL( SCIPhashmapCreate(&targetconsdata->varmap, SCIPblkmem(scip), SCIPcalcHashtableSize(10 * sourceconsdata->nvars)) );

      /* insert variables and determine binary variables */
      targetconsdata->binaryvars->clear();
      for (unsigned int j = 0; j < targetconsdata->nvars; ++j)
      {
         SCIP_CALL( SCIPhashmapInsert(targetconsdata->varmap, targetconsdata->vars[j], (void*) (size_t) j) );
         if ( SCIPvarGetType(targetconsdata->vars[j]) == SCIP_VARTYPE_BINARY )
            targetconsdata->binaryvars->push_back(j);
      }
   }

   SCIP_CALL( SCIPcreateCons(scip, targetcons, SCIPconsGetName(sourcecons), conshdlr, targetconsdata,
         SCIPconsIsInitial(sourcecons), SCIPconsIsSeparated(sourcecons),
         SCIPconsIsEnforced(sourcecons), SCIPconsIsChecked(sourcecons),
         SCIPconsIsPropagated(sourcecons), SCIPconsIsLocal(sourcecons),
         SCIPconsIsModifiable(sourcecons), SCIPconsIsDynamic(sourcecons),
         SCIPconsIsRemovable(sourcecons), SCIPconsIsStickingAtNode(sourcecons)) );

   return SCIP_OKAY;
}


/** constraint enforcing method of constraint handler for LP solutions */
static
SCIP_DECL_CONSENFOLP(consEnfolpSymmetries)
{
   assert( scip != 0 );
   assert( conshdlr != 0 );
   assert( strcmp(SCIPconshdlrGetName(conshdlr), CONSHDLR_NAME) == 0 );
   assert( result != 0 );

   /* do nothing */
   *result = SCIP_FEASIBLE;

   return SCIP_OKAY;
}


/** constraint enforcing method of constraint handler for pseudo solutions */
static
SCIP_DECL_CONSENFOPS(consEnfopsSymmetries)
{
   assert( scip != 0 );
   assert( conshdlr != 0 );
   assert( strcmp(SCIPconshdlrGetName(conshdlr), CONSHDLR_NAME) == 0 );
   assert( result != 0 );

   /* do nothing */
   *result = SCIP_FEASIBLE;

   return SCIP_OKAY;
}


/** feasibility check method of constraint handler for primal solutions */
static
SCIP_DECL_CONSCHECK(consCheckSymmetries)
{
   assert( scip != 0 );
   assert( conshdlr != 0 );
   assert( strcmp(SCIPconshdlrGetName(conshdlr), CONSHDLR_NAME) == 0 );
   assert( result != 0 );

   /* do nothing */
   *result = SCIP_FEASIBLE;

   return SCIP_OKAY;
}


/** presolving method of constraint handler */
static
SCIP_DECL_CONSPRESOL(consPresolSymmetries)
{
   assert( scip != 0 );
   assert( conshdlr != 0 );
   assert( strcmp(SCIPconshdlrGetName(conshdlr), CONSHDLR_NAME) == 0 );
   assert( result != 0 );

   *result = SCIP_DIDNOTRUN;

   return SCIP_OKAY;
}


/** presolving deinitialization method of constraint handler (called after presolving has been finished) */
static
SCIP_DECL_CONSEXITPRE(consExitpreSymmetries)
{
   assert( scip != 0 );
   assert( conshdlr != 0 );
   assert( strcmp(SCIPconshdlrGetName(conshdlr), CONSHDLR_NAME) == 0 );

   /* skip if we are in a restart */
   if ( SCIPgetNRuns(scip) > 1 )
      return SCIP_OKAY;

   SCIPdebugMessage("Presolving method of symmetry constraint handler ...\n");

   SCIP_CONSHDLRDATA* conshdlrdata = SCIPconshdlrGetData(conshdlr);
   assert( conshdlrdata != 0 );

   /* compute symmetries if requested during presolving */
   if ( conshdlrdata->detectsympresol )
   {
      for (int c = 0; c < nconss; ++c)
      {
         assert( conss[c] != 0 );
         SCIP_CONSDATA* consdata = SCIPconsGetData(conss[c]);
         assert( consdata != 0 );

         if ( consdata->computedsym )
            continue;

         SCIPdebugMessage("Exitpre method of symmetries constraint <%s>.\n", SCIPconsGetName(conss[c]));

         SCIP_CALL( determineSymmetry(scip, conshdlrdata, consdata) );
      }
   }

   return SCIP_OKAY;
}


/** variable rounding lock method of constraint handler */
static
SCIP_DECL_CONSLOCK(consLockSymmetries)
{
   assert( scip != 0 );
   assert( conshdlr != 0 );
   assert( strcmp(SCIPconshdlrGetName(conshdlr), CONSHDLR_NAME) == 0 );

   /* do nothing */
   return SCIP_OKAY;
}


/** copy method for constraint handler plugins (called when SCIP copies plugins) */
static
SCIP_DECL_CONSHDLRCOPY(conshdlrCopySymmetries)
{
   assert( scip != 0 );
   assert( conshdlr != 0 );
   assert( strcmp(SCIPconshdlrGetName(conshdlr), CONSHDLR_NAME) == 0 );
   assert( valid != 0 );

   SCIP_CALL( SCIPincludeConshdlrSymmetries(scip) );
   *valid = TRUE;

   return SCIP_OKAY;
}


/** constraint copying method of constraint handler */
static
SCIP_DECL_CONSCOPY(consCopySymmetries)
{  /*lint --e{715}*/
   assert( scip != 0 );
   assert( cons != 0 );
   assert( sourcescip != 0 );
   assert( sourceconshdlr != 0 );
   assert( strcmp(SCIPconshdlrGetName(sourceconshdlr), CONSHDLR_NAME) == 0 );
   assert( valid != 0 );

   /* create empty symmetry constraint */
   SCIP_CALL( SCIPcreateConsSymmetriesPerms(scip, cons, "empty", 0, 0, 0, 0, 0, 0) );

   /* Do not copy, since we do not know how to efficiently copy symmetry information, but declare copy to be valid in
    * order to not disturb sub-SCIP heuristics. */
   *valid = TRUE;

   return SCIP_OKAY;
}


/** include store bool constraint handler */
extern
SCIP_RETCODE SCIPincludeConshdlrSymmetries(
   SCIP*                 scip                /**< SCIP data structure */
   )
{
   SCIP_CONSHDLRDATA* conshdlrdata;
   SCIP_CALL( SCIPallocMemory(scip, &conshdlrdata) );
   assert( conshdlrdata != 0 );
   conshdlrdata->probspecrequire = 0;
   conshdlrdata->probspecrequirefixed = 0;

   /* include constraint handler */
   SCIP_CONSHDLR* conshdlr = 0;
   SCIP_CALL( SCIPincludeConshdlrBasic(scip, &conshdlr, CONSHDLR_NAME, CONSHDLR_DESC,
         CONSHDLR_ENFOPRIORITY, CONSHDLR_CHECKPRIORITY, CONSHDLR_EAGERFREQ, CONSHDLR_NEEDSCONS,
         consEnfolpSymmetries, consEnfopsSymmetries, consCheckSymmetries, consLockSymmetries, conshdlrdata) );
   assert( conshdlr != 0 );

   SCIP_CALL( SCIPsetConshdlrFree(scip, conshdlr, consFreeSymmetries) );
   SCIP_CALL( SCIPsetConshdlrInitpre(scip, conshdlr, consInitpreSymmetries) );
   SCIP_CALL( SCIPsetConshdlrDelete(scip, conshdlr, consDeleteSymmetries) );
   SCIP_CALL( SCIPsetConshdlrCopy(scip, conshdlr, conshdlrCopySymmetries, consCopySymmetries) );
   SCIP_CALL( SCIPsetConshdlrTrans(scip, conshdlr, consTransSymmetries) );
#if SCIP_VERSION >= 320
   SCIP_CALL( SCIPsetConshdlrPresol(scip, conshdlr, consPresolSymmetries, CONSHDLR_MAXPREROUNDS, SCIP_PRESOLTIMING_EXHAUSTIVE) );
#else
   SCIP_CALL( SCIPsetConshdlrPresol(scip, conshdlr, consPresolSymmetries, CONSHDLR_MAXPREROUNDS, CONSHDLR_DELAYPRESOL) );
#endif
   SCIP_CALL( SCIPsetConshdlrExitpre(scip, conshdlr, consExitpreSymmetries) );

   SCIP_CALL( SCIPaddBoolParam(scip,
         "constraints/" CONSHDLR_NAME"/usesymheur", "use heuristic symmetric subgroup detection (only symmetric groups)",
         &conshdlrdata->usesymheur, TRUE, DEFAULT_USESYMHEUR, 0, 0) );

   SCIP_CALL( SCIPaddBoolParam(scip,
         "constraints/" CONSHDLR_NAME"/uselayeredgraph", "use layered graph to compute symmetries",
         &conshdlrdata->uselayeredgraph, TRUE, DEFAULT_USELAYEREDGRAPH, 0, 0) );

   SCIP_CALL( SCIPaddBoolParam(scip,
         "constraints/" CONSHDLR_NAME"/detectsympresol",
         "Should the symmetry be detected within presolving (otherwise before presol)?",
         &conshdlrdata->detectsympresol, TRUE, DEFAULT_DETECTSYMPRESOL, 0, 0) );

   SCIP_CALL( SCIPaddBoolParam(scip,
         "constraints/" CONSHDLR_NAME"/analyzecomponents",
         "Should the components of the symmetry group be analyzed (purely for informational purpose)?",
         &conshdlrdata->analyzecomponents, TRUE, DEFAULT_ANALYZECOMPONENTS, 0, 0) );

   SCIP_CALL( SCIPaddBoolParam(scip,
         "constraints/" CONSHDLR_NAME"/producegapoutput",
         "Should the generators be output in GAP format?",
         &conshdlrdata->producegapoutput, TRUE, DEFAULT_PRODUCEGAPOUTPUT, 0, 0) );

   SCIP_CALL( SCIPaddIntParam(scip,
         "constraints/" CONSHDLR_NAME"/maxgenerators", "limit on the number of generators that should be produced within symmetry detection (0 = no limit)",
         &conshdlrdata->maxgenerators, TRUE, DEFAULT_MAXGENERATORS, 0, INT_MAX, 0, 0) );

   return SCIP_OKAY;
}



/*
 * External methods
 */

/** create symmetries constraint */
SCIP_RETCODE SCIPcreateConsSymmetries(
   SCIP*                 scip,               /**< SCIP data structure */
   SCIP_CONS**           cons,               /**< pointer to hold the created constraint */
   const char*           name                /**< name of constraint */
   )
{
   assert( scip != 0 );

   /* find the symmetries constraint handler */
   SCIP_CONSHDLR* conshdlr = SCIPfindConshdlr(scip, CONSHDLR_NAME);
   if ( conshdlr == 0 )
   {
      SCIPerrorMessage("symmetries constraint handler not found\n");
      return SCIP_PLUGINNOTFOUND;
   }

   /* create constraint data */
   SCIP_CONSDATA* consdata = 0;
   SCIP_CALL( SCIPallocBlockMemory(scip, &consdata) );
   new (&consdata->group) PermutationGroupPtr();
   new (&consdata->binaryvars) uintvecPtr(new std::vector<unsigned int>());
   consdata->vars = 0;
   consdata->nvars = 0;
   consdata->varmap = 0;
   consdata->computedsym = FALSE;

   SCIPdebugMessage("Creating symmetries constraint <%s>.\n", name);

   /* create constraint */
   SCIP_CALL( SCIPcreateCons(scip, cons, name, conshdlr, consdata, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE));

   return SCIP_OKAY;
}


/** create symmetries constraint */
SCIP_RETCODE SCIPcreateConsSymmetriesPerms(
   SCIP*                 scip,               /**< SCIP data structure */
   SCIP_CONS**           cons,               /**< pointer to hold the created constraint */
   const char*           name,               /**< name of constraint */
   unsigned int          nvars,              /**< number of variables taken into account */
   SCIP_VAR**            vars,               /**< array with variables taken into account */
   unsigned int          nperms,             /**< number of permutations */
   unsigned int**        perms,              /**< generators of symmetry group: permutations as array of arrays (nperms x nvars) */
   unsigned int          nbasevars,          /**< number of base variables */
   unsigned int*         basevars            /**< indices of variables that form a base for the strong generating set of permutations */
   )
{
   /* find the symmetries constraint handler */
   SCIP_CONSHDLR* conshdlr = SCIPfindConshdlr(scip, CONSHDLR_NAME);
   if ( conshdlr == 0 )
   {
      SCIPerrorMessage("symmetries constraint handler not found\n");
      return SCIP_PLUGINNOTFOUND;
   }

   /* create constraint data */
   SCIP_CONSDATA* consdata = 0;
   SCIP_CALL( SCIPallocBlockMemory(scip, &consdata) );
   new (&consdata->group) PermutationGroupPtr();
   new (&consdata->binaryvars) uintvecPtr(new std::vector<unsigned int>());
   consdata->vars = 0;
   consdata->nvars = 0;
   consdata->varmap = 0;
   consdata->computedsym = FALSE;

   SCIP_CONSHDLRDATA* conshdlrdata = SCIPconshdlrGetData(conshdlr);
   assert( conshdlrdata != 0 );

   /* compute symmetry group */
   SCIP_CALL( determineSymmetryGroupFromPerms(scip, consdata, nvars, vars, nperms, perms, nbasevars, basevars, conshdlrdata->analyzecomponents) );
   consdata->computedsym = TRUE;

   SCIPdebugMessage("Creating symmetries constraint <%s>.\n", name);

   /* create constraint */
   SCIP_CALL( SCIPcreateCons(scip, cons, name, conshdlr, consdata, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE));

   return SCIP_OKAY;
}


/** return symmetry group information */
PermutationGroupPtr SYMgetSymmetryGroup(
   SCIP_CONSHDLR*        conshdlr            /**< symmetry constraint handler */
   )
{
   assert( conshdlr != 0 );
   assert( strcmp(SCIPconshdlrGetName(conshdlr), CONSHDLR_NAME) == 0 );

   SCIP_CONS** conss = SCIPconshdlrGetConss(conshdlr);
   int nconss = SCIPconshdlrGetNConss(conshdlr);

   if ( nconss <= 0 )
   {
      return PermutationGroupPtr();
   }

   /* arbitrarly get first constraint */
   SCIP_CONSDATA* consdata = SCIPconsGetData(conss[0]);

   return consdata->group;
}


/** return symmetry information */
SCIP_RETCODE SYMgetSymmetryData(
   SCIP*                 scip,               /**< SCIP instancs */
   SCIP_CONSHDLR*        conshdlr,           /**< symmetry constraint handler */
   PermutationGroupPtr&  group,              /**< symmetry group */
   unsigned int&         nvars,              /**< number of variables taken into account */
   SCIP_VAR**&           vars,               /**< array with variables taken into account */
   uintvecPtr&           binaryvars,         /**< indices of binary variables taken into account (use vector for later use) */
   SCIP_HASHMAP*&        varmap              /**< map of variables to indices in vars array */
   )
{
   assert( conshdlr != 0 );
   assert( strcmp(SCIPconshdlrGetName(conshdlr), CONSHDLR_NAME) == 0 );

   SCIP_CONS** conss = SCIPconshdlrGetConss(conshdlr);
   int nconss = SCIPconshdlrGetNConss(conshdlr);

   if ( nconss <= 0 )
   {
      SCIPerrorMessage("Could not find symmetry handling constraint.\n");
      return SCIP_PLUGINNOTFOUND;
   }

   /* arbitrarly get first constraint */
   SCIP_CONSDATA* consdata = SCIPconsGetData(conss[0]);

   if ( ! consdata->computedsym )
   {
      if ( SCIPgetStage(scip) != SCIP_STAGE_PRESOLVING )
      {
         SCIPerrorMessage("Cannot call symmetry detection outside of presolving.\n");
         return SCIP_INVALIDCALL;
      }

      SCIP_CONSHDLRDATA* conshdlrdata = SCIPconshdlrGetData(conshdlr);
      assert( conshdlrdata != 0 );

      /* determine symmetry here */
      SCIP_CALL( determineSymmetry(scip, conshdlrdata, consdata) );
   }

   group = consdata->group;
   vars = consdata->vars;
   nvars = consdata->nvars;
   binaryvars = consdata->binaryvars;
   varmap = consdata->varmap;

   return SCIP_OKAY;
}

/** return symmetry information */
SCIP_RETCODE SYMgetLocalSymmetry(
   SCIP*                 scip,               /**< SCIP instancs */
   SCIP_CONSHDLR*        conshdlr,           /**< symmetry constraint handler */
   PermutationGroupPtr&  group               /**< symmetry group */
   )
{
   assert( conshdlr != 0 );
   assert( strcmp(SCIPconshdlrGetName(conshdlr), CONSHDLR_NAME) == 0 );

   SCIP_CONSHDLRDATA* conshdlrdata = SCIPconshdlrGetData(conshdlr);
   assert( conshdlrdata != 0 );

   SCIP_CALL( determineLocalSymmetry(scip, conshdlrdata, group) );

   return SCIP_OKAY;
}

/** replace symmetry group by the stabilizer over the given set of variable indices */
extern
SCIP_RETCODE SYMreplaceSymmetryWithStabilizer(
   SCIP*                 scip,               /**< SCIP instancs */
   SCIP_CONSHDLR*        conshdlr,           /**< symmetry constraint handler */
   uintvec&              set                 /**< indices of variables for taking the stabilizer */
   )
{
   assert( conshdlr != 0 );
   assert( strcmp(SCIPconshdlrGetName(conshdlr), CONSHDLR_NAME) == 0 );

   if ( set.size() > 0 )
   {
      SCIP_CONS** conss = SCIPconshdlrGetConss(conshdlr);
      int nconss = SCIPconshdlrGetNConss(conshdlr);

      if ( nconss <= 0 )
         return SCIP_OKAY;

      /* arbitrarly get first constraint */
      SCIP_CONSDATA* consdata = SCIPconsGetData(conss[0]);

      consdata->group = PermutationGroupPtr(consdata->group->setStabilizer(set));
   }

   return SCIP_OKAY;
}

/** specify problem type for which we need symmetries */
void SYMsetProbSpecRequirement(
   SCIP_CONSHDLR*        conshdlr,           /**< symmetry constraint handler */
   SCIP_PROBSPEC         type                /**< variable types the callee is interested in */
   )
{
   assert( conshdlr != 0 );
   assert( strcmp(SCIPconshdlrGetName(conshdlr), CONSHDLR_NAME) == 0 );

   SCIP_CONSHDLRDATA* conshdlrdata = SCIPconshdlrGetData(conshdlr);
   assert( conshdlrdata != 0 );

   conshdlrdata->probspecrequire |= static_cast<int>(type);
}

/** specify problem type which symmetries must fix */
void SYMsetProbSpecRequirementFixed(
   SCIP_CONSHDLR*        conshdlr,           /**< symmetry constraint handler */
   SCIP_PROBSPEC         fixedtype           /**< variable types that callee wants to have fixed */
   )
{
   assert( conshdlr != 0 );
   assert( strcmp(SCIPconshdlrGetName(conshdlr), CONSHDLR_NAME) == 0 );

   SCIP_CONSHDLRDATA* conshdlrdata = SCIPconshdlrGetData(conshdlr);
   assert( conshdlrdata != 0 );

   conshdlrdata->probspecrequirefixed |= static_cast<int>(fixedtype);
}

/** whether symmetry should be computed for presolved system */
SCIP_Bool SYMdetectSymmetriesPresolved(
   SCIP_CONSHDLR*        conshdlr            /**< symmetry constraint handler */
   )
{
   assert( conshdlr != 0 );
   assert( strcmp(SCIPconshdlrGetName(conshdlr), CONSHDLR_NAME) == 0 );

   SCIP_CONSHDLRDATA* conshdlrdata = SCIPconshdlrGetData(conshdlr);
   assert( conshdlrdata != 0 );

   return conshdlrdata->detectsympresol;
}

/** Compute independent component of groups
 *
 *  The group can be written as a direct product of these components.
 */
SCIP_RETCODE determineComponents(
   const boost::shared_ptr<permlib::PermutationGroup>& group, /**< symmetry group */
   std::list<ComponentPtr>& components                        /**< components on output */
   )
{
   const PermutationGroup& bsgs = *group;

   std::vector< PERMUTATION::ptr > groupPermutations(bsgs.S.begin(), bsgs.S.end());

   BOOST_FOREACH(const PERMUTATION::ptr& p, groupPermutations)
   {
      boost::dynamic_bitset<> pSupport(bsgs.n);
      for (unsigned int i = 0; i < bsgs.n; ++i)
      {
         if ( *p / i != i )
            pSupport.set(i, 1);
      }

      bool merged = false;
      std::list<ComponentPtr>::iterator mergePosition = components.end();
      for (std::list<ComponentPtr>::iterator it = components.begin(); it != components.end(); /* increment in loop */ )
      {
         boost::dynamic_bitset<> mergedSupport(pSupport & (*it)->bitset);

         if ( mergedSupport.any() )
         {
            if ( mergePosition == components.end() )
            {
               mergePosition = it;
               (*it)->bitset |= pSupport;
               (*it)->generators.push_back(p);
               ++it;
            }
            else
            {
               BOOST_ASSERT( mergePosition != it );
               (*mergePosition)->bitset |= (*it)->bitset;
               (*mergePosition)->generators.insert((*mergePosition)->generators.end(), (*it)->generators.begin(), (*it)->generators.end());
               it = components.erase(it);
            }
            merged = true;
         }
         else
         {
            ++it;
         }
      }
      if (!merged)
      {
         ComponentPtr c(new Component(pSupport, p, bsgs.n));
         components.push_back(c);
      }
   }

   return SCIP_OKAY;
}

/** analyze group type of a given component of a symmetry group */
SCIP_RETCODE analyzeComponent(
   const ComponentPtr&   comp,               /**< component to be analyzed */
   permlib::GroupType*&  componentType,      /**< return type of component */
   std::vector<unsigned int>& componentIndices /**< variable indices for component */
   )
{
   TypeRecognition<PERMUTATION, SchreierTreeTransversal<PERMUTATION> >
      componentRecognition = typeRecognition(comp, componentIndices);
   componentType = componentRecognition.analyze();
   assert( componentType );

#ifdef SCIP_OUTPUT
   // FIXME
   for (unsigned int i = 0; i < comp->degree; ++i)
   {
      if ( comp->bitset[i] )
         SCIPdebugPrintf("%d, ", i+1);
   }
   SCIPdebugPrintf("\n");
   SCIPdebugPrintf(" type: ");
   componentType->writeToStream(std::cout);
   SCIPdebugPrintf("\n");
#endif
   return SCIP_OKAY;
}

/** extract large symmetric diagonal subgroup form component */
permlib::GroupType* largeSymmetricDiagonalSubgroup(
   const ComponentPtr&   comp,               /**< component to be investigated */
   std::vector<unsigned int> orbitCharacteristic /**< characteristic vector of orbit */
   )
{
   std::vector<unsigned int> componentIndices;
   TypeRecognition<PERMUTATION, SchreierTreeTransversal<PERMUTATION> >
      componentRecognition = typeRecognition(comp, componentIndices);
   return componentRecognition.largeSymmetricDiagonalSubgroup(orbitCharacteristic);
}

// kate: space-indent on; indent-width 3; mixedindent off; indent-mode cstyle;
