// -*- C++ -*-
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*    This file is part of "mipsymmetries" a collection of                   */
/*    symmetry handling methods.                                             */
/*                                                                           */
/*    Copyright (C) 2010-2015  Marc Pfetsch, Thomas Rehn                     */
/*                                                                           */
/*                                                                           */
/*    Based on SCIP  --- Solving Constraint Integer Programs                 */
/*       SCIP is distributed under the terms of the SCIP Academic Licence,   */
/*       see file COPYING in the SCIP distribution.                          */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
#pragma ident "@(#) $Id: cons_symmetries.h 267 2012-07-21 16:45:56Z pfetsch $"

/**@file   cons_symmetries.h
 * @brief  constraint handler for storing symmetry information about current problem
 * @author Marc Pfetsch
 * @author Thomas Rehn
 *
 * This constraint handler computes symmetries of the problem and stores this information in adequate form. It does not
 * perform additional actions.
 */

/*---+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8----+----9----+----0----+----1----+----2*/

#ifndef __SCIP_CONS_SYMMETRIES_H_
#define __SCIP_CONS_SYMMETRIES_H_

#include <scip/scip.h>
#include <permlib/permlib_api.h>
#include <permlib/abstract_bsgs.h>
#include <permlib/test/group_type.h>
#include <boost/shared_ptr.hpp>

#include "type_probspec.h"


/* useful typedefs */
typedef boost::shared_ptr<permlib::AbstractPermutationGroup> PermutationGroupPtr;
typedef std::vector<unsigned int> uintvec;
typedef boost::shared_ptr<uintvec> uintvecPtr;

/** include symmetry constraint handler */
extern
SCIP_RETCODE SCIPincludeConshdlrSymmetries(
   SCIP*                 scip                /**< SCIP data structure */
   );

/** create symmetries constraint */
extern
SCIP_RETCODE SCIPcreateConsSymmetries(
   SCIP*                 scip,               /**< SCIP data structure */
   SCIP_CONS**           cons,               /**< pointer to hold the created constraint */
   const char*           name                /**< name of constraint */
   );

/** create symmetries constraint */
extern
SCIP_RETCODE SCIPcreateConsSymmetriesPerms(
   SCIP*                 scip,               /**< SCIP data structure */
   SCIP_CONS**           cons,               /**< pointer to hold the created constraint */
   const char*           name,               /**< name of constraint */
   unsigned int          nvars,              /**< number of variables taken into account */
   SCIP_VAR**            vars,               /**< array with variables taken into account */
   unsigned int          nperms,             /**< number of permutations */
   unsigned int**        perms,              /**< generators of symmetry group: permutations as array of arrays (nperms x nvars) */
   unsigned int          nbasevars,          /**< number of base variables */
   unsigned int*         basevars            /**< indices of variables that form a base for the strong generating set of permutations */
   );

/** return symmetry group information */
extern
PermutationGroupPtr SYMgetSymmetryGroup(
   SCIP_CONSHDLR*        conshdlr            /**< symmetry constraint handler */
   );

/** return symmetry information */
extern
SCIP_RETCODE SYMgetSymmetryData(
   SCIP*                 scip,               /**< SCIP instancs */
   SCIP_CONSHDLR*        conshdlr,           /**< symmetry constraint handler */
   PermutationGroupPtr&  group,              /**< symmetry group */
   unsigned int&         nvars,              /**< number of variables taken into account */
   SCIP_VAR**&           vars,               /**< array with variables taken into account */
   uintvecPtr&           binaryvars,         /**< indices of binary variables taken into account (use vector for later use) */
   SCIP_HASHMAP*&        varmap              /**< map of variables to indices in vars array */
   );

/** return local symmetry information */
extern
SCIP_RETCODE SYMgetLocalSymmetry(
   SCIP*                 scip,               /**< SCIP instancs */
   SCIP_CONSHDLR*        conshdlr,           /**< symmetry constraint handler */
   PermutationGroupPtr&  group               /**< symmetry group */
   );

/** replace symmetry group by the stabilizer over the given set of variable indices */
extern
SCIP_RETCODE SYMreplaceSymmetryWithStabilizer(
   SCIP*                 scip,               /**< SCIP instancs */
   SCIP_CONSHDLR*        conshdlr,           /**< symmetry constraint handler */
   uintvec&              set                 /**< indices of variables for taking the stabilizer */
   );

/** specify problem type for which we need symmetries */
extern
void SYMsetProbSpecRequirement(
   SCIP_CONSHDLR*        conshdlr,           /**< symmetry constraint handler */
   SCIP_PROBSPEC         type                /**< variable types the callee is interested in */
   );

/** specify problem type which symmetries must fix */
extern
void SYMsetProbSpecRequirementFixed(
   SCIP_CONSHDLR*        conshdlr,           /**< symmetry constraint handler */
   SCIP_PROBSPEC         fixedtype           /**< variable types that callee wants to have fixed */
   );

/** whether symmetry should be computed for presolved system */
extern
SCIP_Bool SYMdetectSymmetriesPresolved(
   SCIP_CONSHDLR*        conshdlr            /**< symmetry constraint handler */
   );



/** @struct Component
 *  @brief  Symmetry component
 */
struct Component
{
   boost::dynamic_bitset<> bitset;
   std::list<permlib::PERMUTATION::ptr> generators;
   unsigned int degree;

   /** constructor */
   Component(const boost::dynamic_bitset<>& b, const permlib::PERMUTATION::ptr& p, unsigned int degree_) :
      bitset(b), degree(degree_)
   {
      generators.push_back(p);
   }
};
typedef boost::shared_ptr<Component> ComponentPtr;

/** Compute independent component of groups
 *
 *  The group can be written as a direct product of these components.
 */
extern
SCIP_RETCODE determineComponents(
   const boost::shared_ptr<permlib::PermutationGroup>& group, /**< symmetry group */
   std::list<ComponentPtr>& components                        /**< components on output */
   );

/** analyze group type of a given component of a symmetry group */
extern
SCIP_RETCODE analyzeComponent(
   const ComponentPtr&   comp,               /**< component to be analyzed */
   permlib::GroupType*&  componentType,      /**< return type of component */
   std::vector<unsigned int>& componentIndices /**< variable indices for component */
   );

/** extract large symmetric diagonal subgroup form component */
extern
permlib::GroupType* largeSymmetricDiagonalSubgroup(
   const ComponentPtr&   comp,               /**< component to be investigated */
   std::vector<unsigned int> orbitCharacteristic /**< characteristic vector of orbit */
   );

#endif
