/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*    This file is part of "mipsymmetries" a collection of                   */
/*    symmetry handling methods.                                             */
/*                                                                           */
/*    Copyright (C) 2010-2015  Marc Pfetsch, Thomas Rehn                     */
/*                                                                           */
/*                                                                           */
/*    Based on SCIP  --- Solving Constraint Integer Programs                 */
/*       SCIP is distributed under the terms of the SCIP Academic Licence,   */
/*       see file COPYING in the SCIP distribution.                          */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
#pragma ident "@(#) $Id: handle_symmetries.cpp 1161 2015-10-25 16:57:57Z pfetsch $"
// #define PERMLIB_DEBUGMODE 1
// #define SCIP_OUTPUT
/**@file   handle_symmetries.cpp
 * @brief  Code needed to handle symmetries, mainly symmetry detection
 * @author Marc Pfetsch
 * @author Thomas Rehn
 */

/*---+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8----+----9----+----0----+----1----+----2*/

#include "handle_symmetries.h"
#include "type_probspec.h"

#include <cassert>
#include <unistd.h>
#include <sys/times.h>
#include <cstring>

#include <vector>
#include <list>
#include <map>

#include <permlib/permlib_api.h>
#include <permlib/construct/known_bsgs_construction.h>
#include <permlib/test/type_recognition_heuristic.h>

#include <bliss-0.73/graph.hh>

#include <scip/cons_linear.h>
#include <scip/cons_knapsack.h>
#include <scip/cons_varbound.h>
#include <scip/cons_setppc.h>
#include <scip/cons_logicor.h>
#include <scip/cons_xor.h>

using namespace permlib;

typedef SymmetricGroupRecognitionHeuristic<PERMUTATION> SymGroupHeur;


/*
 * Data structures
 */

/** define types for constraints */
enum SYMconstype
{
   SYMtypeInequality = 0,                    /**< linear inequality */
   SYMtypeEquation   = 1,                    /**< linear equation */
   SYMtypeXOR        = 2                     /**< XOR constraint */
};
typedef enum SYMconstype SYMconsType;


/** struct for bliss callback */
struct BLISS_Data
{
   SCIP*                 scip;               /**< SCIP pointer */
   unsigned int          nvars;              /**< number of variables */
   unsigned int          nperms;             /**< number of permutions */
   unsigned int**        perms;              /**< generators of symmetry group: permutations as array of arrays (nperms x nvars) */
   unsigned int          nmaxperms;          /**< maximal number of permutations */
   unsigned int          nlevels;            /**< number of "levels" in the graph" */
};


/** struct for bliss callback (used by heuristic) */
struct BLISS_DataSymHeuristic
{
   SCIP*                 scip;               /**< SCIP pointer */
   unsigned int          nvars;              /**< number of variables */
   unsigned int          nlevels;            /**< number of "levels" in the graph" */
   unsigned int          ignoredgens;        /**< number of generators that have been ignored by heuristic detection */
   SymGroupHeur*         sheur;              /**< pointer to symmetric group detection heuristic */
};


/** A collection of internal variables for MIP symmetry graph construction */
class GraphRawData
{
public:
   GraphRawData(SCIP* scip_)
      :  scip(scip_),
         nvars(SCIPgetNVars(scip_)),
         vars(SCIPgetVars(scip_)),
         nconss(SCIPgetNConss(scip_)),
         matcoef(0),
         rhscoef(0),
         rhstype(0),
         matrhsidx(0),
         matvaridx(0),
         matidx(0),
         rhsidx(0),
         nmatcoef(0),
         nrhscoef(0),
         nmaxmatcoef(100 * nvars),
         matcoeforig(0),
         rhscoeforig(0),
         T(1)
   {
   }

   ~GraphRawData()
   {
      SCIPfreeBlockMemoryArrayNull(scip, &rhsidx, 2 * nconss);
      SCIPfreeBlockMemoryArrayNull(scip, &rhstype, 2 * nconss);
      SCIPfreeBlockMemoryArrayNull(scip, &rhscoef, 2 * nconss);
      SCIPfreeBlockMemoryArrayNull(scip, &matvaridx, nmaxmatcoef);
      SCIPfreeBlockMemoryArrayNull(scip, &matrhsidx, nmaxmatcoef);
      SCIPfreeBlockMemoryArrayNull(scip, &matidx, nmaxmatcoef);
      SCIPfreeBlockMemoryArrayNull(scip, &matcoef, nmaxmatcoef);
   }

   SCIP_RETCODE initialize()
   {
      SCIP_CALL( SCIPallocBlockMemoryArray(scip, &matcoef, nmaxmatcoef) );
      SCIP_CALL( SCIPallocBlockMemoryArray(scip, &matidx, nmaxmatcoef) );
      SCIP_CALL( SCIPallocBlockMemoryArray(scip, &matrhsidx, nmaxmatcoef) );
      SCIP_CALL( SCIPallocBlockMemoryArray(scip, &matvaridx, nmaxmatcoef) );
      SCIP_CALL( SCIPallocBlockMemoryArray(scip, &rhscoef, 2 * nconss) );
      SCIP_CALL( SCIPallocBlockMemoryArray(scip, &rhstype, 2 * nconss) );
      SCIP_CALL( SCIPallocBlockMemoryArray(scip, &rhsidx, 2 * nconss) );
      return SCIP_OKAY;
   }

   /* data */
   SCIP*                 scip;               /**< scip pointer */
   int                   nvars;              /**< number of variables */
   SCIP_VAR**            vars;               /**< pointer to variables array by SCIP */
   int                   nconss;             /**< number of constraints */
   SCIP_Real*            matcoef;            /**< nonzero coefficients appearing in the matrix */
   SCIP_Real*            rhscoef;            /**< rhs coefficients */
   SYMconsType*          rhstype;            /**< type of constraint */
   unsigned int*         matrhsidx;          /**< indices of rhs corresponding to matrix entries */
   unsigned int*         matvaridx;          /**< indices of variables for matrix entries */
   int*                  matidx;             /**< indices in mat(rhs/var)idx array corresponding to matrix coefficients */
   int*                  rhsidx;             /**< indices in rhstype array corresponding to rhs coefficients */
   unsigned int          nmatcoef;           /**< number of different coefficients in matrix */
   unsigned int          nrhscoef;           /**< number of different coefficients in rhs */
   unsigned int          nmaxmatcoef;        /**< maximal number of matrix coefficients (will be increase on demand) */
   SCIP_Real*            matcoeforig;        /**< original coefficients in matrix */
   SCIP_Real*            rhscoeforig;        /**< original coefficients in rhs */
   unsigned int          T;                  /**< number of levels */
};


/** pair of rhs values and type of constraint */
typedef std::pair<SCIP_Real, SYMconsType>      RhsPair;



/**@class Varcmp
 * @brief Class to compare two variables according to objective, lower and upper bound, and variable type
 */
class Varcmp
{
public:
   Varcmp(
      SCIP*              s,                  /**< SCIP pointer */
      int                fixedtype,          /**< variable types that must be fixed by symmetries */
      SCIP_Bool          local               /**< use local variable bounds */
      )
   {
      scip_ = s;
      fixedtype_ = fixedtype;
      local_ = local;
   }

   bool operator() (SCIP_VAR* const &cvara, SCIP_VAR* const &cvarb) const
   {
      SCIP_VAR* vara = const_cast<SCIP_VAR*>(cvara);
      SCIP_VAR* varb = const_cast<SCIP_VAR*>(cvarb);

      /* determine special ordering w.r.t. fixed variables */
      bool symfixeda = SymmetryFixVar(fixedtype_, vara);
      bool symfixedb = SymmetryFixVar(fixedtype_, varb);

      /* if both variables are fixed, we sort only by their unique index */
      if ( symfixeda && symfixedb )
         return SCIPvarGetIndex(vara) < SCIPvarGetIndex(varb);
      else
      {
         /* if one variable is fixed, it is less than the other */
         if ( symfixeda )
            return true;
         if ( symfixedb )
            return false;
      }

      /* first check objective coefficients */
      if ( SCIPisLT(scip_, SCIPvarGetObj(vara), SCIPvarGetObj(varb)) )
         return true;
      if ( SCIPisLT(scip_, SCIPvarGetObj(varb), SCIPvarGetObj(vara)) )
         return false;

      SCIP_Real lba = SCIPvarGetLbGlobal(vara);
      SCIP_Real lbb = SCIPvarGetLbGlobal(varb);

      SCIP_Real uba = SCIPvarGetUbGlobal(vara);
      SCIP_Real ubb = SCIPvarGetUbGlobal(varb);

      if ( local_ )
      {
         if ( SCIPvarIsBinary(vara) && lba + 0.5 < uba )
         {
            int nbdchg = SCIPvarGetNBdchgInfosLb(vara);
            for (int i = 0; i < nbdchg; ++i)
            {
               SCIP_BDCHGINFO* info = SCIPvarGetBdchgInfoLb(vara, i);
               assert( SCIPbdchginfoGetBoundtype(info) == SCIP_BOUNDTYPE_LOWER );
               if ( SCIPbdchginfoGetChgtype(info) == SCIP_BOUNDCHGTYPE_BRANCHING )
               {
                  assert( SCIPbdchginfoGetNewbound(info) > lba );
                  lba = SCIPbdchginfoGetNewbound(info);
               }
            }
         }

         if ( SCIPvarIsBinary(varb) && lbb + 0.5 < ubb )
         {
            int nbdchg = SCIPvarGetNBdchgInfosLb(varb);
            for (int i = 0; i < nbdchg; ++i)
            {
               SCIP_BDCHGINFO* info = SCIPvarGetBdchgInfoLb(varb, i);
               assert( SCIPbdchginfoGetBoundtype(info) == SCIP_BOUNDTYPE_LOWER );
               if ( SCIPbdchginfoGetChgtype(info) == SCIP_BOUNDCHGTYPE_BRANCHING )
               {
                  assert( SCIPbdchginfoGetNewbound(info) > lbb );
                  lbb = SCIPbdchginfoGetNewbound(info);
               }
            }
         }
      }

      /* if still undecided, take lower bound */
      if ( SCIPisLT(scip_, lba, lbb) )
         return true;
      if ( SCIPisLT(scip_, lbb, lba) )
         return false;

      if ( local_ )
      {
         if ( SCIPvarIsBinary(vara) && lba + 0.5 < uba )
         {
            int nbdchg = SCIPvarGetNBdchgInfosUb(vara);
            for (int i = 0; i < nbdchg; ++i)
            {
               SCIP_BDCHGINFO* info = SCIPvarGetBdchgInfoUb(vara, i);
               assert( SCIPbdchginfoGetBoundtype(info) == SCIP_BOUNDTYPE_UPPER );
               if ( SCIPbdchginfoGetChgtype(info) == SCIP_BOUNDCHGTYPE_BRANCHING )
               {
                  assert( SCIPbdchginfoGetNewbound(info) < uba );
                  uba = SCIPbdchginfoGetNewbound(info);
               }
            }
         }

         if ( SCIPvarIsBinary(varb) && lbb + 0.5 < ubb )
         {
            int nbdchg = SCIPvarGetNBdchgInfosUb(varb);
            for (int i = 0; i < nbdchg; ++i)
            {
               SCIP_BDCHGINFO* info = SCIPvarGetBdchgInfoUb(varb, i);
               assert( SCIPbdchginfoGetBoundtype(info) == SCIP_BOUNDTYPE_UPPER );
               if ( SCIPbdchginfoGetChgtype(info) == SCIP_BOUNDCHGTYPE_BRANCHING )
               {
                  assert( SCIPbdchginfoGetNewbound(info) < ubb );
                  ubb = SCIPbdchginfoGetNewbound(info);
               }
            }
         }
      }

      /* if still undecided, take upper bound */
      if ( SCIPisLT(scip_, uba, ubb) )
         return true;
      if ( SCIPisLT(scip_, ubb, uba) )
         return false;

      /* if still undecided, take variable type */
      if ( SCIPvarGetType(vara) < SCIPvarGetType(varb) )
         return true;
      /* test for "SCIPvarGetType(varb) < SCIPvarGetType(vara)" not needed (would return false anyway) */

      return false;
   }

private:
   SCIP*                 scip_;              /**< SCIP pointer */
   int                   fixedtype_;         /**< variable types that must be fixed by symmetries */
   SCIP_Bool             local_;             /**< use local variable bounds */
};


/**@class Realcmp
 * @brief Class to compare real values within a feasibility tolerance
 */
class Realcmp
{
public:
   Realcmp(
      SCIP*              s                   /**< SCIP pointer */
      )
   {
      scip_ = s;
   }

   bool operator() (const SCIP_Real& a, const SCIP_Real& b) const
   {
      if ( SCIPisLT(scip_, a, b) )
         return true;
      return false;
   }

private:
   SCIP*                 scip_;              /**< SCIP pointer */
};


/**@class Realuintcmp
 * @brief Class to compare pairs of real values and unsigned int values within a feasibility tolerance
 *
 * Note that the second entry is usually a SCIP_Bool.
 */
class Rhspaircmp
{
public:
   Rhspaircmp(
      SCIP*              s                   /**< SCIP pointer */
      )
   {
      scip_ = s;
   }

   bool operator() (const RhsPair& a, const RhsPair& b) const
   {
      /* first compare second component (faster) */
      if ( a.second < b.second )
         return true;
      if ( a.second > b.second )
         return false;
      assert( a.second == b.second );
      if ( SCIPisLT(scip_, a.first, b.first) )
         return true;
      /* test for SCIPisLT(scip_, b.first, a.first) not needed (would return false anyway) */
      return false;
   }

private:
   SCIP*                 scip_;              /**< SCIP pointer */
};


/* useful typedefs */
typedef std::map<SCIP_VAR*, unsigned int, Varcmp>      VarMapType;
typedef std::map<SCIP_Real, unsigned int, Realcmp>     RealMapType;
typedef std::map<RhsPair, unsigned int, Rhspaircmp>    RhsPairMapType;





/*
 * Local methods
 */


/** callback function for bliss */
static
void blisshook(
   void*                 user_param,         /**< parameter supplied at call to bliss */
   unsigned int          n,                  /**< size of aut vector */
   const unsigned int*   aut                 /**< automorphism */
   )
{
   assert( aut != 0 );
   assert( user_param != 0 );

   BLISS_Data* data = (BLISS_Data*) user_param;
   assert( data->scip != 0 );
   assert( data->perms != 0 );
   assert( data->nvars < n );

#if 0
   SCIPdebugMessage("Bliss callback function.\n");
#endif

   /* check whether we need to resize */
   if ( data->nperms >= data->nmaxperms )
   {
      int newsize = SCIPcalcMemGrowSize(data->scip, data->nperms);
      SCIP_RETCODE retcode = SCIPreallocBlockMemoryArray(data->scip, &data->perms, data->nmaxperms, newsize);
      if ( retcode != SCIP_OKAY )
         return;
   }

   /* copy first part of automorphism */
   bool isIdentity = true;
   unsigned int* p = 0;
   SCIP_RETCODE retcode = SCIPallocBlockMemoryArray(data->scip, &p, data->nvars);
   if ( retcode != SCIP_OKAY )
      return;
   for (unsigned int j = 0; j < data->nvars; ++j)
   {
      assert( j * data->nlevels < n );

      /* convert index of variable-level 0-nodes to variable indices */
      p[j] = aut[j * data->nlevels] / data->nlevels;
      if ( p[j] != j )
         isIdentity = false;
   }

   /*  ignore trivial generators, i.e. generators that only permute the constraints */
   if ( isIdentity )
   {
#if 0
      SCIPdebugMessage("ignoring trivial permutation\n");
#endif
      SCIPfreeBlockMemoryArray(data->scip, &p, data->nvars);
   }
   else
      data->perms[data->nperms++] = p;
}


/** callback function for bliss (used by heuristic) */
static
void blisshookheur(
   void*                 user_param,         /**< parameter supplied at call to bliss */
   unsigned int          n,                  /**< size of aut vector */
   const unsigned int*   aut                 /**< automorphism */
   )
{
   assert( aut != 0 );
   assert( user_param != 0 );

   BLISS_DataSymHeuristic* data = (BLISS_DataSymHeuristic*) user_param;
   assert( data->scip != 0 );
   assert( data->sheur != 0 );

   /* copy first part of automorphism */
   bool isIdentity = true;
   unsigned int* p = 0;
   SCIP_RETCODE retcode = SCIPallocBlockMemoryArray(data->scip, &p, data->nvars);
   if ( retcode != SCIP_OKAY )
      return;
   for (unsigned int j = 0; j < data->nvars; ++j)
   {
      assert( j * data->nlevels < n );

      /* convert index of variable-level 0-nodes to variable indices */
      p[j] = aut[j * data->nlevels] / data->nlevels;
      if ( p[j] != j )
         isIdentity = false;
   }

   /*  ignore trivial generators, i.e. generators that only permute the constraints */
   if ( ! isIdentity )
   {
      PERMUTATION perm(p, p+data->nvars);
      if ( ! data->sheur->addGenerator(perm) )
         ++data->ignoredgens;
   }
   SCIPfreeBlockMemoryArray(data->scip, &p, data->nvars);
}


/** transforms given variables, scalars, and constant to the corresponding active variables, scalars, and constant
 *
 *  @p constant needs to be initialized!
 */
static
SCIP_RETCODE getActiveVariables(
   SCIP*                 scip,               /**< SCIP data structure */
   SCIP_VAR***           vars,               /**< pointer to vars array to get active variables for */
   SCIP_Real**           scalars,            /**< pointer to scalars a_1, ..., a_n in linear sum a_1*x_1 + ... + a_n*x_n + c */
   int*                  nvars,              /**< pointer to number of variables and values in vars and vals array */
   SCIP_Real*            constant,           /**< pointer to constant c in linear sum a_1*x_1 + ... + a_n*x_n + c */
   SCIP_Bool             transformed         /**< transformed constraint? */
   )
{
   int requiredsize;
   int v;

   assert( scip != 0 );
   assert( vars != 0 );
   assert( scalars != 0 );
   assert( *vars != 0 );
   assert( *scalars != 0 );
   assert( nvars != 0 );
   assert( constant != 0 );

   if ( transformed )
   {
      SCIP_CALL( SCIPgetProbvarLinearSum(scip, *vars, *scalars, nvars, *nvars, constant, &requiredsize, TRUE) );

      if ( requiredsize > *nvars )
      {
         SCIP_CALL( SCIPreallocBufferArray(scip, vars, requiredsize) );
         SCIP_CALL( SCIPreallocBufferArray(scip, scalars, requiredsize) );

         SCIP_CALL( SCIPgetProbvarLinearSum(scip, *vars, *scalars, nvars, requiredsize, constant, &requiredsize, TRUE) );
         assert( requiredsize <= *nvars );
      }
   }
   else
   {
      for (v = 0; v < *nvars; ++v)
      {
         SCIP_CALL( SCIPvarGetOrigvarSum(&(*vars)[v], &(*scalars)[v], constant) );
      }
   }
   return SCIP_OKAY;
}


/** Fill in matrix elements into coefficient arrays. */
static
SCIP_RETCODE collectCoefficients(
   SCIP*                 scip,               /**< SCIP data structure */
   SCIP_VAR**            linvars,            /**< array of linear variables */
   SCIP_Real*            linvals,            /**< array of linear coefficients values (or 0 if all linear coefficient values are 1) */
   int                   nlinvars,           /**< number of linear variables */
   SCIP_Real             lhs,                /**< left hand side */
   SCIP_Real             rhs,                /**< right hand side */
   SCIP_Bool             istransformed,      /**< whether the constraint is transformed */
   SCIP_Bool             isxor,              /**< whether the constraint is an XOR constraint */
   SCIP_Real*&           matcoef,            /**< matrix coefficients */
   int*&                 matidx,             /**< indices in mat(rhs/var)idx array corresponding to matrix coefficients */
   unsigned int*&        matrhsidx,          /**< indices of rhs corresponding to matrix entries */
   unsigned int*&        matvaridx,          /**< indices of variables corresponding to matrix entries */
   unsigned int&         nmatcoef,           /**< number of matrix coefficients */
   unsigned int&         nmaxmatcoef,        /**< maximal number of matrix coefficients */
   SCIP_Real*            rhscoef,            /**< rhs coefficients */
   SYMconsType*          rhstype,            /**< type of constraints */
   int*                  rhsidx,             /**< indices in rhstype array corresponding to rhs coefficients */
   unsigned int&         nrhscoef            /**< number of rhs coefficients */
   )
{
   assert( scip != 0 );

   assert( nlinvars == 0 || linvars != 0 );
   assert( lhs <= rhs );

   /* do nothing if constraint is empty */
   if ( nlinvars == 0 )
      return SCIP_OKAY;

   /* Ignore redundant constraints! */
   if ( SCIPisInfinity(scip, -lhs) && SCIPisInfinity(scip, rhs) )
      return SCIP_OKAY;

   /* get active variables */
   SCIP_VAR** vars;
   SCIP_Real* vals;
   SCIP_Real constant = 0.0;
   int nvars;

   /* duplicate variable and value array */
   nvars = nlinvars;
   SCIP_CALL( SCIPduplicateBufferArray(scip, &vars, linvars, nvars) );
   if ( linvals != 0 )
   {
      SCIP_CALL( SCIPduplicateBufferArray(scip, &vals, linvals, nvars) );
   }
   else
   {
      SCIP_CALL( SCIPallocBufferArray(scip, &vals, nvars) );
      for (int v = 0; v < nvars; ++v)
         vals[v] = 1.0;
   }
   assert( vals != 0 );
   assert( vars != 0 );

   /* get active variables */
   SCIP_CALL( getActiveVariables(scip, &vars, &vals, &nvars, &constant, istransformed) );

   /* check whether constraint is non-empty after transformation to active variables */
   if ( nvars <= 0 )
   {
      SCIPfreeBufferArray(scip, &vals);
      SCIPfreeBufferArray(scip, &vars);
      return SCIP_OKAY;
   }

   /* handle constant */
   if ( ! SCIPisInfinity(scip, -lhs) )
      lhs -= constant;
   if ( ! SCIPisInfinity(scip, rhs) )
      rhs -= constant;

   /* check whether we have to resize */
   if ( nmatcoef + nvars > nmaxmatcoef )
   {
      int newsize = SCIPcalcMemGrowSize(scip, nmatcoef + nvars);
      assert( newsize >= 0 );
      SCIP_CALL( SCIPreallocBlockMemoryArray(scip, &matidx, nmaxmatcoef, newsize) );
      SCIP_CALL( SCIPreallocBlockMemoryArray(scip, &matrhsidx, nmaxmatcoef, newsize) );
      SCIP_CALL( SCIPreallocBlockMemoryArray(scip, &matvaridx, nmaxmatcoef, newsize) );
      SCIP_CALL( SCIPreallocBlockMemoryArray(scip, &matcoef, nmaxmatcoef, newsize) );
      SCIPdebugMessage("Resized matrix coefficients from %u to %d.\n", nmaxmatcoef, newsize);
      nmaxmatcoef = (unsigned int) newsize;
   }

   /* check lhs/rhs */
   if ( SCIPisEQ(scip, lhs, rhs) )
   {
      assert( !SCIPisInfinity(scip, rhs) );

      /* equality constraint */
      rhscoef[nrhscoef] = rhs;
      if ( isxor )
         rhstype[nrhscoef] = SYMtypeXOR;
      else
         rhstype[nrhscoef] = SYMtypeEquation;
      rhsidx[nrhscoef] = (int) nrhscoef;

      for (int j = 0; j < nvars; ++j)
      {
         assert( nmatcoef < nmaxmatcoef );
         matidx[nmatcoef] = (int) nmatcoef;
         matrhsidx[nmatcoef] = nrhscoef;
         assert( 0 <= SCIPvarGetProbindex(vars[j]) && SCIPvarGetProbindex(vars[j]) < SCIPgetNVars(scip) );
         matvaridx[nmatcoef] = (unsigned) SCIPvarGetProbindex(vars[j]);
         matcoef[nmatcoef] = vals[j];
         ++nmatcoef;
      }
      ++nrhscoef;
   }
   else
   {
      if ( !SCIPisInfinity(scip, -lhs) )
      {
         rhscoef[nrhscoef] = -lhs;
         rhstype[nrhscoef] = SYMtypeInequality;
         rhsidx[nrhscoef] = (int) nrhscoef;

         for (int j = 0; j < nvars; ++j)
         {
            assert( nmatcoef < nmaxmatcoef );
            matidx[nmatcoef] = (int) nmatcoef;
            matrhsidx[nmatcoef] = nrhscoef;
            matvaridx[nmatcoef] = (unsigned) SCIPvarGetProbindex(vars[j]);
            assert( 0 <= SCIPvarGetProbindex(vars[j]) && SCIPvarGetProbindex(vars[j]) < SCIPgetNVars(scip) );
            matcoef[nmatcoef] = -vals[j];
            ++nmatcoef;
         }
         ++nrhscoef;
      }
      if ( !SCIPisInfinity(scip, rhs) )
      {
         rhscoef[nrhscoef] = rhs;
         rhstype[nrhscoef] = SYMtypeInequality;
         rhsidx[nrhscoef] = (int) nrhscoef;

         for (int j = 0; j < nvars; ++j)
         {
            assert( nmatcoef < nmaxmatcoef );
            matidx[nmatcoef] = (int) nmatcoef;
            matrhsidx[nmatcoef] = nrhscoef;
            assert( 0 <= SCIPvarGetProbindex(vars[j]) && SCIPvarGetProbindex(vars[j]) < SCIPgetNVars(scip) );
            matvaridx[nmatcoef] = (unsigned) SCIPvarGetProbindex(vars[j]);
            matcoef[nmatcoef] = vals[j];
            ++nmatcoef;
         }
         ++nrhscoef;
      }
   }

   SCIPfreeBufferArray(scip, &vals);
   SCIPfreeBufferArray(scip, &vars);

   return SCIP_OKAY;
}


#ifndef NDEBUG
/** checks whether given permutations are a symmetry of an IP
 *
 *  We need the matrix and rhs in the original order. This is needed to speed up the comparison
 *  process. The matrix is needed in the right order to easily check rows. The rhs is used because
 *  of cache effects.
 */
static
SCIP_RETCODE checkSymmetriesAreSymmetries(
   SCIP*                 scip,               /**< SCIP data structure */
   SCIP_VAR**            vars,               /**< variables */
   int                   nvars,              /**< number of linear variables */
   const SCIP_Real*      matcoef,            /**< matrix coefficients in original order! */
   const unsigned int*   matrhsidx,          /**< indices of rhs corresponding to matrix entries */
   const unsigned int*   matvaridx,          /**< indices of variables corresponding to matrix entries */
   unsigned int          nmatcoef,           /**< number of matrix coefficients */
   const SCIP_Real*      rhscoef,            /**< rhs coefficients in original order! */
   const SYMconsType*    rhstype,            /**< type of constraints */
   unsigned int          nrhscoef,           /**< number of rhs coefficients */
   unsigned int          nperms,             /**< number of permutations */
   unsigned int**        perms,              /**< permutations */
   int                   fixedtype           /**< variable types that must be fixed by symmetries */
   )
{
   SCIPdebugMessage("Checking whether symmetries are symmetries (generators: %u).\n", nperms);

   /* set up dense arrow for permuted row */
   SCIP_Real* permrow = 0;
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &permrow, nvars) );

   /* set up map between rows and first entry in matcoef array */
   unsigned int* rhsmatbeg = 0;
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &rhsmatbeg, nrhscoef) );
   for (unsigned int i = 0; i < nrhscoef; ++i)
      rhsmatbeg[i] = nmatcoef + 1;

   /* build map from rhs into matrix */
   unsigned int oldrhs = nrhscoef + 1;
   for (unsigned int j = 0; j < nmatcoef; ++j)
   {
      unsigned int rhs = matrhsidx[j];
      if ( rhs != oldrhs )
      {
         assert( rhs < nrhscoef );
         rhsmatbeg[rhs] = j;
         oldrhs = rhs;
      }
   }

   /* create row */
   for (int j = 0; j < nvars; ++j)
      permrow[j] = 0.0;

   /* check all gnerators */
   for (unsigned int p = 0; p < nperms; ++p)
   {
      SCIPdebugMessage("verifying automorphism group generator #%u ...\n", p);
      unsigned int* P = perms[p];
      assert( P != 0 );

      for (int j = 0; j < nvars; ++j)
      {
         if ( SymmetryFixVar(fixedtype, vars[j]) && P[j] != static_cast<unsigned int>(j) )
         {
            SCIPdebugMessage("permutation does not fix types %d, moving variable %d\n", fixedtype, j);
            return SCIP_ERROR;
         }
      }

      /* check all constraints == rhs */
      for (unsigned int r1 = 0; r1 < nrhscoef; ++r1)
      {
         /* fill row into permrow (dense) */
         unsigned int j = rhsmatbeg[r1];
         assert( j < nmatcoef );
         assert( matrhsidx[j] == r1 ); /* note: row cannot be empty by construction */

         /* loop through row */
         unsigned int npermuted = 0;
         while ( j < nmatcoef && matrhsidx[j] == r1 )
         {
            assert( matvaridx[j] < (unsigned int) nvars );
            unsigned int varidx = P[matvaridx[j]];
            assert( varidx < (unsigned int) nvars );
            if ( varidx != matvaridx[j] )
               ++npermuted;
            assert( SCIPisZero(scip, permrow[varidx]) );
            permrow[varidx] = matcoef[j];
            ++j;
         }

         /* if row is not affected by permutation, we do not have to check it */
         if ( npermuted > 0 )
         {
            /* check other rows (sparse) */
            bool found = false;
            for (unsigned int r2 = 0; r2 < nrhscoef; ++r2)
            {
               /* a permutation must map constraints of the same type and respect rhs coefficients */
               if ( rhstype[r1] == rhstype[r2] && SCIPisEQ(scip, rhscoef[r1], rhscoef[r2]) )
               {
                  j = rhsmatbeg[r2];
                  assert( j < nmatcoef );
                  assert( matrhsidx[j] == r2 );
                  assert( matvaridx[j] < (unsigned int) nvars );

                  /* loop through row r2 and check whether it is equal to permuted row r */
                  while (j < nmatcoef && matrhsidx[j] == r2 && SCIPisEQ(scip, permrow[matvaridx[j]], matcoef[j]) )
                     ++j;

                  /* check whether rows are completely equal */
                  if ( j >= nmatcoef || matrhsidx[j] != r2 )
                  {
                     /* perm[p] is indeed a symmetry */
                     found = true;
                     break;
                  }
               }
            }

            assert( found );
            if ( ! found ) /*lint !e774*/
            {
#ifdef SCIP_DEBUG
               print_iterable(perms[p], perms[p] + nvars, 0, "perm no symmetry: ");
#endif
               return SCIP_ERROR;
            }
         }

         /* reset permrow */
         j = rhsmatbeg[r1];
         while ( j < nmatcoef && matrhsidx[j] == r1 )
         {
            unsigned int varidx = P[matvaridx[j]];
            permrow[varidx] = 0.0;
            ++j;
         }
      }
   }

   SCIPfreeBlockMemoryArray(scip, &rhsmatbeg, nrhscoef);
   SCIPfreeBlockMemoryArray(scip, &permrow, nvars);

   return SCIP_OKAY;
}
#endif


/** Construct colored graph for symmetry computations
 *
 *  Construct bipartite graph:
 *  - Each variable gets a different node.
 *  - Each constraint gets a different node.
 *  - Each matrix coefficient gets a different node that is conntected to the two nodes
 *    corresponding to the constraint and variable.
 *
 *  Each different variable, rhs, and matrix coefficient type gets a different color that is
 *  attached to the corresponding entries.
 */
static
SCIP_RETCODE fillGraphByColoredCoefficients(
   bliss::Graph&         G,                  /**< Graph to be constructed */
   SCIP*                 scip,               /**< SCIP instance */
   SCIP_VAR**            vars,               /**< SCIP variables */
   int                   nvars,              /**< number of variables */
   const VarMapType&     varmap,             /**< map from variables to indices */
   const RealMapType&    matcoefmap,         /**< map from matrix coefficients to indices */
   const RhsPairMapType& rhscoefmap,         /**< map from rhs coefficients to indices */
   unsigned int          nuniquematcoef,     /**< number of unique matrix coefficients */
   unsigned int          nuniquerhscoef,     /**< number of unique rhs coefficients */
   unsigned int          nuniquevars,        /**< number of unique variables */
   const SCIP_Real*      matcoef,            /**< matrix coefficients */
   const int*            matidx,             /**< indices in mat(rhs/var)idx array corresponding to matrix coefficients */
   const unsigned int*   matrhsidx,          /**< indices of rhs corresponding to matrix entries */
   const unsigned int*   matvaridx,          /**< indices of variables corresponding to matrix entries */
   unsigned int          nmatcoef,           /**< number of matrix coefficients */
   const SCIP_Real*      rhscoef,            /**< rhs coefficients */
   const SYMconsType*    rhstype,            /**< type of constraints */
   const int*            rhsidx,             /**< indices in rhstype array corresponding to rhs coefficients */
   unsigned int          nrhscoef,           /**< number of rhs coefficients */
   unsigned int&         nnodes,             /**< number of nodes in graph */
   unsigned int&         nedges,             /**< number of edges in graph */
   SCIP_Bool&            success             /**< whether the construction was successful */
   )
{
   SCIPdebugMessage("Building graph with colored coefficient nodes.\n");
   nnodes = 0;
   nedges = 0;
   success = FALSE;

   /* add nodes for variables */
   for (unsigned int v = 0; v < (unsigned int) nvars; ++v)
   {
      SCIP_VAR* var = vars[v];
      assert( var != 0 );

      /* find color corresponding to variable */
      assert( varmap.find(var) != varmap.end() );
      const unsigned int color = varmap.find(var)->second;
      assert( color < nuniquevars );

      (void) G.add_vertex(color);
      ++nnodes;
   }
   assert( (int) G.get_nof_vertices() == nvars );

   /* store nodes corresponding to rhs (we need the original order of rows) */
   unsigned int* rhsnodemap;
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &rhsnodemap, nrhscoef) );
#ifndef NDEBUG
   for (unsigned int j = 0; j < nrhscoef; ++j)
      rhsnodemap[j] = nrhscoef + 1;
#endif

   /* add nodes for rhs of constraints */
   for (unsigned int c = 0; c < nrhscoef; ++c)
   {
      int idx = rhsidx[c];
      assert( 0 <= idx && idx < (int) nrhscoef );

      /* find color corresponding to rhs coefficient distinguishing types of constraints */
      const RhsPair valPair = std::make_pair(rhscoef[c], rhstype[idx]);
      assert( rhscoefmap.find(valPair) != rhscoefmap.end() );
      const unsigned int color = rhscoefmap.find(valPair)->second;
      assert( color < nuniquerhscoef );

      unsigned int node = G.add_vertex(nuniquevars + color);
      assert( node == (unsigned int) nvars + c );
      rhsnodemap[idx] = node;
      ++nnodes;
   }
   assert( G.get_nof_vertices() == (unsigned int) nvars + nrhscoef );

   typedef std::pair<unsigned int, unsigned int> InterPair;
   typedef std::map<InterPair, unsigned int> IntermediatesMap;
   IntermediatesMap groupedIntermediateNodes;
   /* Grouping of nodes depends on the number of nodes in the bipartite graph class.
    * If there are more variables than constraints, we group by constraints.
    * That is, given several variable nodes which are incident to one constraint node by the same color,
    * we join these variable nodes to the constaint node by only one intermediate node.
    */
   const bool groupByConstraints = nrhscoef < static_cast<unsigned int>(nvars);
   if (groupByConstraints)
      SCIPdebugMessage("group intermediate nodes by constraints\n");
   else
      SCIPdebugMessage("group intermediate nodes by variables\n");

   /* "colored" edges based on all matrix coefficients */
   unsigned int nusedcolors = nuniquevars + nuniquerhscoef;
   for (unsigned int j = 0; j < nmatcoef; ++j)
   {
      int idx = matidx[j];
      assert( 0 <= idx && idx < (int) nmatcoef );

      /* find color corresponding to matrix coefficient */
      SCIP_Real val = matcoef[j];
      assert( matcoefmap.find(val) != matcoefmap.end() );
      const unsigned int color = matcoefmap.find(val)->second;
      assert( color < nuniquematcoef );

      assert( matrhsidx[idx] < nrhscoef );
      assert( matvaridx[idx] < (unsigned int) nvars );
      const unsigned int rhsnode = rhsnodemap[matrhsidx[idx]];
      const unsigned int varnode = matvaridx[idx];
      assert( rhsnode < G.get_nof_vertices() );
      assert( varnode < G.get_nof_vertices() );
      assert( (unsigned int) nvars <= rhsnode && rhsnode < (unsigned int) nvars + nrhscoef );

      /* if we have only one colors, we do not need intermediate nodes */
      if (nuniquematcoef == 1)
      {
         G.add_edge(varnode, rhsnode);
         ++nedges;
      }
      else
      {
         InterPair key;
         if (groupByConstraints)
            key = std::make_pair(matrhsidx[idx], color);
         else
            key = std::make_pair(matvaridx[idx], color);
         IntermediatesMap::const_iterator keyIt = groupedIntermediateNodes.find(key);
         unsigned int intermediatenode = 0;
         if (keyIt == groupedIntermediateNodes.end())
         {
            intermediatenode = G.add_vertex(nusedcolors + color);
            groupedIntermediateNodes[ key ] = intermediatenode;
            ++nnodes;
         }
         else
         {
            intermediatenode = (*keyIt).second;
         }
         assert( intermediatenode >= static_cast<unsigned int>(nvars + nrhscoef) );
         /* determine whether graph would be too large for bliss (can only handle unsigned int) */
         if ( intermediatenode >= INT_MAX )
         {
            SCIPfreeBlockMemoryArray(scip, &rhsnodemap, nrhscoef);
            return SCIP_OKAY;
         }
         G.add_edge(varnode, intermediatenode);
         G.add_edge(rhsnode, intermediatenode);
         nedges += 2;
      }
   }

   SCIPfreeBlockMemoryArray(scip, &rhsnodemap, nrhscoef);

   success = TRUE;
   return SCIP_OKAY;
}


/** Construct colored graph for symmetry computations
 *
 *  The colored graph is constructed as explained in the nauty user guide. It adds a single node for
 *  each different matrix coefficient and colors it using logarithmically many levels @a T.
 */
static
SCIP_RETCODE fillGraphByColorLayers(
   bliss::Graph&         G,                  /**< Graph to be constructed */
   SCIP*                 scip,               /**< SCIP instance */
   unsigned int          T,                  /**< number of levels */
   SCIP_VAR**            vars,               /**< SCIP variables */
   int                   nvars,              /**< number of variables */
   const VarMapType&     varmap,             /**< map from variables to indices */
   const RealMapType&    matcoefmap,         /**< map from matrix coefficients to indices */
   const RhsPairMapType& rhscoefmap,         /**< map from rhs coefficients to indices */
   unsigned int          nuniquematcoef,     /**< number of unique matrix coefficients */
   unsigned int          nuniquerhscoef,     /**< number of unique rhs coefficients */
   unsigned int          nuniquevars,        /**< number of unique variables */
   const SCIP_Real*      matcoef,            /**< matrix coefficients */
   const int*            matidx,             /**< indices in mat(rhs/var)idx array corresponding to matrix coefficients */
   const unsigned int*   matrhsidx,          /**< indices of rhs corresponding to matrix entries */
   const unsigned int*   matvaridx,          /**< indices of variables corresponding to matrix entries */
   unsigned int          nmatcoef,           /**< number of matrix coefficients */
   const SCIP_Real*      rhscoef,            /**< rhs coefficients */
   const SYMconsType*    rhstype,            /**< type of constraints */
   const int*            rhsidx,             /**< indices in rhstype array corresponding to rhs coefficients */
   unsigned int          nrhscoef,           /**< number of rhs coefficients */
   unsigned int&         nnodes,             /**< number of nodes in graph */
   unsigned int&         nedges,             /**< number of edges in graph */
   SCIP_Bool&            success             /**< whether the construction was successful */
   )
{
   SCIPdebugMessage("Building graph with %u layers.\n", T);
   nnodes = 0;
   nedges = 0;
   success = FALSE;

   /* determine whether graph would be too large for bliss (can only handle unsigned int) */
   unsigned long int g = T * (nrhscoef + (unsigned int) nvars);
   if ( g > (unsigned int) INT_MAX )
      return SCIP_OKAY;

   /* add nodes for variables */
   for (unsigned int v = 0; v < (unsigned int) nvars; ++v)
   {
      /* find color corresponding to variable */
      assert( varmap.find(vars[v]) != varmap.end() );
      const unsigned int color = varmap.find(vars[v])->second;
      assert( color < nuniquevars );

      /* add nodes in each layer and edges to previous layer */
      for (unsigned int t = 0; t < T; ++t)
      {
         unsigned int node = G.add_vertex(T * color + t);
         assert( node == v * T + t );
         ++nnodes;

         /* connect all vars across "color" layers */
         if ( t > 0 )
         {
            G.add_edge(node, node - 1);
            ++nedges;
         }
      }
   }
   assert( G.get_nof_vertices() == (unsigned int) nvars * T );

   /* store nodes corresponding to rhs (we need the original order of rows) */
   unsigned int* rhsnodemap = 0;
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &rhsnodemap, nrhscoef) );
#ifndef NDEBUG
   for (unsigned int j = 0; j < nrhscoef; ++j)
      rhsnodemap[j] = nrhscoef + 1;
#endif

   /* add nodes and edges for rhs of constraints */
   for (unsigned int c = 0; c < nrhscoef; ++c)
   {
      int idx = rhsidx[c];
      assert( 0 <= idx && idx < (int) nrhscoef );

      /* find color corresponding to rhs coefficient distinguishing types of constraints */
      const RhsPair valPair = std::make_pair(rhscoef[c], rhstype[idx]);
      assert( rhscoefmap.find(valPair) != rhscoefmap.end() );
      const unsigned int color = rhscoefmap.find(valPair)->second;
      assert( color < nuniquerhscoef );

      /* add nodes in each layer and add edges to previous layer */
      for (unsigned int t = 0; t < T; ++t)
      {
         const unsigned int node = G.add_vertex(T * (nuniquevars + color) + t);
         assert( node == T * (c + (unsigned int) nvars) + t );
         ++nnodes;

         /* connect all rhs across "color" layers */
         if ( t == 0 )
            rhsnodemap[idx] = node;
         else
         {
            G.add_edge(node, node - 1);
            ++nedges;
         }
      }
   }
   assert( G.get_nof_vertices() == T * (nrhscoef + (unsigned int) nvars) );

   /* add edges for each different coefficient in the matrix */
   for (unsigned int j = 0; j < nmatcoef; ++j)
   {
      int idx = matidx[j];
      assert( 0 <= idx && idx < (int) nmatcoef );

      /* find color corresponding to matrix coefficient */
      SCIP_Real val = matcoef[j];
      assert( matcoefmap.find(val) != matcoefmap.end() );
      const unsigned int color = matcoefmap.find(val)->second + 1;  /* add 1 because of bit testing below */
      assert( color <= nuniquematcoef );

      unsigned int idxrhs = rhsnodemap[matrhsidx[idx]];
      unsigned int idxvar = matvaridx[idx];
      assert( (unsigned int) nvars * T <= idxrhs && idxrhs < T * (nrhscoef + (unsigned int) nvars) );
      assert( idxvar < (unsigned int) nvars );

      for (unsigned int t = 0; t < T; ++t)
      {
         /* check bits */
         if ( (color & (1 << t)) )
         {
            unsigned int source = idxvar * T + t;
            unsigned int target = idxrhs + t;
            G.add_edge(source, target);
            ++nedges;
         }
      }
   }

   SCIPfreeBlockMemoryArray(scip, &rhsnodemap, nrhscoef);

   success = TRUE;
   return SCIP_OKAY;
}


/** Constructs a graph for a given MIP that encodes its formulation symmetry */
static
SCIP_RETCODE constructGraph(
   SCIP*                 scip,               /**< SCIP pointer */
   SCIP_Bool             uselayeredgraph,    /**< use layered graph for symmetry detection */
   int                   fixedtype,          /**< variable types that must be fixed by symmetries */
   SCIP_Bool             local,              /**< Use local variable bounds? */
   bliss::Graph*&        G,                  /**< pointer to a bliss graph */
   GraphRawData&         rawData,            /**< a bunch of internal variables that may be useful for verification */
   unsigned int&         nnodes,             /**< number of nodes in graph */
   unsigned int&         nedges              /**< number of edges in graph */
   )
{
   assert( scip != 0 );

   nnodes = 0;
   nedges = 0;

   /* get all global constraints */
   int nconss = rawData.nconss;
   SCIP_CONS** conss = SCIPgetConss(scip);

   /* exit if no constraints are available */
   if ( nconss == 0 )
      return SCIP_OKAY;

   /* get variables */
   int nvars = SCIPgetNVars(scip);
   SCIP_VAR** vars = SCIPgetVars(scip);

   SCIPdebugMessage("Detecting %ssymmetry on %d variables and %d constraints.\n", local ? "local " : "", nvars, nconss);

   unsigned int& nmaxmatcoef = rawData.nmaxmatcoef;
   SCIP_Real*& matcoef = rawData.matcoef;
   SCIP_Real*& rhscoef = rawData.rhscoef;
   SYMconsType*& rhstype = rawData.rhstype;
   unsigned int*& matrhsidx = rawData.matrhsidx;
   unsigned int*& matvaridx = rawData.matvaridx;
   int*& matidx = rawData.matidx;
   int*& rhsidx = rawData.rhsidx;
   unsigned int& nmatcoef = rawData.nmatcoef;
   unsigned int& nrhscoef = rawData.nrhscoef;

   /* prepare temorary constraint data */
   SCIP_VAR** consvars = 0;
   SCIP_Real* consvals = 0;
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &consvars, nvars) );
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &consvals, nvars) );

   /* loop through all constraints */
   for (int c = 0; c < nconss; ++c)
   {
      /* get constraint */
      SCIP_CONS* cons = conss[c];
      assert( cons != 0 );

      /* get constraint handler */
      SCIP_CONSHDLR* conshdlr = SCIPconsGetHdlr(cons);
      assert( conshdlr != 0 );
      const char* conshdlrname = SCIPconshdlrGetName(conshdlr);

      /* check type of constraint */
      if ( strcmp(conshdlrname, "linear") == 0 )
      {
         SCIP_CALL( collectCoefficients(scip, SCIPgetVarsLinear(scip, cons), SCIPgetValsLinear(scip, cons),
               SCIPgetNVarsLinear(scip, cons), SCIPgetLhsLinear(scip, cons), SCIPgetRhsLinear(scip, cons),
               SCIPconsIsTransformed(cons), FALSE, matcoef, matidx, matrhsidx, matvaridx, nmatcoef, nmaxmatcoef,
               rhscoef, rhstype, rhsidx, nrhscoef) );
      }
      else if ( strcmp(conshdlrname, "setppc") == 0 )
      {
         SCIP_VAR** linvars = SCIPgetVarsSetppc(scip, cons);
         int nconsvars = SCIPgetNVarsSetppc(scip, cons);

         switch ( SCIPgetTypeSetppc(scip, cons) )
         {
         case SCIP_SETPPCTYPE_PARTITIONING :
            SCIP_CALL( collectCoefficients(scip, linvars, 0, nconsvars, 1.0, 1.0, SCIPconsIsTransformed(cons), FALSE,
                  matcoef, matidx, matrhsidx, matvaridx, nmatcoef, nmaxmatcoef,
                  rhscoef, rhstype, rhsidx, nrhscoef) );
            break;
         case SCIP_SETPPCTYPE_PACKING :
            SCIP_CALL( collectCoefficients(scip, linvars, 0, nconsvars, -SCIPinfinity(scip), 1.0, SCIPconsIsTransformed(cons), FALSE,
                  matcoef, matidx, matrhsidx, matvaridx, nmatcoef, nmaxmatcoef,
                  rhscoef, rhstype, rhsidx, nrhscoef) );
            break;
         case SCIP_SETPPCTYPE_COVERING :
            SCIP_CALL( collectCoefficients(scip, linvars, 0, nconsvars, 1.0, SCIPinfinity(scip), SCIPconsIsTransformed(cons), FALSE,
                  matcoef, matidx, matrhsidx, matvaridx, nmatcoef, nmaxmatcoef,
                  rhscoef, rhstype, rhsidx, nrhscoef) );
            break;
         default:
            SCIPerrorMessage("Unknown setppc type %d.\n", SCIPgetTypeSetppc(scip, cons));
            return SCIP_ERROR;
         }
      }
      else if ( strcmp(conshdlrname, "xor") == 0 )
      {
         /* get number of variables of XOR constraint (should include integer variable) */
         int nconsvars;
         SCIP_Bool success;
         SCIP_CALL( SCIPgetConsNVars(scip, cons, &nconsvars, &success) );
         assert( success );
         assert( nconsvars <= nvars );
         assert( nconsvars == SCIPgetNVarsXor(scip, cons) + 1 );

         /* get variables of XOR constraint */
         SCIP_CALL( SCIPgetConsVars(scip, cons, consvars, nconsvars, &success) );
         assert( success );

         for (int v = 0; v < nconsvars; ++v)
         {
            /* mark integer variable with a coefficient of 2 to distinguish it from the other variables */
            if ( SCIPvarGetType(consvars[v]) == SCIP_VARTYPE_INTEGER )
               consvals[v] = 2.0;
            else
               consvals[v] = 1.0;
         }

         SCIP_CALL( collectCoefficients(scip, consvars, consvals, nconsvars,
               (SCIP_Real) SCIPgetRhsXor(scip, cons), (SCIP_Real) SCIPgetRhsXor(scip, cons), SCIPconsIsTransformed(cons), TRUE,
               matcoef, matidx, matrhsidx, matvaridx, nmatcoef, nmaxmatcoef, rhscoef, rhstype, rhsidx, nrhscoef) );
      }
      else if ( strcmp(conshdlrname, "logicor") == 0 )
      {
         SCIP_CALL( collectCoefficients(scip, SCIPgetVarsLogicor(scip, cons), 0, SCIPgetNVarsLogicor(scip, cons),
               1.0, SCIPinfinity(scip), SCIPconsIsTransformed(cons), FALSE,
               matcoef, matidx, matrhsidx, matvaridx, nmatcoef, nmaxmatcoef, rhscoef, rhstype, rhsidx, nrhscoef) );
      }
      else if ( strcmp(conshdlrname, "knapsack") == 0 )
      {
         SCIP_Longint* weights;

         SCIP_VAR** linvars = SCIPgetVarsKnapsack(scip, cons);
         int nconsvars = SCIPgetNVarsKnapsack(scip, cons);
         assert( nconsvars <= nvars );
         assert( consvals != 0 );

         /* copy Longint array to SCIP_Real array */
         weights = SCIPgetWeightsKnapsack(scip, cons);
         for (int v = 0; v < nconsvars; ++v)
            consvals[v] = (SCIP_Real) weights[v];

         SCIP_CALL( collectCoefficients(scip, linvars, consvals, nconsvars, -SCIPinfinity(scip),
               (SCIP_Real) SCIPgetCapacityKnapsack(scip, cons), SCIPconsIsTransformed(cons), FALSE,
               matcoef, matidx, matrhsidx, matvaridx, nmatcoef, nmaxmatcoef, rhscoef, rhstype, rhsidx, nrhscoef) );
      }
      else if ( strcmp(conshdlrname, "varbound") == 0 )
      {
         consvars[0] = SCIPgetVarVarbound(scip, cons);
         consvars[1] = SCIPgetVbdvarVarbound(scip, cons);

         consvals[0] = 1.0;
         consvals[1] = SCIPgetVbdcoefVarbound(scip, cons);

         SCIP_CALL( collectCoefficients(scip, consvars, consvals, 2, SCIPgetLhsVarbound(scip, cons),
               SCIPgetRhsVarbound(scip, cons), SCIPconsIsTransformed(cons), FALSE,
               matcoef, matidx, matrhsidx, matvaridx, nmatcoef, nmaxmatcoef, rhscoef, rhstype, rhsidx, nrhscoef) );
      }
      else if ( strcmp(conshdlrname, "symmetries") == 0 )
      {
         /* skip */
      }
      else if ( strcmp(conshdlrname, "bounddisjunction") == 0 )
      {
         /* currently assume bound disjunctions are o.k. for local symmetry groups */
         if ( ! local )
         {
            /* @todo we need to handle bounddisjunctions if local symmetry groups are considered */
            SCIPerrorMessage("Cannot determine symmetries for constraint <%s> of constraint handler <%s>.\n",
               SCIPconsGetName(cons), SCIPconshdlrGetName(conshdlr) );
            return SCIP_ERROR;
         }
      }
      else
      {
         SCIPerrorMessage("Cannot determine symmetries for constraint <%s> of constraint handler <%s>.\n",
            SCIPconsGetName(cons), SCIPconshdlrGetName(conshdlr) );
         return SCIP_ERROR;
      }
   }
   assert( (int) nrhscoef <= 2 * nconss );
   assert( nrhscoef > 0 ); /* cannot have empty rows! */

   SCIPfreeBlockMemoryArray(scip, &consvals, nvars);
   SCIPfreeBlockMemoryArray(scip, &consvars, nvars);

#ifndef NDEBUG
   /* if symmetries have to be checked store original matrix and rhs coefficient arrays */
   SCIP_Real*& matcoeforig = rawData.matcoeforig;
   SCIP_CALL( SCIPduplicateBlockMemoryArray(scip, &matcoeforig, matcoef, nmatcoef) );
   SCIP_Real*& rhscoeforig = rawData.rhscoeforig;
   SCIP_CALL( SCIPduplicateBlockMemoryArray(scip, &rhscoeforig, rhscoef, nrhscoef) );
#endif

   /* determine nonredundant list of coefficients: first sort */
   SCIPsortRealInt(matcoef, matidx, nmatcoef);
   SCIPsortRealInt(rhscoef, rhsidx, nrhscoef);

#ifdef SCIP_DEBUG
   for (unsigned int j = 0; j < nrhscoef; ++j)
      assert( SCIPisEQ(scip, rhscoef[j], rhscoeforig[rhsidx[j]]) );
#endif

   /* create maps for coefficients to indices */

   /* map SCIP variables */
   Varcmp varcmp(scip, fixedtype, local);
   VarMapType varmap(varcmp);

   /* map SCIP_Reals by using the SCIP feasibility tolerance */
   Realcmp realcmp(scip);
   RealMapType matcoefmap(realcmp);

   Rhspaircmp rhspaircmp(scip);
   RhsPairMapType rhscoefmap(rhspaircmp);

   /* determine number of different coefficents */
   unsigned int nuniquematcoef = 0;
   unsigned int nuniquerhscoef = 0;
   unsigned int nuniquevars = 0;

   /* find interchangable variables: same objective, lower and upper bounds, and variable type (see Varcmp above) */
   for (unsigned int j = 0; j < (unsigned int) nvars; ++j)
   {
      SCIP_VAR* var = vars[j];
      assert( var != 0 );

      if ( varmap.find(var) == varmap.end() )
      {
         varmap[var] = nuniquevars++;
#ifdef SCIP_OUTPUT
         SCIPdebugMessage("detected variable <%s> (probindex: %d, obj: %g, lb: %g, ub: %g, type: %d) - color %u\n",
            SCIPvarGetName(var), SCIPvarGetProbindex(var), SCIPvarGetObj(var), SCIPvarGetLbGlobal(var), SCIPvarGetUbGlobal(var),
            SCIPvarGetType(var), nuniquevars-1);
#endif
      }
   }

   /* find different matrix entries (use sorting to avoid too many map calls) */
   SCIP_Real oldcoef = SCIP_INVALID;
   for (unsigned int j = 0; j < nmatcoef; ++j)
   {
      assert( 0 <= matidx[j] && matidx[j] < (int) nmatcoef );
      SCIP_Real val = matcoef[j];
      if ( ! SCIPisEQ(scip, val, oldcoef) )
      {
         assert( matcoefmap.find(val) == matcoefmap.end() );
         matcoefmap[val] = nuniquematcoef++;
         oldcoef = val;
#ifdef SCIP_OUTPUT
         SCIPdebugMessage("have new matrix entry %f - color: %u\n", val, nuniquematcoef-1);
#endif
      }
   }

   /* find different rhs */
   int idx = rhsidx[0];
   assert( 0 <= idx && idx < (int) nrhscoef );
   oldcoef = SCIP_INVALID;
   SYMconsType oldtype = rhstype[idx];
   for (unsigned int j = 0; j < nrhscoef; ++j)
   {
      idx = rhsidx[j];
      assert( 0 <= idx && idx < (int) nrhscoef );
      SCIP_Real val = rhscoef[j];
      if ( ! SCIPisEQ(scip, val, oldcoef) || oldtype != rhstype[idx] )
      {
         const RhsPair valPair = std::make_pair(val, rhstype[idx]);
         if ( rhscoefmap.find(valPair) == rhscoefmap.end() )
         {
            rhscoefmap[valPair] = nuniquerhscoef++;
#ifdef SCIP_OUTPUT
            SCIPdebugMessage("have new rhs %f, type: %u - color: %u\n", val, rhstype[idx], nuniquerhscoef-1);
#endif
         }
         oldcoef = val;
         oldtype = rhstype[idx];
      }
   }
   assert( nuniquerhscoef > 0 );
   assert( nuniquematcoef > 0 );
   assert( nuniquevars > 0 );

   SCIPdebugMessage("Number of detected different variables: %u (total: %d).\n", nuniquevars, nvars);
   SCIPdebugMessage("Number of detected different nonzero coefficients: rhs: %u (total %u), matrix: %u (total %u).\n",
      nuniquerhscoef, nrhscoef, nuniquematcoef, nmatcoef);

   /* create bliss graph */
   G = new bliss::Graph();

   /* compute number of layers for edge-color emulation */
   unsigned int& T = rawData.T;
   T = (unsigned int) std::ceil( log2((double) nuniquematcoef + 1.0) );
   assert( T > 0 );

   /* estimate size of graph based on edge-color emulation */
   const unsigned int V1 = T * (nvars + nrhscoef);
   const unsigned int E1 = T * (nvars + nrhscoef - 2) + std::min(T/2, 1u) * nmatcoef;

   /* estimate size of graph based on vertices for matrix coefficients */
   const unsigned int V2 = (unsigned) nvars + nrhscoef + nmatcoef;
   const unsigned int E2 = 2 * nmatcoef;

   SCIPdebugMessage("Estimated size of construction #1: |V| = %u,|E| = %u.\n", V1, E1);
   SCIPdebugMessage("Estimated size of construction #2: |V| = %u,|E| = %u.\n", V2, E2);

   //   if (V2 < V1 && E2 < E1)
   SCIP_Bool success = FALSE;
   if ( uselayeredgraph )
   {
      SCIP_CALL( fillGraphByColorLayers(*G, scip, T, vars, nvars, varmap, matcoefmap, rhscoefmap,
            nuniquematcoef, nuniquerhscoef, nuniquevars, matcoef, matidx, matrhsidx, matvaridx, nmatcoef,
            rhscoef, rhstype, rhsidx, nrhscoef, nnodes, nedges, success) );
   }
   else
   {
      T = 1;
      SCIP_CALL( fillGraphByColoredCoefficients(*G, scip, vars, nvars, varmap, matcoefmap, rhscoefmap,
            nuniquematcoef, nuniquerhscoef, nuniquevars, matcoef, matidx, matrhsidx, matvaridx, nmatcoef,
            rhscoef, rhstype, rhsidx, nrhscoef, nnodes, nedges, success) );
   }
   if ( ! success )
   {
      delete G;
      G = 0;
      return SCIP_OKAY;
   }

#ifdef SCIP_OUTPUT
   G->write_dot("debug.dot");
#endif

   SCIPdebugMessage("Symmetry detection graph has %u nodes.\n", G->get_nof_vertices());

   return SCIP_OKAY;
}

/** Compute symmetry group of global binary MIP */
SCIP_RETCODE computeSymmetry(
   SCIP*                 scip,               /**< SCIP pointer */
   unsigned int          maxgenerators,      /**< maximal number of generators constructed (= 0 if unlimited) */
   SCIP_Bool             uselayeredgraph,    /**< use layered graph for symmetry detection */
   int                   fixedtype,          /**< variable types that must be fixed by symmetries */
   SCIP_Bool             local,              /**< Use local variable bounds? */
   SCIP_VAR***           permvars,           /**< variables on which permutations act */
   unsigned int*         npermvars,          /**< number of variables */
   unsigned int***       perms,              /**< permutations */
   unsigned int*         nperms,             /**< number of permutations */
   unsigned int*         nmaxperms,          /**< maximal number of permutations (needed for freeing storage) */
   unsigned int**        baseindices,        /**< indices of variables that form a base for the strong generating set of permutations */
   unsigned int*         nbaseindices,       /**< number of base variables */
   unsigned int*         nbases              /**< number of bases */
   )
{
   assert( scip != 0 );
   assert( permvars != 0 );
   assert( npermvars != 0 );
   assert( perms != 0 );
   assert( nperms != 0 );
   assert( nmaxperms != 0 );
   assert( baseindices != 0 );
   assert( nbaseindices != 0 );

   /* init */
   *permvars = 0;
   *npermvars = 0;
   *perms = 0;
   *nperms = 0;
   *nmaxperms = 0;
   *baseindices = 0;
   *nbaseindices = 0;
   *nbases = 0;

   /* start to measure time */
   struct tms timer_beg;
   struct tms timer_end;
   times(&timer_beg);

   GraphRawData rawData(scip);
   bliss::Graph* G = 0;

   SCIP_CALL( rawData.initialize() );
   unsigned int nnodes = 0;
   unsigned int nedges = 0;
   SCIP_CALL( constructGraph(scip, uselayeredgraph, fixedtype, local, G, rawData, nnodes, nedges) );

   if ( G == 0 )
   {
      SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Graph construction failed.\n");
      return SCIP_OKAY;
   }
   SCIPdebugMessage("Starting symmetry detection.\n");

   /* compute automorphisms */
   bliss::Stats stats;
   BLISS_Data data;
   data.scip = scip;
   data.nvars = (unsigned) rawData.nvars;
   data.nperms = 0;
   data.nmaxperms = (unsigned) 100 * rawData.nvars;
   data.nlevels = rawData.T;
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &data.perms, data.nmaxperms) );
   /* Prefer splitting partition cells corresponding to variables over those corresponding
    * to inequalities. This is because we are only interested in the action
    * of the automorphism group on the variables, and we need a base for this action */
   G->set_splitting_heuristic(bliss::Graph::shs_f);
   /* disable component recursion as advised by Tommi Junttila from bliss */
   G->set_component_recursion(false);

   /* do not use a node limit, but set generator limit */
   G->set_search_limits(0, maxgenerators);

   /* start search */
   G->find_automorphisms(stats, blisshook, (void*) &data);
#ifdef SCIP_OUTPUT
   stats.print(stdout);
#endif

#ifndef NDEBUG
   if ( ! SCIPisStopped(scip) )
   {
      SCIP_CALL( checkSymmetriesAreSymmetries(scip, rawData.vars, rawData.nvars, rawData.matcoeforig, rawData.matrhsidx, rawData.matvaridx, rawData.nmatcoef,
            rawData.rhscoeforig, rawData.rhstype, rawData.nrhscoef, data.nperms, data.perms, fixedtype) );
      SCIPfreeBlockMemoryArray(scip, &rawData.rhscoeforig, rawData.nrhscoef);
      SCIPfreeBlockMemoryArray(scip, &rawData.matcoeforig, rawData.nmatcoef);
   }
#endif

   /* Try to extract BSGS information from bliss.
    * We are especially interested in the information
    * for which base the generators are a strong generating set.
    * (They should be one as they are obtained by some partition backtracking approach).
    *
    * The main obstacle is that our graph contains not only nodes for "variables",
    * but also for "constraints". If bliss splits a partition based on a constraint
    * it is unclear how this induces a base point.
    */
   std::vector<bliss::PathInfo> firstPathInfo = G->get_first_path_info();
   unsigned int nBaseVars = (unsigned int) firstPathInfo.size();
   unsigned int* baseVars = 0;
   unsigned int baseCounter = 0;
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &baseVars, nBaseVars) );
   for (unsigned int i = 0; i < nBaseVars; ++i)
   {
      const unsigned int baseElement = firstPathInfo[i].splitting_element / rawData.T;

      /* ignore all base points that do not correspond to a "variable" node */
      if (baseElement >= static_cast<unsigned int>(rawData.nvars))
         continue;

      /* If bliss automorphism search is aborted due to imposed limits,
       * then we have to remove obviously redundant elements from the base.
       * We need only those which are moved by at least one generator.
       */
      bool moved = false;
      for (unsigned j = 0; j < data.nperms; ++j)
      {
         if (data.perms[j][baseElement] != baseElement)
         {
            moved = true;
            break;
         }
      }

      if (moved)
         baseVars[baseCounter++] = baseElement;
   }

   SCIPdebugMessage("found %u from %u possible baseVars\n", baseCounter, nBaseVars);

   /* output time */
   times(&timer_end);
   if ( ! local )
   {
      double t = (timer_end.tms_utime - timer_beg.tms_utime) / (double)sysconf(_SC_CLK_TCK);
      if ( uselayeredgraph )
         SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Time for symmetry detection:\t\t%4.2f \t(layered graph, %u nodes, %u edges)\n", t, nnodes, nedges);
      else
         SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Time for symmetry detection:\t\t%4.2f \t(matrix graph, %u nodes, %u edges)\n", t, nnodes, nedges);
      if ( maxgenerators == 0 )
         SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Number of generators:\t\t\t%u \t(max: -)\n", data.nperms);
      else
         SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Number of generators:\t\t\t%u \t(max: %u)\n", data.nperms, maxgenerators);
   }

   /* free graph */
   delete G;

   /* prepare return values */
   *permvars = rawData.vars;
   *npermvars = (unsigned int) rawData.nvars;
   *perms = data.perms;
   *nperms = data.nperms;
   *nmaxperms = data.nmaxperms;
   *baseindices = baseVars;
   *nbaseindices = nBaseVars;
   *nbases = baseCounter;

   return SCIP_OKAY;
}


/** Compute a subgroup of the symmetry group of global MIP by transposition heuristics */
SCIP_RETCODE computeSymmetryByHeuristic(
   SCIP*                 scip,               /**< SCIP pointer */
   SCIP_Bool             uselayeredgraph,    /**< use layered graph for symmetry detection */
   int                   fixedtype,          /**< variable types that must be fixed by symmetries */
   SCIP_Bool             local,              /**< Use local variable bounds? */
   SCIP_VAR***           permvars,           /**< variables on which permutations act */
   unsigned int*         npermvars,          /**< number of variables */
   unsigned int**        orbits,             /**< orbits */
   unsigned int*         norbits,            /**< number of orbits */
   unsigned int**        orbitsizes,         /**< array of size of each orbit sizes into which the entries in the array orbits are grouped */
   unsigned int*         norbitsizes         /**< number of entries in array orbitsizes */
   )
{
   assert( scip != 0 );
   assert( permvars != 0 );
   assert( npermvars != 0 );
   assert( orbits != 0 );
   assert( norbits != 0 );
   assert( orbitsizes != 0 );
   assert( norbitsizes != 0 );

   /* init */
   *permvars = 0;
   *npermvars = 0;
   *orbits = 0;
   *norbits = 0;
   *orbitsizes = 0;
   *norbitsizes = 0;

   /* start to measure time */
   struct tms timer_beg;
   struct tms timer_end;
   times(&timer_beg);

   GraphRawData rawData(scip);
   bliss::Graph* G = 0;

   SCIP_CALL( rawData.initialize() );
   unsigned int nnodes = 0;
   unsigned int nedges = 0;
   SCIP_CALL( constructGraph(scip, uselayeredgraph, fixedtype, local, G, rawData, nnodes, nedges) );

   if ( G == 0 )
   {
      SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Graph construction failed.\n");
      return SCIP_OKAY;
   }

   SCIPdebugMessage("Starting symmetry detection.\n");

   /* compute automorphisms */
   bliss::Stats stats;
   BLISS_DataSymHeuristic data;
   data.scip = scip;
   data.nvars = (unsigned) rawData.nvars;
   data.sheur = new SymmetricGroupRecognitionHeuristic<PERMUTATION>(rawData.nvars);
   data.nlevels = rawData.T;
   data.ignoredgens = 0;

   /* start search */
   G->find_automorphisms(stats, blisshookheur, (void*) &data);
#ifdef SCIP_OUTPUT
   stats.print(stdout);
#endif

#ifndef NDEBUG
   SCIPfreeBlockMemoryArray(scip, &rawData.rhscoeforig, rawData.nrhscoef);
   SCIPfreeBlockMemoryArray(scip, &rawData.matcoeforig, rawData.nmatcoef);
#endif

   /* output time */
   times(&timer_end);
   if ( ! local )
   {
      double t = (timer_end.tms_utime - timer_beg.tms_utime) / (double)sysconf(_SC_CLK_TCK);
      if ( uselayeredgraph )
         SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Time for symmetry detection:\t\t%4.2f \t(layered graph, %u nodes, %u edges)\n", t, nnodes, nedges);
      else
         SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Time for symmetry detection:\t\t%4.2f \t(matrix graph, %u nodes, %u edges)\n", t, nnodes, nedges);
   }

   /* prepare return values */
   *permvars = rawData.vars;
   *npermvars = (unsigned int) rawData.nvars;
   std::vector<std::list<dom_int> > symmetricGroupOrbits;
   data.sheur->symmetricGroupOrbits(symmetricGroupOrbits);

   /* output whether we found the full symmetry group */
   if ( ! local )
   {
      if ( data.ignoredgens == 0 )
         SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Heuristic symmetry detection:\t\t%u orbits of full symmetry group\n", (unsigned int) symmetricGroupOrbits.size());
      else
         SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Heuristic symmetry detection:\t\t%u orbits, ignored %u generators\n", (unsigned int) symmetricGroupOrbits.size(), data.ignoredgens);
   }

   *norbits = (unsigned) rawData.nvars;
   *norbitsizes = (unsigned int) symmetricGroupOrbits.size();
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, orbits, *norbits) );
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, orbitsizes, *norbitsizes) );

   std::vector<std::list<dom_int> >::const_iterator vit;
   unsigned int i_orbits = 0;
   unsigned int i_orbitsizes = 0;
   for (vit = symmetricGroupOrbits.begin(); vit != symmetricGroupOrbits.end(); ++vit)
   {
      (*orbitsizes)[i_orbitsizes] = (*vit).size();
      ++i_orbitsizes;

      for (std::list<dom_int>::const_iterator lit = (*vit).begin(); lit != (*vit).end(); ++lit)
      {
         (*orbits)[i_orbits] = *lit;
         ++i_orbits;
      }
   }

   delete data.sheur;

   return SCIP_OKAY;
}


/* vim: set tabstop=3 expandtab: */
// kate: space-indent on; indent-width 3; mixedindent off; indent-mode cstyle;
