/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*    This file is part of "mipsymmetries" a collection of                   */
/*    symmetry handling methods.                                             */
/*                                                                           */
/*    Copyright (C) 2010-2015  Marc Pfetsch, Thomas Rehn                     */
/*                                                                           */
/*                                                                           */
/*    Based on SCIP  --- Solving Constraint Integer Programs                 */
/*       SCIP is distributed under the terms of the SCIP Academic Licence,   */
/*       see file COPYING in the SCIP distribution.                          */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
#pragma ident "@(#) $Id: main.cpp 35 2011-06-23 14:47:43Z rehn $"

/**@file   main.cpp
 * @brief  main file for testing computation of IP symmetries
 * @author Marc Pfetsch
 */

/*---+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8----+----9----+----0----+----1----+----2*/


#include <scip/scip.h>
#include <scip/scipdefplugins.h>

#include "cons_subtreebool.h"
#include "cons_symmetries.h"
#include "branch_first.h"
#include "prop_isoprune.h"
#include "branch_orbital.h"
#include "prop_orbitalfixing.h"
#include "presol_lpsymfix.h"
#include "presol_symbreak.h"
#include "presol_orbitprobing.h"
#include "readArguments.h"

#include <iostream>
#include <fstream>

#include "handle_symmetries.h"
#include <permlib/permlib_api.h>
#include <permlib/construct/known_bsgs_construction.h>


/** run scip with commandline arguments */
static
SCIP_RETCODE runSCIP(
   int                   argc,
   char**                argv
   )
{
   SCIP* scip = 0;
   const char* filename;
   const char* settingsname;
   const char* solutionfile;
   SCIP_Real timeLimit;
   SCIP_Real memLimit;
   SCIP_Longint nodeLimit;
   SCIP_Bool noSymmetry = FALSE;
   SCIP_Bool onlypre = FALSE;
   SCIP_Bool useorbital = FALSE;
   SCIP_Bool useorbitalfixing = FALSE;
   SCIP_Bool useisoprune = FALSE;
   SCIP_Bool usesymbreak = FALSE;
   SCIP_Bool useorbitprobing = FALSE;
   SCIP_Bool uselpsymfix = FALSE;
   SCIP_Bool usebranchfirst = FALSE;
   SCIP_Bool forcesymmetrydetection = FALSE;
   int requiredsymmetry = 0;
   SCIP_RETCODE retcode;
   SCIP_Real cutoffvalue = SCIP_INVALID;
   int dispFreq;
   int permseed;

   /* parse command line arguments */
   retcode = readArguments(argc, argv, &filename, &solutionfile, &settingsname, &timeLimit, &memLimit, &nodeLimit, &dispFreq, &noSymmetry, &onlypre, &permseed, &cutoffvalue);
   if ( retcode != SCIP_OKAY )
      exit(1);
   assert( filename != 0 );

   /* initialize SCIP */
   SCIP_CALL( SCIPcreate(&scip) );

   SCIPprintVersion(scip, 0);

   SCIPinfoMessage(scip, 0, "\n");
   SCIPinfoMessage(scip, 0, "Symmetry handling routines - (c) Marc Pfetsch, Thomas Rehn [svn: %s]\n", SVN_REV);
   SCIPinfoMessage(scip, 0, "\n");

   /* include default SCIP plugins */
   SCIP_CALL( SCIPincludeDefaultPlugins(scip) );

   /* add subtreebool constraint handler */
   SCIP_CALL( SCIPincludeConshdlrSubtreebool(scip) );

   /* add symmetries constraint handler */
   SCIP_CALL( SCIPincludeConshdlrSymmetries(scip) );

   /* include orbit probing presolver */
   SCIP_CALL( SCIPincludePresolOrbitprobing(scip) );

   /* include orbital branching */
   SCIP_CALL( SCIPincludeBranchruleOrbital(scip) );

   /* include isomorphism pruning */
   SCIP_CALL( SCIPincludePropIsoprune(scip) );

   /* include orbital fixing */
   SCIP_CALL( SCIPincludePropOrbitalfixing(scip) );

   /* include lpsymfix presolver */
   SCIP_CALL( SCIPincludePresolLpsymfix(scip) );

   /* include symmetry breaking presolver */
   SCIP_CALL( SCIPincludePresolSymbreak(scip) );



   /* --------------------------------------------------------------------- */
   /* turn off several SCIP-components: */

   /* gateextraction seems to modify the problem in a way that cannot be captured by our symmetry detection. */
   SCIP_CALL( SCIPsetIntParam(scip, "presolving/gateextraction/maxrounds", 0) );
   SCIPinfoMessage(scip, 0, "\nTurned off presolver <gateextraction>.\n");

   /* --------------------------------------------------------------------- */
   /* handle permutations */
   if ( permseed >= 0 )
   {
      SCIP_CALL( SCIPsetIntParam(scip, "misc/permutationseed", permseed) );
   }

   /* --------------------------------------------------------------------- */

   /* add our own parameters */
   SCIP_CALL( SCIPaddBoolParam(scip,
         "useorbital", "Use orbital branching?",
         &useorbital, TRUE, FALSE, 0, 0) );

   SCIP_CALL( SCIPaddBoolParam(scip,
         "useisoprune", "Use isomorphism pruning?",
         &useisoprune, TRUE, FALSE, 0, 0) );

   SCIP_CALL( SCIPaddBoolParam(scip,
         "useorbitalfixing", "Use orbital fixing?",
         &useorbitalfixing, TRUE, FALSE, 0, 0) );

   SCIP_CALL( SCIPaddBoolParam(scip,
         "usebranchfirst", "Branch on first variable?",
         &usebranchfirst, TRUE, FALSE, 0, 0) );

   SCIP_CALL( SCIPaddBoolParam(scip,
         "usesymbreak", "Use symmetry breaking inequalities?",
         &usesymbreak, TRUE, FALSE, 0, 0) );

   SCIP_CALL( SCIPaddBoolParam(scip,
         "useorbitprobing", "Use orbit probing?",
         &useorbitprobing, TRUE, FALSE, 0, 0) );

   SCIP_CALL( SCIPaddBoolParam(scip,
         "uselpsymfix", "Use projection of symmetric LP-variables onto the fixed space?",
         &uselpsymfix, TRUE, FALSE, 0, 0) );

   SCIP_CALL( SCIPaddBoolParam(scip,
         "forcesymmetrydetection", "Compute symmetry group even if it is not used?",
         &forcesymmetrydetection, TRUE, FALSE, 0, 0) );

   SCIP_CALL( SCIPaddIntParam(scip,
         "requiredsymmetry", "Required symmetry if forcesymmetrydetection is turned on: 0 - all symmetries; 1 - symmetry on binary variables & integer variables fixed; 2 - symmetry on continuous variable & fixed integer and binary variables",
         &requiredsymmetry, TRUE, 0, 0, 2, 0, 0) );

   /* set time, node, and memory limit */
   if ( ! SCIPisInfinity(scip, timeLimit) )
      SCIP_CALL( SCIPsetRealParam(scip, "limits/time", timeLimit) );
   if ( ! SCIPisInfinity(scip, memLimit) )
      SCIP_CALL( SCIPsetRealParam(scip, "limits/memory", memLimit) );
   if ( nodeLimit < SCIP_LONGINT_MAX )
      SCIP_CALL( SCIPsetLongintParam(scip, "limits/nodes", nodeLimit) );
   if ( dispFreq >= 0 )
      SCIP_CALL( SCIPsetIntParam(scip, "display/freq", dispFreq) );

   /* check for parameters */
   if ( settingsname != 0 )
   {
      if ( SCIPfileExists(settingsname) )
      {
         SCIPinfoMessage(scip, 0, "reading parameter file <%s> ...\n\n", settingsname);
         SCIP_CALL( SCIPreadParams(scip, settingsname) );
      }
      else
      {
         SCIPerrorMessage("parameter file <%s> not found - using default parameters.\n", settingsname);
      }
   }

   if ( usebranchfirst )
   {
      SCIP_CALL( SCIPincludeBranchruleFirst(scip) );
   }

   /* deactivate symmetry if requested */
   if ( noSymmetry )
   {
      SCIP_CALL( SCIPsetBoolParam(scip, "useorbital", FALSE) );
      SCIP_CALL( SCIPsetBoolParam(scip, "useisoprune", FALSE) );
      SCIP_CALL( SCIPsetBoolParam(scip, "useorbitalfixing", FALSE) );
      SCIP_CALL( SCIPsetBoolParam(scip, "usesymbreak", FALSE) );
      SCIP_CALL( SCIPsetBoolParam(scip, "useorbitprobing", FALSE) );
      SCIP_CALL( SCIPsetBoolParam(scip, "uselpsymfix", FALSE) );
   }

   /* output changed parameters */
   SCIPinfoMessage(scip, 0, "Changed settings:\n");
   SCIP_CALL( SCIPwriteParams(scip, 0, FALSE, TRUE) );
   SCIPinfoMessage(scip, 0, "\n");

   /* solve problem */
   if ( ! onlypre )
      SCIPinfoMessage(scip, 0, "\nsolving problem ...\n\n");
   else
      SCIPinfoMessage(scip, 0, "\nrunning preprocessing ...\n\n");

   /* read problem */
   SCIP_CALL( SCIPreadProb(scip, filename, 0) );

   /* add symmetry handling constraint */
   SCIP_CONS* cons;
   SCIP_CALL( SCIPcreateConsSymmetries(scip, &cons, "symmetries") );
   SCIP_CALL( SCIPaddCons(scip, cons) );
   SCIP_CALL( SCIPreleaseCons(scip, &cons) );

   /* possibly read solution */
   if ( solutionfile != 0 )
   {
      SCIP_CALL( SCIPreadSol(scip, solutionfile) );
   }

   /* handle cutoff */
   if ( cutoffvalue != SCIP_INVALID )
   {
      SCIPinfoMessage(scip, 0, "\nSetting cutoff value to %g.\n\n", cutoffvalue);
      SCIP_CALL( SCIPsetObjlimit(scip, cutoffvalue) );
      SCIP_CALL( SCIPsetHeuristics(scip, SCIP_PARAMSETTING_OFF, TRUE) );
   }

   if ( forcesymmetrydetection )
   {
      SCIP_CONSHDLR* conshdlr = SCIPfindConshdlr(scip, "symmetries");
      if ( conshdlr == 0 )
      {
         SCIPerrorMessage("Could not find constraint handler <symmetries>.\n");
         return SCIP_ERROR;
      }

      switch ( requiredsymmetry )
      {
      case 0:
         SYMsetProbSpecRequirement(conshdlr, SCIP_PROBSPEC_BINARY);
         SYMsetProbSpecRequirement(conshdlr, SCIP_PROBSPEC_INTEGER);
         SYMsetProbSpecRequirement(conshdlr, SCIP_PROBSPEC_REAL);
         break;

      case 1:
         SYMsetProbSpecRequirement(conshdlr, SCIP_PROBSPEC_BINARY);
         SYMsetProbSpecRequirementFixed(conshdlr, SCIP_PROBSPEC_INTEGER);
         break;

      case 2:
         SYMsetProbSpecRequirement(conshdlr, SCIP_PROBSPEC_REAL);
         SYMsetProbSpecRequirementFixed(conshdlr, SCIP_PROBSPEC_BINARY);
         SYMsetProbSpecRequirementFixed(conshdlr, SCIP_PROBSPEC_INTEGER);
         break;

      default:
         SCIPerrorMessage("Wrong type for symmetry requirements: %d.\n", requiredsymmetry);
         return SCIP_ERROR;
      }
   }

   if ( onlypre )
   {
      SCIP_CALL( SCIPpresolve(scip) );
   }
   else
   {
      SCIP_CALL( SCIPsolve(scip) );
   }

   SCIP_CALL( SCIPprintStatistics(scip, 0) );
#if 0
   if ( SCIPgetBestSol(scip) != 0 )
   {
      SCIP_CALL( SCIPprintSol(scip, SCIPgetBestSol(scip), 0, FALSE) );
   }
#endif

   SCIP_CALL( SCIPfreeProb(scip) );

   // SCIPprintMemoryDiagnostic(scip);

   /* free */
   SCIP_CALL( SCIPfree(&scip) );

   BMScheckEmptyMemory();

   return SCIP_OKAY;
}



/** main function */
int
main(
   int                   argc,
   char**                argv
   )
{
   SCIP_RETCODE retcode;

   retcode = runSCIP(argc, argv);
   if ( retcode != SCIP_OKAY )
   {
      SCIPprintError(retcode);
      return -1;
   }

   return 0;
}

// kate: space-indent on; indent-width 3; mixedindent off; indent-mode cstyle;
