/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*    This file is part of "mipsymmetries" a collection of                   */
/*    symmetry handling methods.                                             */
/*                                                                           */
/*    Copyright (C) 2010-2015  Marc Pfetsch, Thomas Rehn                     */
/*                                                                           */
/*                                                                           */
/*    Based on SCIP  --- Solving Constraint Integer Programs                 */
/*       SCIP is distributed under the terms of the SCIP Academic Licence,   */
/*       see file COPYING in the SCIP distribution.                          */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
#pragma ident "@(#) $Id: maxkcol.cpp 1216 2015-10-27 10:03:31Z pfetsch $"

/**@file   maxkcol.cpp
 * @brief  main file for testing isomorphism pruning on the maximum k-colorable subgraph problem
 * @author Marc Pfetsch
 */

/*---+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8----+----9----+----0----+----1----+----2*/


#include <scip/scip.h>
#include <scip/scipdefplugins.h>

#include "cons_subtreebool.h"
#include "cons_symmetries.h"
#include "branch_first.h"
#include "prop_isoprune.h"
#include "branch_orbital.h"
#include "prop_orbitalfixing.h"
#include "presol_lpsymfix.h"
#include "presol_symbreak.h"
#include "graph.h"
#include "readArguments.h"

#include <iostream>
#include <fstream>

#include "handle_symmetries.h"
#include <permlib/permlib_api.h>
#include <permlib/construct/known_bsgs_construction.h>

/* get next number in string */
static
long get_next_number(char*& s)
{
   /* skip whitespace */
   while ( isspace(*s) )
      ++s;
   long tmp = std::atol(s);
   /* number */
   while ( (*s != 0) && (! isspace(*s)) )
      ++s;
   return tmp;
}


/** read graph */
static
SCIP_RETCODE readGraph(
   SCIP*                 scip,               /**< SCIP data structure */
   Graph&                G,                  /**< empty graph at the beginning */
   const char*           filename            /**< name of the input file */
   )
{
   std::ifstream file(filename, std::ios::in);
   if (! file)
   {
      SCIPerrorMessage("Error: could not open %s\n", filename);
      return SCIP_NOFILE;
   }

   char* buffer = new char [1024];
   int line = 1;

   SCIPinfoMessage(scip, 0, "Reading file %s ...\n", filename);

   /* get number of nodes and edges */
   do
   {
      (void) file.getline(buffer, 1023);
      ++line;
   }
   while ( ! file.fail() && ! file.eof() && buffer[0] != 'p');

   if ( file.eof() || file.fail() )
   {
      SCIPerrorMessage("%s: error! Could not find line starting with 'p'.\n", filename);
      delete [] buffer;
      return SCIP_INVALIDDATA;
   }

   if ( buffer[2] != 'e' || buffer[3] != 'd' || buffer[4] != 'g' || buffer[5] != 'e' )
   {
      SCIPerrorMessage("%s: line %d: error! Line starting with 'p' must continue with 'edge'!\n%s\n", filename, line, buffer);
      delete [] buffer;
      return SCIP_INVALIDDATA;
   }

   char* str = &(buffer[6]);

   /* if line reads 'edges' (non-standard!), instead of 'edge'. */
   if ( *str == 's' )
      ++str;

   /* read sizes */
   int Ln = get_next_number(str);
   int Lm = get_next_number(str);

   if ( Ln <= 0 )
   {
      SCIPerrorMessage("%s: line %d: error! Number of vertices must be positive!\n%s\n", filename, line, buffer);
      delete [] buffer;
      return SCIP_INVALIDDATA;
   }
   if ( Lm < 0 )
   {
      SCIPerrorMessage("%s: line %d: error! Number of edges must be nonnegative!\n%s\n", filename, line, buffer);
      delete [] buffer;
      return SCIP_INVALIDDATA;
   }

   unsigned int ln = (unsigned int) Ln;
   unsigned int lm = (unsigned int) Lm;

   /* collection of edges */
   std::list<std::pair<unsigned int,unsigned int> > EdgeList;

   /* read vertices and edges */
   (void) file.getline(buffer, 1023);
   ++line;
   assert( ! file.fail() );

   std::vector<bool> covered(ln,false);
   while ( ! file.eof() && ! file.fail() )
   {
      str = buffer;
      if ( str[0] == 'e' )
      {
         ++str;
         int source = get_next_number(str) -1;
         int target = get_next_number(str) -1;

         /* check for source */
         if ( source < 0 || source >= (int) ln || target < 0 || target >= (int) ln )
         {
            SCIPerrorMessage("%s: line %d: error! Node number invalid!\n%s\n", filename, line, buffer);
            delete [] buffer;
            return SCIP_INVALIDDATA;
         }

         /* ignore self-loops */
         if ( source != target )
            EdgeList.push_back( std::make_pair((unsigned int) source, (unsigned int) target) );

         covered[source] = true;
         covered[target] = true;
      }
      (void) file.getline(buffer, 1023);
      ++line;
   }

   /* create vertices */
   std::vector<Vertex> M(ln);
   unsigned int nIsolatedNodes = 0;
   for (unsigned int i = 0; i < (unsigned int) ln; ++i)
   {
      Vertex v = boost::add_vertex(G);
      M[i] = v;
      if (! covered[i] )
         ++nIsolatedNodes;
   }

   if ( nIsolatedNodes > 0 )
   {
      SCIPwarningMessage(scip, "Graph contains %d isolated nodes!\n", nIsolatedNodes );
   }

   /* create edges */
   unsigned int nEdges = 0;
   unsigned int duplicateEdges = 0;
   std::list<std::pair<unsigned int,unsigned int> >::const_iterator eit, eend = EdgeList.end();
   for (eit = EdgeList.begin(); eit != eend; ++eit)
   {
      unsigned int source = eit->first;
      unsigned int target = eit->second;

      assert( source < ln );
      assert( target < ln );

      Vertex s = M[source];
      Vertex t = M[target];

      /* if edge is not already present */
      if (! boost::edge(s, t, G).second )
      {
         std::pair<Edge, bool> p = boost::add_edge(s, t, G);
         ++nEdges;
         if (! p.second)
         {
            SCIPerrorMessage("Could not create edge {%u,%u}.\n", source, target);
            delete [] buffer;
            return SCIP_INVALIDDATA;
         }
      }
      else
         ++duplicateEdges;
   }
   SCIPinfoMessage(scip, 0, "Duplicate edges: %u\n", duplicateEdges);

   if ( nEdges + duplicateEdges != lm )
   {
      SCIPerrorMessage("%s: line %d: error! Found only %u. There should be %u.\n", filename, nEdges + duplicateEdges, lm);
      delete [] buffer;
      return SCIP_INVALIDDATA;
   }

   delete [] buffer;

   return SCIP_OKAY;
}




/** run scip with commandline arguments */
static
SCIP_RETCODE runSCIP(
   int                   argc,
   char**                argv
   )
{
   SCIP* scip = 0;
   const char* filename;
   const char* settingsname;
   const char* solutionfile;
   unsigned int maxcolors;
   SCIP_Real timeLimit;
   SCIP_Real memLimit;
   SCIP_Longint nodeLimit;
   SCIP_Bool noSymmetry = FALSE;
   SCIP_Bool useorbital = FALSE;
   SCIP_Bool useorbitalfixing = FALSE;
   SCIP_Bool useisoprune = FALSE;
   SCIP_Bool usesymbreak = FALSE;
   SCIP_Bool uselpsymfix = FALSE;
   SCIP_Bool usebranchfirst = FALSE;
   SCIP_Bool forcesymmetrydetection = FALSE;
   SCIP_RETCODE retcode;
   SCIP_Real cutoffvalue = SCIP_INVALID;
   int dispFreq;
   int permseed;

   /* parse command line arguments */
   retcode = readArgumentsC(argc, argv, &filename, &maxcolors, &solutionfile, &settingsname,
      &timeLimit, &memLimit, &nodeLimit, &dispFreq, &noSymmetry, &permseed, &cutoffvalue);
   if ( retcode != SCIP_OKAY )
      exit(retcode);
   assert( filename != 0 );

   /* initialize SCIP */
   SCIP_CALL( SCIPcreate(&scip) );

   SCIPprintVersion(scip, 0);

   SCIPinfoMessage(scip, 0, "\n");
   SCIPinfoMessage(scip, 0, "Symmetry handling routines - (c) Marc Pfetsch, Thomas Rehn [svn: %s]\n", SVN_REV);
   SCIPinfoMessage(scip, 0, "Maximum k-colorable subgraph problem test.\n");
   SCIPinfoMessage(scip, 0, "\n");

   /* include default SCIP plugins */
   SCIP_CALL( SCIPincludeDefaultPlugins(scip) );

   /* add subtreebool constraint handler */
   SCIP_CALL( SCIPincludeConshdlrSubtreebool(scip) );

   /* add symmetries constraint handler */
   SCIP_CALL( SCIPincludeConshdlrSymmetries(scip) );

   /* include orbital branching */
   SCIP_CALL( SCIPincludeBranchruleOrbital(scip) );

   /* include isomorphism pruning */
   SCIP_CALL( SCIPincludePropIsoprune(scip) );

   /* include orbital fixing */
   SCIP_CALL( SCIPincludePropOrbitalfixing(scip) );

   /* include lpsymfix presolver */
   SCIP_CALL( SCIPincludePresolLpsymfix(scip) );

   /* include symmetry breaking presolver */
   SCIP_CALL( SCIPincludePresolSymbreak(scip) );


   /* --------------------------------------------------------------------- */
   /* turn off several SCIP-components: */

   /* gateextraction seems to modify the problem in a way that cannot be captured by our symmetry detection. */
   SCIP_CALL( SCIPsetIntParam(scip, "presolving/gateextraction/maxrounds", 0) );
   SCIPinfoMessage(scip, 0, "\nTurned off presolver <gateextraction>.\n");

   /* --------------------------------------------------------------------- */
   /* handle permutations */
   if ( permseed >= 0 )
   {
      SCIPsetIntParam(scip, "misc/permutationseed", permseed);
   }

   /* --------------------------------------------------------------------- */

   /* add our own parameters */
   SCIP_CALL( SCIPaddBoolParam(scip,
         "useorbital", "use orbital branching",
         &useorbital, TRUE, FALSE, 0, 0) );

   SCIP_CALL( SCIPaddBoolParam(scip,
         "useisoprune", "use isomorphism pruning",
         &useisoprune, TRUE, FALSE, 0, 0) );

   SCIP_CALL( SCIPaddBoolParam(scip,
         "useorbitalfixing", "use orbital fixing",
         &useorbitalfixing, TRUE, FALSE, 0, 0) );

   SCIP_CALL( SCIPaddBoolParam(scip,
         "usebranchfirst", "branch on first variable",
         &usebranchfirst, TRUE, FALSE, 0, 0) );

   SCIP_CALL( SCIPaddBoolParam(scip,
         "usesymbreak", "use symmetry breaking inequalities",
         &usesymbreak, TRUE, FALSE, 0, 0) );

   SCIP_CALL( SCIPaddBoolParam(scip,
         "uselpsymfix", "use projection of symmetric LP-variables onto the fixed space",
         &uselpsymfix, TRUE, FALSE, 0, 0) );

   SCIP_CALL( SCIPaddBoolParam(scip,
         "forcesymmetrydetection", "compute symmetry group even if it is not used",
         &forcesymmetrydetection, TRUE, TRUE, 0, 0) );

   /* set time, node, and memory limit */
   if ( ! SCIPisInfinity(scip, timeLimit) )
      SCIP_CALL( SCIPsetRealParam(scip, "limits/time", timeLimit) );
   if ( ! SCIPisInfinity(scip, memLimit) )
      SCIP_CALL( SCIPsetRealParam(scip, "limits/memory", memLimit) );
   if ( nodeLimit < SCIP_LONGINT_MAX )
      SCIP_CALL( SCIPsetLongintParam(scip, "limits/nodes", nodeLimit) );
   if ( dispFreq >= 0 )
      SCIP_CALL( SCIPsetIntParam(scip, "display/freq", dispFreq) );

   /* check for parameters */
   if ( settingsname != 0 )
   {
      if ( SCIPfileExists(settingsname) )
      {
         SCIPinfoMessage(scip, 0, "reading parameter file <%s> ...\n\n", settingsname);
         SCIP_CALL( SCIPreadParams(scip, settingsname) );
      }
      else
      {
         SCIPerrorMessage("parameter file <%s> not found - using default parameters.\n", settingsname);
      }
   }

   if ( usebranchfirst )
   {
      SCIP_CALL( SCIPincludeBranchruleFirst(scip) );
   }

   /* deactivate symmetry if requested */
   if ( noSymmetry )
   {
      SCIP_CALL( SCIPsetBoolParam(scip, "usesymbreak", FALSE) );
      SCIP_CALL( SCIPsetBoolParam(scip, "useorbital", FALSE) );
      SCIP_CALL( SCIPsetBoolParam(scip, "useisoprune", FALSE) );
   }

   /* output changed parameters */
   SCIPinfoMessage(scip, 0, "Changed settings:\n");
   SCIP_CALL( SCIPwriteParams(scip, 0, FALSE, TRUE) );
   SCIPinfoMessage(scip, 0, "\n");

   /* create Problem */
   SCIP_CALL( SCIPcreateProb(scip, "maxkcol", 0, 0, 0, 0, 0, 0, 0) );

   /* maximize */
   SCIP_CALL( SCIPsetObjsense(scip, SCIP_OBJSENSE_MAXIMIZE) );

   /* read solution */
   if ( solutionfile != 0 )
   {
      SCIP_CALL( SCIPtransformProb(scip) );
      SCIPinfoMessage(scip, 0, "reading solution <%s> ...\n\n", filename);
      SCIP_CALL( SCIPreadSol(scip, solutionfile) );
   }

   /* handle cutoff */
   if ( cutoffvalue != SCIP_INVALID )
   {
      SCIPinfoMessage(scip, 0, "\nSetting cutoff value to %g.\n\n", cutoffvalue);
      SCIP_CALL( SCIPsetObjlimit(scip, cutoffvalue) );
      SCIP_CALL( SCIPsetHeuristics(scip, SCIP_PARAMSETTING_OFF, TRUE) );
   }

   /* read graph */
   Graph G;
   SCIP_CALL( readGraph(scip, G, filename) );

   SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "\nGraph data:\n");
   SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Number of nodes:\t\t%d\n", boost::num_vertices(G));
   SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Number of edges:\t\t%d\n", boost::num_edges(G));

   /* set up problem */
   unsigned int n = boost::num_vertices(G);
   SCIP_VAR*** origxvars;
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &origxvars, n) );
   char name[SCIP_MAXSTRLEN];

   for (unsigned int i = 0; i < n; ++i)
   {
      SCIP_CALL( SCIPallocBlockMemoryArray(scip, &origxvars[i], maxcolors) );

      for (unsigned int j = 0; j < maxcolors; ++j)
      {
         (void) SCIPsnprintf(name, SCIP_MAXSTRLEN, "x%u#%u", i, j);
         SCIP_CALL( SCIPcreateVar(scip, &(origxvars[i][j]), name, 0.0, 1.0, 1.0,
               SCIP_VARTYPE_BINARY, TRUE, FALSE, 0, 0, 0, 0, 0) );
         SCIP_CALL( SCIPaddVar(scip, origxvars[i][j]) );
      }
   }

   /* add packing constraints */
   SCIP_CONS* cons;
   SCIP_VAR** vars;
   int size = MAX(maxcolors, 2);
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &vars, size) );

   for (unsigned int i = 0; i < n; ++i)
   {
      for (unsigned int j = 0; j < maxcolors; ++j)
         vars[j] = origxvars[i][j];

      (void) SCIPsnprintf(name, SCIP_MAXSTRLEN, "pack%u", i);
      SCIP_CALL( SCIPcreateConsSetpack(scip, &cons, name, (int) maxcolors, vars,
            TRUE, TRUE, TRUE, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE) );
      SCIP_CALL( SCIPaddCons(scip, cons) );
      SCIP_CALL( SCIPreleaseCons(scip, &cons) );
   }

   /* add edge constraints */
   EdgeIterator eit, eend;
   for (boost::tie(eit, eend) = boost::edges(G); eit != eend; ++eit)
   {
      Vertex source = boost::source(*eit, G);
      Vertex target = boost::target(*eit, G);

      for (unsigned int j = 0; j < maxcolors; ++j)
      {
         vars[0] = origxvars[source][j];
         vars[1] = origxvars[target][j];

         (void) SCIPsnprintf(name, SCIP_MAXSTRLEN, "edge%u#%u#%u", source, target, j);
         SCIP_CALL( SCIPcreateConsSetpack(scip, &cons, name, 2, vars,
               TRUE, TRUE, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE) );
         SCIP_CALL( SCIPaddCons(scip, cons) );
         SCIP_CALL( SCIPreleaseCons(scip, &cons) );
      }
   }

   if ( ! noSymmetry )
   {
      if ( forcesymmetrydetection )
      {
         /* add symmetry handling constraint */
         SCIP_CALL( SCIPcreateConsSymmetries(scip, &cons, "symmetries") );
         SCIP_CALL( SCIPaddCons(scip, cons) );
         SCIP_CALL( SCIPreleaseCons(scip, &cons) );

         SCIP_CONSHDLR* conshdlr = SCIPfindConshdlr(scip, "symmetries");
         if ( conshdlr == 0 )
         {
            SCIPerrorMessage("Could not find constraint handler <symmetries>.\n");
            return SCIP_ERROR;
         }
         (void) SYMsetProbSpecRequirement(conshdlr, SCIP_PROBSPEC_BINARY);
         (void) SYMsetProbSpecRequirement(conshdlr, SCIP_PROBSPEC_INTEGER);
         (void) SYMsetProbSpecRequirement(conshdlr, SCIP_PROBSPEC_REAL);
      }
      else
      {
         const unsigned int nperms = 2;
         const unsigned int npermvars = n * maxcolors;
         unsigned int* perms[2];
         unsigned int* basevars;
         unsigned int nbasevars;

         SCIP_VAR** permvars;
         SCIP_CALL( SCIPallocBlockMemoryArray(scip, &permvars, npermvars) );
         int c = 0;
         for (unsigned int i = 0; i < n; ++i)
         {
            for (unsigned int j = 0; j < maxcolors; ++j)
            {
               permvars[c] = origxvars[i][j];
               ++c;
            }
         }

         for (unsigned int i = 0; i < nperms; ++i)
         {
            SCIP_CALL( SCIPallocBlockMemoryArray(scip, &perms[i], npermvars) );
         }

         for (unsigned int i = 0; i < n; ++i)
         {
            for (unsigned int j = 0; j < maxcolors; ++j)
            {
               perms[0][i * maxcolors + j] = i * maxcolors + ( (j+1)%maxcolors );
            }
            perms[1][i * maxcolors] = i * maxcolors + 1;
            perms[1][i * maxcolors + 1] = i * maxcolors + 0;
            for (unsigned int j = 2; j < maxcolors; ++j)
            {
               perms[1][i * maxcolors + j] = i * maxcolors + j;
            }
         }

         /* ???????????????????? @todo check this! */
         nbasevars = maxcolors;
         SCIP_CALL( SCIPallocBlockMemoryArray(scip, &basevars, nbasevars) );
         for (unsigned int j = 0; j < maxcolors; ++j)
            basevars[j] = j;

         /* add symmetry handling constraint */
         SCIP_CALL( SCIPcreateConsSymmetriesPerms(scip, &cons, "symmetries", npermvars, permvars, nperms, perms, nbasevars, basevars) );
         SCIP_CALL( SCIPaddCons(scip, cons) );
         SCIP_CALL( SCIPreleaseCons(scip, &cons) );

         /* free memory */
         SCIPfreeBlockMemoryArray(scip, &basevars, nbasevars);
         SCIPfreeBlockMemoryArray(scip, &permvars, npermvars);
         for (unsigned int i = 0; i < nperms; ++i)
         {
            SCIPfreeBlockMemoryArray(scip, &perms[i], npermvars);
         }
      }
   }

   SCIPfreeBlockMemoryArray(scip, &vars, size);

#ifdef SCIP_DEBUG
   SCIP_CALL( SCIPwriteOrigProblem(scip, "maxkcol.cip", "cip", FALSE) );
#endif

   /* solve problem */
   SCIPinfoMessage(scip, 0, "\nsolving problem ...\n\n");
   SCIP_CALL( SCIPsolve(scip) );

   SCIP_CALL( SCIPprintStatistics(scip, 0) );
   // SCIP_CALL( SCIPprintSol(scip, SCIPgetBestSol(scip), 0, FALSE) );

   for (unsigned int i = 0; i < n; ++i)
   {
      for (unsigned int j = 0; j < maxcolors; ++j)
         SCIP_CALL( SCIPreleaseVar(scip, &origxvars[i][j]) );
      SCIPfreeBlockMemoryArray(scip, &origxvars[i], maxcolors);
   }
   SCIPfreeBlockMemoryArray(scip, &origxvars, n);

   /* free */
   SCIP_CALL( SCIPfree(&scip) );

   BMScheckEmptyMemory();

   return SCIP_OKAY;
}



/** main function */
int
main(
   int                   argc,
   char**                argv
   )
{
   SCIP_RETCODE retcode;

   retcode = runSCIP(argc, argv);
   if ( retcode != SCIP_OKAY )
   {
      SCIPprintError(retcode);
      return -1;
   }

   return 0;
}
