/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*    This file is part of "mipsymmetries" a collection of                   */
/*    symmetry handling methods.                                             */
/*                                                                           */
/*    Copyright (C) 2010-2015  Marc Pfetsch, Thomas Rehn                     */
/*                                                                           */
/*                                                                           */
/*    Based on SCIP  --- Solving Constraint Integer Programs                 */
/*       SCIP is distributed under the terms of the SCIP Academic Licence,   */
/*       see file COPYING in the SCIP distribution.                          */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
#pragma ident "@(#) $Id: orbital_fixing.cpp 1048 2015-10-18 12:06:36Z pfetsch $"

/**@file   orbital_fixing.cpp
 * @brief  methods to perform orbital fixing
 * @author Marc Pfetsch
 * @author Thomas Rehn
 *
 * We fix (binary) variables in orbits if certain variables are fixed to 0 or 1.
 *
 * The idea is described, for instance, in@par
 *  F. Margot: Symmetry in integer linear programming@n
 *  in: M. Juenger, T. Liebling, D. Naddef, G.L. Nemhauser, W. Pulleyblank, G. Reinelt, G. Rinaldi, L. Wolsey (Eds.),@n
 *  50 Years of Integer Programming 1958-2008, Springer-Verlag, Berlin, Heidelberg, 2010, pp. 647-681. (Chapter 17).
 */

/*---+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8----+----9----+----0----+----1----+----2*/

#include "orbital_fixing.h"

#include <list>
#include <set>

#include <boost/scoped_ptr.hpp>

using namespace permlib;

/* useful defines */
typedef AbstractPermutationGroup::OrbitList OrbitList;
typedef boost::scoped_ptr<OrbitList> OrbitListPtr;
typedef std::set<permlib::dom_int>::const_iterator OrbitSetIt;
typedef OrbitList::const_iterator OrbIt;


/** perform orbital fixing
 *
 *  Note that we do not have to distinguish between variables that have been fixed or branched to 1, since the
 *  stabilizer is with respect to the variables that have been branched to 1. Thus, if an orbit contains a variable that
 *  has been branched to 1, the whole orbit only contains variables that have been branched to 1 - and nothing can be
 *  fixed.
 */
SCIP_RETCODE orbitalFixing(
   SCIP*                 scip,               /**< SCIP pointer */
   SCIP_VAR**            vars,               /**< variables */
   int                   nvars,              /**< number of variables */
   const PermGroupPtr&   group,              /**< local symmetry group or stabilizer of binary variables branched to 1 */
   SCIP_Bool&            infeasible,         /**< whether we detected that the problem is infeasible */
   unsigned int&         nfixedone,          /**< number of variables fixed to 1 */
   unsigned int&         ngen                /**< number of domain propagations performed */
   )
{
   SCIP_Bool tightened;

   ngen = 0;
   nfixedone = 0;

   /* get orbits of group */
   const OrbitListPtr childOrbits(group->orbits());

   SCIPdebugMessage("Perform orbital fixing on %lu orbits.\n", childOrbits->size());

   /* check all orbits */
   for (OrbIt it = childOrbits->begin(); it != childOrbits->end(); ++it)
   {
      /* skip orbits that are too small */
      if ( it->size() < 2 )
         continue;

      /* check all variables in the orbit */
      bool havefixedone = false;
      bool havefixedzero = false;
      OrbitSetIt oit, oend = it->end();
      for (oit = it->begin(); oit != oend; ++oit)
      {
         const unsigned int idx = *oit;
         assert( (int) idx < nvars );
         SCIP_VAR* var = vars[idx];
         assert( var != 0 );

         /* check whether variable is not binary (and not implicit integer!) */
         if ( SCIPvarGetType(var) != SCIP_VARTYPE_BINARY )
         {
            /* skip orbit if there are non-binary variables */
            havefixedone = false;
            havefixedzero = false;
            break;
         }

         /* if variable is fixed to 1 -> can fix all variables in orbit to 1, see paper for an explanation. */
         if ( SCIPvarGetLbLocal(var) > 0.5 )
            havefixedone = true;

         /* check for zero-fixed variables */
         if ( SCIPvarGetUbLocal(var) < 0.5 )
            havefixedzero = true;
      }

      /* check consistency */
      if ( havefixedone && havefixedzero )
      {
         infeasible = TRUE;
         return SCIP_OKAY;
      }

      /* fix all variables to 0 if there is one variable fixed to 0 */
      if ( havefixedzero )
      {
         assert( ! havefixedone );
         assert( oend == it->end() );
         for (oit = it->begin(); oit != oend; ++oit)
         {
            const unsigned int idx = *oit;
            assert( (int) idx < nvars );
            SCIP_VAR* var = vars[idx];

            /* only variables that are not yet fixed to 0 */
            if ( SCIPvarGetUbLocal(var) > 0.5 )
            {
               SCIPdebugMessage("can fix <%s> (index %u) to 0.\n", SCIPvarGetName(var), idx);
               assert( SCIPvarGetType(var) == SCIP_VARTYPE_BINARY );
               /* due to aggregation, var might already be fixed to 1, so do not put assert here */

               /* do not use SCIPinferBinvarProp(), since conflict analysis is not valid */
               SCIP_CALL( SCIPtightenVarUb(scip, var, 0.0, FALSE, &infeasible, &tightened) );
               if ( infeasible )
                  return SCIP_OKAY;
               if ( tightened )
                  ++ngen;
            }
         }
      }

      /* fix all variables to 1 if there is one variable fixed to 1 */
      if ( havefixedone )
      {
         assert( ! havefixedzero );
         assert( oend == it->end() );
         for (oit = it->begin(); oit != oend; ++oit)
         {
            const unsigned int idx = *oit;
            assert( (int) idx < nvars );
            SCIP_VAR* var = vars[idx];

            /* only variables that are not yet fixed to 1 */
            if ( SCIPvarGetLbLocal(var) < 0.5)
            {
               SCIPdebugMessage("can fix <%s> (index %u) to 1.\n", SCIPvarGetName(vars[idx]), idx);
               assert( SCIPvarGetType(var) == SCIP_VARTYPE_BINARY );
               /* due to aggregation, var might already be fixed to 0, so do not put assert here */

               /* do not use SCIPinferBinvarProp(), since conflict analysis is not valid */
               SCIP_CALL( SCIPtightenVarLb(scip, var, 1.0, FALSE, &infeasible, &tightened) );
               if ( infeasible )
                  return SCIP_OKAY;
               if ( tightened )
               {
                  ++ngen;
                  ++nfixedone;
               }
            }
         }
      }
   }

   return SCIP_OKAY;
}
