/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*    This file is part of "mipsymmetries" a collection of                   */
/*    symmetry handling methods.                                             */
/*                                                                           */
/*    Copyright (C) 2010-2015  Marc Pfetsch, Thomas Rehn                     */
/*                                                                           */
/*                                                                           */
/*    Based on SCIP  --- Solving Constraint Integer Programs                 */
/*       SCIP is distributed under the terms of the SCIP Academic Licence,   */
/*       see file COPYING in the SCIP distribution.                          */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
#pragma ident "@(#) $Id: presol_lpsymfix.cpp 124 2011-10-19 18:19:40Z pfetsch $"

/**@file   presol_lpsymfix.cpp
 * @ingroup PRESOLVERS
 * @brief  presolver for projecting symmetric LP-variable onto the fixed space
 * @author Marc Pfetsch
 * @author Thomas Rehn
 */

/*---+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8----+----9----+----0----+----1----+----2*/

#include "presol_lpsymfix.h"
#include "cons_symmetries.h"

#include <vector>
#include <list>
#include <set>

/* presolver properties */
#define PRESOL_NAME            "lpsymfix"
#define PRESOL_DESC            "presolver for projecting symmetric LP-variables onto the fixed space"
#define PRESOL_PRIORITY         -10000000    /**< priority of the presolver (>= 0: before, < 0: after constraint handlers) */
#define PRESOL_MAXROUNDS               -1    /**< maximal number of presolving rounds the presolver participates in (-1: no limit) */
#define PRESOL_DELAY                FALSE    /**< should presolver be delayed, if other presolvers found reductions? */
#define PRESOL_TIMING   SCIP_PRESOLTIMING_EXHAUSTIVE   /**< timing for presolving */

using namespace permlib;


/* useful typedefs */
typedef AbstractPermutationGroup::OrbitList OrbitList;
typedef std::set<permlib::dom_int>::const_iterator OrbitSetIt;
typedef OrbitList::const_iterator OrbIt;


/*
 * Data structures
 */

/** presolver data */
struct SCIP_PresolData
{
   PermutationGroupPtr   group;              /**< symmetry group */
   unsigned int          nvars;              /**< number of variables */
   SCIP_VAR**            vars;               /**< variables in constraint */
   SCIP_HASHMAP*         varmap;             /**< map of variables to indices in vars array */
   SCIP_Bool             enabled;            /**< run presolver? */
   SCIP_CONSHDLR*        symmetriesconshdlr; /**< pointer to cons_symmetries constraint handler */
};



/*
 * Callback methods of presolver
 */

/** destructor of presolver to free user data (called when SCIP is exiting) */
static
SCIP_DECL_PRESOLFREE(presolFreeLpsymfix)
{  /*lint --e{715}*/
   SCIP_PRESOLDATA* presoldata;

   assert( scip != 0 );
   assert( presol != 0 );

   presoldata = SCIPpresolGetData(presol);
   assert( presoldata != 0 );

   presoldata->group.reset();
   SCIPfreeMemory(scip, &presoldata);

   return SCIP_OKAY;
}


/** initialization method of presolver (called after problem was transformed) */
SCIP_DECL_PRESOLINIT(presolInitLpsymfix)
{
   SCIP_PRESOLDATA* presoldata;

   assert( scip != 0 );
   assert( presol != 0 );

   SCIPdebugMessage("Initpre method of symmetric lp projection presolver ...\n");

   presoldata = SCIPpresolGetData(presol);
   assert( presoldata != 0 );

   /* check whether we should run */
   SCIP_CALL( SCIPgetBoolParam(scip, "uselpsymfix", &presoldata->enabled) );

   if ( presoldata->enabled )
   {
      /* we only want symmetries of real variables, all other types must be fixed */
      SYMsetProbSpecRequirement(presoldata->symmetriesconshdlr, SCIP_PROBSPEC_REAL);

      SYMsetProbSpecRequirementFixed(presoldata->symmetriesconshdlr, SCIP_PROBSPEC_BINARY);
      SYMsetProbSpecRequirementFixed(presoldata->symmetriesconshdlr, SCIP_PROBSPEC_INTEGER);
   }

   return SCIP_OKAY;
}


/** execution method of presolver */
static
SCIP_DECL_PRESOLEXEC(presolExecLpsymfix)
{  /*lint --e{715,818}*/
   SCIP_PRESOLDATA* presoldata;
   int oldnaggrvars;

   assert( scip != 0 );
   assert( presol != 0 );
   assert( result != 0 );

   *result = SCIP_DIDNOTRUN;

   presoldata = SCIPpresolGetData(presol);
   assert( presoldata != 0 );

   /* possibly skip presolver */
   if ( ! presoldata->enabled )
      return SCIP_OKAY;

   /* skip presolving if we are not at the end */
   if ( ! SCIPisPresolveFinished(scip) )
      return SCIP_OKAY;

   /* possibly stop */
   if ( SCIPisStopped(scip) )
      return SCIP_OKAY;

   SCIPdebugMessage("Presolving method of symmetric lp projection presolver ...\n");

   /* get symmetry if not already present */
   if ( ! presoldata->group )
   {
      SCIPdebugMessage("Get symmetry information ...\n");

      /* get symmetry information */
      uintvecPtr dummy(new uintvec());
      SCIP_CALL( SYMgetSymmetryData(scip, presoldata->symmetriesconshdlr, presoldata->group, presoldata->nvars, presoldata->vars, dummy, presoldata->varmap) );
      if (! presoldata->group)
         presoldata->enabled = FALSE;        /* to avoid later calls */
   }

   *result = SCIP_DIDNOTFIND;
   oldnaggrvars = *naggrvars;

   /* possibly stop */
   if ( SCIPisStopped(scip) )
      return SCIP_OKAY;

   /* collect continuous variables */
   uintvec realvars;
   for (unsigned int j = 0; j < presoldata->nvars; ++j)
   {
      SCIP_VAR* var = presoldata->vars[j];
      assert( var != 0 );

      if ( SCIPvarGetType(var) == SCIP_VARTYPE_CONTINUOUS || SCIPvarGetType(var) == SCIP_VARTYPE_IMPLINT )
         realvars.push_back(j);
   }

   /* exit if there are not enough conitinuous variables to project */
   if ( realvars.size() <= 1 )
      return SCIP_OKAY;

   /* We must not project onto the fixed space if real variables are moved by a permutation that also moves a discrete
    * variable. Our 'SYMsetProbSpecRequirementFixed' should make sure that all symmetries fix discrete variables */

   /* get all orbits for continuous variables */
   boost::scoped_ptr<OrbitList> orbits(presoldata->group->orbits(realvars));
   SCIPdebugMessage("Number of orbits: %lu.\n", (unsigned long) orbits->size());

   /* loop through orbits */
   for (OrbIt it = orbits->begin(); it != orbits->end(); ++it)
   {
      /* skip singletons */
      if ( it->size() <= 1 )
         continue;

      /* aggregate all variables in orbit */
      SCIP_Bool infeasible;
      SCIP_Bool redundant;
      SCIP_Bool aggregated;

      OrbitSetIt oit = it->begin();
      OrbitSetIt oend = it->end();

      /* store first variable and skip to second variable */
      unsigned int firstidx = *oit++;
      assert( firstidx < presoldata->nvars );
      SCIP_VAR* firstvar = presoldata->vars[firstidx];
      assert( SCIPvarGetType(firstvar) == SCIP_VARTYPE_CONTINUOUS || SCIPvarGetType(firstvar) == SCIP_VARTYPE_IMPLINT );
#ifndef NDEBUG
      SCIP_Real obj = 0.0;
      SCIP_CALL( SCIPvarGetAggregatedObj(firstvar, &obj) );
#endif

      /* loop through other variables in orbit */
      for (; oit != oend; ++oit)
      {
         unsigned int varidx = *oit;
         assert( varidx < presoldata->nvars );
         SCIP_VAR* var = presoldata->vars[varidx];

         /* all variables in orbit should be continuous (or implicit continuous) and have the same objective */
         assert( SCIPvarGetType(var) == SCIP_VARTYPE_CONTINUOUS || SCIPvarGetType(var) == SCIP_VARTYPE_IMPLINT );
#ifndef NDEBUG
         SCIP_Real obj2 = 0.0;
         SCIP_CALL( SCIPvarGetAggregatedObj(var, &obj2) );
#endif

         SCIPdebugMessage("Aggregate: <%s> and <%s>.\n", SCIPvarGetName(firstvar), SCIPvarGetName(var));

         SCIP_CALL( SCIPaggregateVars(scip, firstvar, var, 1.0, -1.0, 0.0, &infeasible, &redundant, &aggregated) );

         if ( infeasible )
         {
            *result = SCIP_CUTOFF;
            return SCIP_OKAY;
         }

         if ( aggregated )
         {
            ++(*naggrvars);
            *result = SCIP_SUCCESS;
         }
      }
   }
   presoldata->enabled = FALSE;

   SCIPdebugMessage("Aggregated variables: %d.\n", *naggrvars - oldnaggrvars);

   return SCIP_OKAY;
}





/*
 * presolver specific interface methods
 */

/** creates the symmetry LP-fix presolver and includes it in SCIP */
SCIP_RETCODE SCIPincludePresolLpsymfix(
   SCIP*                 scip                /**< SCIP data structure */
   )
{
   SCIP_PRESOLDATA* presoldata = 0;

   /* create presolver data */
   SCIP_CALL( SCIPallocMemory(scip, &presoldata) );

   /* we must call the constructor explictly, because memory was m'alloced and not new'ed */
   new (&presoldata->group) PermutationGroupPtr();
   presoldata->vars = 0;
   presoldata->nvars = 0;
   presoldata->varmap = 0;

   /* determine cons_symmetries constraint handler */
   SCIP_CONSHDLR* conshdlr = SCIPfindConshdlr(scip, "symmetries");
   if ( conshdlr == 0 )
   {
      SCIPerrorMessage("Could not find constraint handler <symmetries>.\n");
      return SCIP_ERROR;
   }
   presoldata->symmetriesconshdlr = conshdlr;

   /* include presolver */
   SCIP_PRESOL* presol = 0;
#if SCIP_VERSION >= 320
   SCIP_CALL( SCIPincludePresolBasic(scip, &presol, PRESOL_NAME, PRESOL_DESC, PRESOL_PRIORITY, PRESOL_MAXROUNDS, PRESOL_TIMING,
         presolExecLpsymfix, presoldata) );
#else
   SCIP_CALL( SCIPincludePresolBasic(scip, &presol, PRESOL_NAME, PRESOL_DESC, PRESOL_PRIORITY, PRESOL_MAXROUNDS, PRESOL_DELAY, presolExecLpsymfix, presoldata) );
#endif

   /* set additional callbacks */
   SCIP_CALL( SCIPsetPresolFree(scip, presol, presolFreeLpsymfix) );
   SCIP_CALL( SCIPsetPresolInit(scip, presol, presolInitLpsymfix) );

   return SCIP_OKAY;
}

/* vim: set tabstop=3 expandtab: */
// kate: space-indent on; indent-width 3; mixedindent off; indent-mode cstyle;
