/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*    This file is part of "mipsymmetries" a collection of                   */
/*    symmetry handling methods.                                             */
/*                                                                           */
/*    Copyright (C) 2010-2015  Marc Pfetsch, Thomas Rehn                     */
/*                                                                           */
/*                                                                           */
/*    Based on SCIP  --- Solving Constraint Integer Programs                 */
/*       SCIP is distributed under the terms of the SCIP Academic Licence,   */
/*       see file COPYING in the SCIP distribution.                          */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
#pragma ident "@(#) $Id: presol_orbitprobing.cpp 124 2011-10-19 18:19:40Z pfetsch $"

/**@file   presol_orbitprobing.cpp
 * @ingroup PRESOLVERS
 * @brief  presolver for orbit probing
 * @author Marc Pfetsch
 * @author Thomas Rehn
 *
 * This presolver performs probing on orbits as suggested by orbital branching. That is, it tentatively fixes all
 * variables in an orbit to 0. If the resulting problem is infeasible, it may fix the first variable in the orbit to
 * 1. Conversely, it tentatively fixes the first variable in an orbit to 1. If this is infeasible, it may fix all
 * variables in the orbit to 0. This represents the branches of orbital branching.
 *
 * @todo Branch all variables in an orbit to 1 and fix the first to 0, if this get infeasible. This is an idea mentioned
 * in a talk by Tobias Achterberg at ZIB in 2013.
 */

/*---+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8----+----9----+----0----+----1----+----2*/

#include "presol_orbitprobing.h"
#include "cons_symmetries.h"

#include <cassert>
#include <list>
#include <set>


/* presolver properties */
#define PRESOL_NAME            "orbitprobing"
#define PRESOL_DESC            "presolver for orbit probing"
#define PRESOL_PRIORITY            -50000    /**< priority of the presolver (>= 0: before, < 0: after constraint handlers) */
#define PRESOL_MAXROUNDS               -1    /**< maximal number of presolving rounds the presolver participates in (-1: no limit) */
#define PRESOL_DELAY                 TRUE    /**< should presolver be delayed, if other presolvers found reductions? */
#define PRESOL_TIMING   SCIP_PRESOLTIMING_EXHAUSTIVE   /**< timing for presolving */

/* default parameters */
#define DEFAULT_USEPROBEONE          TRUE    /**< Should orbit probing on the one branch be performed? */
#define DEFAULT_MARKALL              TRUE    /**< Should each variable of a checked orbit me marked not to be tried again as first variable? */

/* other parameters */
#define MAXPROBINGPROPROUNDS          100    /**< maximal number of propagation rounds in probing */
#define MAXDNOM                   10000LL    /**< maximal denominator for simple rational fixed values */


/* useful typedefs */
typedef permlib::AbstractPermutationGroup::OrbitList OrbitList;
typedef std::set<permlib::dom_int>::const_iterator OrbitSetIt;
typedef OrbitList::const_iterator OrbitIt;

using namespace permlib;

/*
 * Data structures
 */

/** presolver data */
struct SCIP_PresolData
{
   PermutationGroupPtr   group;              /**< symmetry group */
   unsigned int          nvars;              /**< number of variables used for computing symmetry */
   SCIP_VAR**            vars;               /**< variables used for computing symmetry */
   uintvecPtr            binaryvars;         /**< indices of binary variables to perform orbital branching on */
   uintvec               varsfixedzero;      /**< variables fixed to 0 by orbit probing */
   uintvec               varsfixedone;       /**< variables fixed to 1 by orbit probing */
   SCIP_HASHMAP*         varmap;             /**< map of variables to indices in vars array */
   SCIP_Bool             performedprobing;   /**< whether we already performed probing */
   SCIP_Bool             useprobeone;        /**< Should orbit probing on the one branch be performed? */
   SCIP_Bool             markall;            /**< Should each variable of a checked orbit me marked not to be tried again as first variable? */
   SCIP_Bool             enabled;            /**< run presolver? */
   SCIP_Bool             early;              /**< run presolver as early as possible if symmetry has been detected in initpre() */
   SCIP_CONSHDLR*        symmetriesconshdlr; /**< pointer to cons_symmetries constraint handler */
};


/*
 * Local methods
 */

/** applies and evaluates probing of a single orbit */
static
SCIP_RETCODE applyProbingOrbit(
   SCIP*                 scip,               /**< SCIP data structure */
   SCIP_Bool             probzero,           /**< Should probing on the zero branch be performed? One-branch otherwise. */
   int                   proprounds,         /**< number of propagation rounds */
   SCIP_VAR*             onevar,             /**< variable to be branched to 1 if probzero is false */
   SCIP_VAR**            vars,               /**< problem variables */
   int                   nvars,              /**< number of problem variables */
   SCIP_Bool*            checkedvar,         /**< store whether probing on these variables has partly been performed (updated here) */
   SCIP_Bool             markall,            /**< Should each variable of a checked orbit me marked not to be tried again as first variable? */
   OrbitIt&              orbitit,            /**< orbit to perform probing on */
   SCIP_Real*            proplbs,            /**< array to store lower bounds after full propagation */
   SCIP_Real*            propubs,            /**< array to store upper bounds after full propagation */
   SCIP_Bool&            skiporbit,          /**< whether we should skip the current orbit */
   SCIP_Bool&            cutoff              /**< pointer to store whether the probing direction is infeasible */
   )
{
   assert( proplbs != NULL );
   assert( propubs != NULL );
   assert( onevar != NULL );

   SCIPdebugMessage("Applying probing on orbit of size %lu (%d-branch) ...\n", orbitit->size(), 1 - probzero);
   skiporbit = FALSE;
   cutoff = FALSE;

   /* start probing mode */
   SCIP_CALL( SCIPstartProbing(scip) );

#ifndef NDEBUG
   /* make sure that variables in orbit have the same signature */
   SCIP_Real firstobj = SCIPvarGetObj(onevar);
   SCIP_Real firstlb = SCIPvarGetLbLocal(onevar);
   SCIP_Real firstub = SCIPvarGetUbLocal(onevar);
   SCIP_VARTYPE firsttype = SCIPvarGetType(onevar);
   SCIP_VARSTATUS firststatus = SCIPvarGetStatus(onevar);
   int firstlocksdown = SCIPvarGetNLocksDown(onevar);
   int firstlocksup = SCIPvarGetNLocksUp(onevar);
#endif

   /* prepare probing */
   OrbitSetIt oend = orbitit->end();
   for (OrbitSetIt oit = orbitit->begin(); oit != oend; ++oit)
   {
      SCIP_VAR* var;

      assert( *oit < (unsigned int) nvars );
      if ( markall )
         checkedvar[*oit] = TRUE;

      var = vars[*oit];
      assert( SCIPvarIsBinary(var) );

      /* check for fixed variables */
      if ( SCIPvarGetLbLocal(var) > 0.5 || SCIPvarGetUbLocal(var) < 0.5 )
      {
         SCIPdebugMessage("Variable <%s> in the orbit is already fixed to %f - skip orbit.\n", SCIPvarGetName(var), SCIPvarGetLbLocal(var));
         skiporbit = TRUE;
         SCIP_CALL( SCIPendProbing(scip) );
         return SCIP_OKAY;
      }

#ifndef NDEBUG
      assert( SCIPisEQ(scip, firstobj, SCIPvarGetObj(var)) );
      assert( SCIPisEQ(scip, firstlb, SCIPvarGetLbLocal(var)) );
      assert( SCIPisEQ(scip, firstub, SCIPvarGetUbLocal(var)) );
      assert( firsttype == SCIPvarGetType(var) );
      assert( firststatus == SCIPvarGetStatus(var) );
      assert( firstlocksdown == SCIPvarGetNLocksDown(var) );
      assert( firstlocksup == SCIPvarGetNLocksUp(var) );
#endif

      if ( probzero )
      {
         SCIP_CALL( SCIPchgVarUbProbing(scip, var, 0.0) );
      }
   }

   /* probe on 1-branch */
   if ( ! probzero )
   {
      SCIP_CALL( SCIPchgVarLbProbing(scip, onevar, 1.0) );
   }

   /* apply propagation of implication graph and clique table */
   SCIP_CALL( SCIPpropagateProbingImplications(scip, &cutoff) );

   /* apply propagation */
   if ( ! cutoff )
   {
      SCIP_CALL( SCIPpropagateProbing(scip, proprounds, &cutoff, NULL) );
   }

   /* evaluate propagation */
   if ( ! cutoff )
   {
      for (int i = 0; i < nvars; ++i)
      {
         proplbs[i] = SCIPvarGetLbLocal(vars[i]);
         propubs[i] = SCIPvarGetUbLocal(vars[i]);
      }
   }

   /* exit probing mode */
   SCIP_CALL( SCIPendProbing(scip) );

   return SCIP_OKAY;
}


/** analyses boundchanges resulting from probing on an orbit and performs deduced fixations, aggregations, and domain tightenings
 *
 *  This function is modified from SCIPanalyzeDeductionsProbing() in prop_probing.c.
 */
SCIP_RETCODE analyzeDeductionsProbing(
   SCIP*                 scip,               /**< SCIP data structure */
   SCIP_PRESOLDATA*      presoldata,         /**< presolver data */
   const OrbitIt&        orbitit,            /**< orbit to branch on */
   int                   nvars,              /**< number of variables which bound changes should be analyzed */
   SCIP_VAR**            vars,               /**< variables which bound changes should be analyzed */
   const SCIP_Real*      leftproplbs,        /**< lower bounds after applying domain propagation in left branch */
   const SCIP_Real*      leftpropubs,        /**< upper bounds after applying domain propagation in left branch */
   const SCIP_Real*      rightproplbs,       /**< lower bounds after applying domain propagation in right branch */
   const SCIP_Real*      rightpropubs,       /**< upper bounds after applying domain propagation in right branch */
   int*                  nfixedvars,         /**< pointer to counter which is increased by the number of deduced variable fixations */
   int*                  nchgbds,            /**< pointer to counter which is increased by the number of deduced bound tightenings */
   SCIP_Bool&            cutoff              /**< buffer to store whether a cutoff is detected */
   )
{
   int j;

   assert( scip != NULL );
   assert( vars != NULL || nvars == 0 );
   assert( leftproplbs != NULL );
   assert( leftpropubs != NULL );
   assert( rightproplbs != NULL );
   assert( rightpropubs != NULL );
   assert( nfixedvars != NULL );
   assert( nchgbds != NULL );

   cutoff = FALSE;

   for (j = 0; j < nvars && !cutoff; ++j)
   {
      SCIP_VAR* var;
      SCIP_Bool varisinteger;
      SCIP_Real oldlb;
      SCIP_Real oldub;
      SCIP_Real newlb;
      SCIP_Real newub;

      assert( vars != NULL );
      var = vars[j];
      assert( var != NULL );

      /* skip variables in orbit */
      if ( orbitit->find(j) != orbitit->end() )
         continue;

      /* skip already fixed variables */
      oldlb = SCIPvarGetLbLocal(var);
      oldub = SCIPvarGetUbLocal(var);
      if ( SCIPisEQ(scip, oldlb, oldub) )
         continue;

      /* new bounds of the variable is the union of the propagated bounds of the left and right case */
      newlb = MIN(leftproplbs[j], rightproplbs[j]);
      newub = MAX(leftpropubs[j], rightpropubs[j]);
      varisinteger = (SCIPvarGetType(var) < SCIP_VARTYPE_CONTINUOUS);

      /* check for fixed variables */
      if ( SCIPisEQ(scip, newlb, newub) )
      {
         SCIP_Real fixval;
         SCIP_Bool fixed = FALSE;

         if ( ! varisinteger )
         {
            /* in both probings, variable j is deduced to the same value: fix variable to this value */
            fixval = SCIPselectSimpleValue(newlb - 0.9 * SCIPepsilon(scip), newub + 0.9 * SCIPepsilon(scip), MAXDNOM);
         }
         else
         {
            fixval = newlb;
         }
         SCIPdebugMessage("Orbit (%lu): fixing <%s> to %f ([%f, %f], type: %d, obj: %f).\n",
            orbitit->size(), SCIPvarGetName(var), fixval, oldlb, oldub, SCIPvarGetType(var), SCIPvarGetObj(var));

         SCIP_CALL( SCIPfixVar(scip, var, fixval, &cutoff, &fixed) );
         if ( fixed )
         {
            ++(*nfixedvars);
            if ( newlb > 0.5 )
               presoldata->varsfixedone.push_back(j);
            else
               presoldata->varsfixedzero.push_back(j);
         }
         continue;
      }
      else
      {
         /* check for bound tightenings */
         SCIP_Bool tightened;
         SCIP_Bool tightenlb;
         SCIP_Bool tightenub;
         SCIP_Bool force;

         if ( varisinteger )
         {
            force = TRUE;
            tightenlb = (newlb > oldlb + 0.5);
            tightenub = (newub < oldub - 0.5);
         }
         else
         {
            force = TRUE;
            tightenlb = SCIPisLbBetter(scip, newlb, oldlb, oldub);
            tightenub = SCIPisUbBetter(scip, newub, oldlb, oldub);
         }

         if ( tightenlb )
         {
            /* in both probings, variable j is deduced to be at least newlb: tighten lower bound */
            SCIP_CALL( SCIPtightenVarLb(scip, var, newlb, force, &cutoff, &tightened) );
            if ( tightened )
            {
               SCIPdebugMessage("tightened lower bound of variable <%s>[%g,%g] to %g due to probing on orbit of size %lu\n",
                  SCIPvarGetName(var), oldlb, oldub, newlb, orbitit->size());
               ++(*nchgbds);
            }
         }

         if ( tightenub && ! cutoff )
         {
            /* in both probings, variable j is deduced to be at most newub: tighten upper bound */
            SCIP_CALL( SCIPtightenVarUb(scip, var, newub, force, &cutoff, &tightened) );
            if ( tightened )
            {
               SCIPdebugMessage("tightened upper bound of variable <%s>[%g,%g] to %g due to probing on orbit of size %lu\n",
                  SCIPvarGetName(var), oldlb, oldub, newub, orbitit->size());
               ++(*nchgbds);
            }
         }
         if ( cutoff )
            break;
      }
   }

   return SCIP_OKAY;
}


/** apply probing on orbits */
static
SCIP_RETCODE applyProbing(
   SCIP*                 scip,               /**< SCIP data structure */
   SCIP_PRESOLDATA*      presoldata,         /**< presolver data */
   int*                  nfixedvars,         /**< pointer to store number of fixed variables */
   int*                  nchgbds,            /**< pointer to store number of changed bounds */
   SCIP_Bool&            cutoff              /**< pointer to store whether cutoff occured */
   )
{
   SCIP_Real* zeroproplbs;
   SCIP_Real* zeropropubs;
   SCIP_Real* oneproplbs;
   SCIP_Real* onepropubs;
   SCIP_Bool* checkedvar;
   SCIP_Bool fixedone = FALSE;
   SCIP_Bool fixedvar = FALSE;
   int nrounds = 0;

   assert( scip != 0 );
   assert( presoldata != 0 );

   cutoff = FALSE;

   PermutationGroupPtr group = presoldata->group;
   assert( group );
   uintvecPtr binaryvars = presoldata->binaryvars;

   /* skip probing if there are not enough binary variables */
   if ( binaryvars->size() < 2 )
      return SCIP_OKAY;

   assert( presoldata->nvars > 0 );
   assert( presoldata->vars != 0 );
   int nvars = presoldata->nvars;
   SCIP_VAR** vars = presoldata->vars;

   /* get temporary memory for storing probing results */
   SCIP_CALL( SCIPallocBufferArray(scip, &zeroproplbs, nvars) );
   SCIP_CALL( SCIPallocBufferArray(scip, &zeropropubs, nvars) );
   SCIP_CALL( SCIPallocBufferArray(scip, &oneproplbs, nvars) );
   SCIP_CALL( SCIPallocBufferArray(scip, &onepropubs, nvars) );
   SCIP_CALL( SCIPallocBufferArray(scip, &checkedvar, nvars) );

   /* mark all variables to be unmarked */
   BMSclearMemoryArray(checkedvar, nvars);

   /* we need to repeat computations with stabilizers if we fixed a variable */
   PermutationGroupPtr stab = group;
   uintvec fixedvarsone;
   uintvec fixedvarszero;

   /* repeat until no fixation took place */
   do
   {
      fixedone = FALSE;
      fixedvar = FALSE;

      fixedvarsone.clear();
      fixedvarszero.clear();
      for (int i = 0; i < nvars; ++i)
      {
         SCIP_VAR* var = vars[i];
         assert( var != NULL );
         if ( SCIPvarIsBinary(var) )
         {
            if ( SCIPvarGetLbLocal(var) > 0.5 )
               fixedvarsone.push_back(i);
            if ( SCIPvarGetUbLocal(var) < 0.5 )
               fixedvarszero.push_back(i);
         }
      }

      try
      {
         stab = PermutationGroupPtr(group->setStabilizer(fixedvarsone));
         stab = PermutationGroupPtr(stab->setStabilizer(fixedvarszero));
      }
      catch (std::bad_alloc& ba)
      {
         SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Aborted orbit probing because we went out of memory.\n");
         /* try to catch bad memory allocation */
         break;
      }

      /* get all orbits */
      boost::scoped_ptr<OrbitList> O(stab->orbits(*binaryvars));

      /* if the number of orbits is so large that all are trivial */
      if ( O->size() >= binaryvars->size() )
         break;

      SCIPdebugMessage("Round: %d - total number of orbits: %lu (fixed to 0: %lu, fixed to 1: %lu).\n",
         nrounds, (unsigned long) O->size(), fixedvarszero.size(), fixedvarsone.size());

      /* loop through orbits */
      OrbitIt iend = O->end();
      for (OrbitIt it = O->begin(); it != iend && ! SCIPisStopped(scip) && ! cutoff; ++it)
      {
         SCIP_Bool localcutoff = FALSE;
         SCIP_Bool skiporbit = FALSE;
         SCIP_VAR* onevar = NULL;

         /* skip trivial orbits */
         if ( it->size() < 2 )
            continue;

         /* take first variable and check whether we have performed probing on it */
         unsigned int firstidx = *(it->begin());
         assert( (int) firstidx < nvars );
         if ( checkedvar[firstidx] )
            continue;

         /* always mark first variable */
         checkedvar[firstidx] = TRUE;

         onevar = vars[firstidx];
         assert( SCIPvarIsBinary(onevar) );

         /* apply probing on the zero branch */
         SCIP_CALL( applyProbingOrbit(scip, TRUE, MAXPROBINGPROPROUNDS, onevar, vars, nvars, checkedvar, presoldata->markall,
               it, zeroproplbs, zeropropubs, skiporbit, localcutoff) );
         if ( skiporbit )
            continue;

         if ( localcutoff )
         {
            /* the variable can be fixed to 1 */
            SCIP_CALL( SCIPfixVar(scip, onevar, 1.0, &cutoff, &fixedone) );
            assert( fixedone );
            if ( cutoff )
               break;

            if ( fixedone )
            {
               SCIPdebugMessage("Fixed first variable <%s> in orbit to 1.0.\n", SCIPvarGetName(onevar));
               ++(*nfixedvars);
               presoldata->varsfixedone.push_back(firstidx);
               fixedvarsone.push_back(firstidx);
               fixedvar = TRUE;
               break;
            }
         }

         /* apply probing on the 1-branch */
         SCIP_CALL( applyProbingOrbit(scip, FALSE, MAXPROBINGPROPROUNDS, onevar, vars, nvars, checkedvar, presoldata->markall,
               it, oneproplbs, onepropubs, skiporbit, localcutoff) );
         if ( skiporbit )
            continue;

         if ( localcutoff )
         {
            SCIP_Bool fixed = FALSE;

            /* the orbit can be fixed to 0 */
            OrbitSetIt oit, oend = it->end();
            for (oit = it->begin(); oit != oend && ! cutoff; ++oit)
            {
               SCIP_Bool fixedit = FALSE;
               assert( *oit < (unsigned int) nvars );

               SCIP_CALL( SCIPfixVar(scip, vars[*oit], 0.0, &cutoff, &fixedit) );
               fixed = fixed || fixedit;
               if ( fixedit )
               {
                  ++(*nfixedvars);
                  presoldata->varsfixedzero.push_back(*oit);
               }
            }
            if ( cutoff )
               break;

            if ( fixed )
            {
               SCIPdebugMessage("Fixed orbit to 0.0.\n");
               fixedvar = TRUE;
               break;
            }
         }

         /* Analyze probing deductions for the two branches - note that we do not store whether variables are fixed to 1
          * here. This is because these deductions have been made without symmetry handling methods. Thus, these
          * variables do not influence the stabilizer, similar to the case in orbital branching where only branching
          * variables are used (and not other fixed variables). */
         int localnfixedvars = 0;
         int localnchgbds = 0;
         SCIP_CALL( analyzeDeductionsProbing(scip, presoldata, it, nvars, vars, zeroproplbs, zeropropubs, oneproplbs, onepropubs,
               &localnfixedvars, &localnchgbds, cutoff) );
         *nfixedvars += localnfixedvars;
         *nchgbds += localnchgbds;

         if ( localnfixedvars > 0 || localnchgbds > 0 )
         {
            fixedvar = TRUE;
            SCIPdebugMessage("Deduced fixings: %d, bound changes: %d.\n", localnfixedvars, localnchgbds);
            break;
         }
      }

      /* if we found a cutoff exit */
      if ( cutoff )
         break;

      ++nrounds;
   }
   while ( fixedvar );

   /* free temporary memory */
   SCIPfreeBufferArray(scip, &checkedvar);
   SCIPfreeBufferArray(scip, &onepropubs);
   SCIPfreeBufferArray(scip, &oneproplbs);
   SCIPfreeBufferArray(scip, &zeropropubs);
   SCIPfreeBufferArray(scip, &zeroproplbs);

   return SCIP_OKAY;
}



/*
 * Callback methods of presolver
 */

/** destructor of presolver to free user data (called when SCIP is exiting) */
static
SCIP_DECL_PRESOLFREE(presolFreeOrbitprobing)
{  /*lint --e{715,818}*/
   SCIP_PRESOLDATA* presoldata;

   assert( scip != 0 );
   assert( presol != 0 );

   presoldata = SCIPpresolGetData(presol);
   assert( presoldata != 0 );

   presoldata->group.reset();
   presoldata->binaryvars.reset();
   SCIPfreeMemory(scip, &presoldata);

   return SCIP_OKAY;
}


/** initialization method of presolver (called after problem was transformed) */
static
SCIP_DECL_PRESOLINIT(presolInitOrbitprobing)
{
   SCIP_PRESOLDATA* presoldata;

   assert( scip != 0 );
   assert( presol != 0 );

   SCIPdebugMessage("Initpre method of orbit probing presolver ...\n");

   presoldata = SCIPpresolGetData(presol);
   assert( presoldata != 0 );

   /* check whether we should run */
   SCIP_CALL( SCIPgetBoolParam(scip, "useorbitprobing", &presoldata->enabled) );

   if ( presoldata->enabled )
   {
      /* work on binary variables while fixing integer variables - same settings as branch_orbital */
      SYMsetProbSpecRequirement(presoldata->symmetriesconshdlr, SCIP_PROBSPEC_BINARY);
      SYMsetProbSpecRequirementFixed(presoldata->symmetriesconshdlr, SCIP_PROBSPEC_INTEGER);
   }

   return SCIP_OKAY;
}


/** presolving initialization method of presolver (called when presolving is about to begin) */
static
SCIP_DECL_PRESOLINITPRE(presolInitpreOrbitprobing)
{  /*lint --e{715}*/
   SCIP_PRESOLDATA* presoldata;

   assert( scip != 0 );
   assert( presol != 0 );

   presoldata = SCIPpresolGetData(presol);
   assert( presoldata != 0 );

   /* perform probing in every run */
   presoldata->performedprobing = FALSE;

   /* check whether we have to run the presolver at the beginning of presolving */
   presoldata->early = ! SYMdetectSymmetriesPresolved(presoldata->symmetriesconshdlr);
   if ( presoldata->early && presoldata->enabled )
   {
      SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Executing presolver <%s> early, since symmetries are computed early.\n\n", SCIPpresolGetName(presol));

      SCIP_CALL( SCIPsetIntParam(scip, "presolving/" PRESOL_NAME"/priority", 90000000) );
#if SCIP_VERSION < 320
      SCIP_CALL( SCIPsetBoolParam(scip, "presolving/" PRESOL_NAME"/delay", FALSE) );
#endif
   }

   return SCIP_OKAY;
}


/** execution method of presolver */
static
SCIP_DECL_PRESOLEXEC(presolExecOrbitprobing)
{  /*lint --e{715,818}*/
   SCIP_PRESOLDATA* presoldata;
   SCIP_Bool cutoff = FALSE;

   assert( scip != 0 );
   assert( presol != 0 );
   assert( result != 0 );

   *result = SCIP_DIDNOTRUN;

   /* avoid recursive callp */
   if ( SCIPinProbing(scip) )
      return SCIP_OKAY;

   presoldata = SCIPpresolGetData(presol);
   assert( presoldata != 0 );

   /* possibly skip presolver */
   if ( ! presoldata->enabled )
      return SCIP_OKAY;

   /* if we already performed probing, stop */
   if ( presoldata->performedprobing )
      return SCIP_OKAY;

   /* skip presolving if we are not at the end */
   if ( ! presoldata->early && ! SCIPisPresolveFinished(scip) )
      return SCIP_OKAY;

   /* avoid calling probing */
   if ( SCIPisStopped(scip) )
      return SCIP_OKAY;

   SCIPdebugMessage("Presolving method of orbit probing presolver ...\n");

   /* get symmetry if not already present */
   if ( ! presoldata->group )
   {
      /* get symmetry information */
      SCIP_CALL( SYMgetSymmetryData(scip, presoldata->symmetriesconshdlr, presoldata->group, presoldata->nvars, presoldata->vars,
            presoldata->binaryvars, presoldata->varmap) );
      if (! presoldata->group)
      {
         presoldata->performedprobing = TRUE;  /* to avoid later calls */
         presoldata->enabled = FALSE;
      }
   }

   /* avoid calling probing */
   if ( SCIPisStopped(scip) )
      return SCIP_OKAY;

   if ( presoldata->group )
   {
      SCIPdebugMessage("Applying orbit probing ...\n");
      SCIP_CALL( applyProbing(scip, presoldata, nfixedvars, nchgbds, cutoff) );
      presoldata->performedprobing = TRUE;

      /* correct group by stabilizers */
      SCIP_CALL( SYMreplaceSymmetryWithStabilizer(scip, presoldata->symmetriesconshdlr, presoldata->varsfixedzero) );
      SCIP_CALL( SYMreplaceSymmetryWithStabilizer(scip, presoldata->symmetriesconshdlr, presoldata->varsfixedone) );

      if ( cutoff )
         *result = SCIP_CUTOFF;
      else
         *result = SCIP_SUCCESS;
   }

   return SCIP_OKAY;
}





/*
 * presolver specific interface methods
 */

/** creates the orbitprobing presolver and includes it in SCIP */
SCIP_RETCODE SCIPincludePresolOrbitprobing(
   SCIP*                 scip                /**< SCIP data structure */
   )
{
   SCIP_PRESOLDATA* presoldata = 0;

   /* create presolver data */
   SCIP_CALL( SCIPallocMemory(scip, &presoldata) );

   /* we must call the constructor explictly, because memory was m'alloced and not new'ed */
   new (&presoldata->group) PermutationGroupPtr();
   new (&presoldata->binaryvars) uintvecPtr();
   presoldata->vars = 0;
   presoldata->nvars = 0;
   presoldata->varmap = 0;
   presoldata->performedprobing = FALSE;
   presoldata->enabled = TRUE;
   presoldata->early = FALSE;

   /* determine cons_symmetries constraint handler */
   SCIP_CONSHDLR* conshdlr = SCIPfindConshdlr(scip, "symmetries");
   if ( conshdlr == 0 )
   {
      SCIPerrorMessage("Could not find constraint handler <symmetries>.\n");
      return SCIP_ERROR;
   }
   presoldata->symmetriesconshdlr = conshdlr;

   /* include presolver */
   SCIP_PRESOL* presol = 0;
#if SCIP_VERSION >= 320
   SCIP_CALL( SCIPincludePresolBasic(scip, &presol, PRESOL_NAME, PRESOL_DESC, PRESOL_PRIORITY, PRESOL_MAXROUNDS, PRESOL_TIMING,
         presolExecOrbitprobing, presoldata) );
#else
   SCIP_CALL( SCIPincludePresolBasic(scip, &presol, PRESOL_NAME, PRESOL_DESC, PRESOL_PRIORITY, PRESOL_MAXROUNDS, PRESOL_DELAY, presolExecOrbitprobing, presoldata) );
#endif

   /* set additional callbacks */
   SCIP_CALL( SCIPsetPresolFree(scip, presol, presolFreeOrbitprobing) );
   SCIP_CALL( SCIPsetPresolInit(scip, presol, presolInitOrbitprobing) );
   SCIP_CALL( SCIPsetPresolInitpre(scip, presol, presolInitpreOrbitprobing) );

   /* add orbit probing presolver parameters */
   SCIP_CALL( SCIPaddBoolParam(scip,
         "presolving/" PRESOL_NAME"/useprobeone",
         "Should probing on the one branch be performed?",
         &presoldata->useprobeone, TRUE, DEFAULT_USEPROBEONE, NULL, NULL) );

   SCIP_CALL( SCIPaddBoolParam(scip,
         "presolving/" PRESOL_NAME"/markall",
         "Should each variable of a checked orbit me marked not to be tried again as first variable?",
         &presoldata->markall, TRUE, DEFAULT_MARKALL, NULL, NULL) );

   return SCIP_OKAY;
}

/** get number of variables fixed to 0 by orbit probing */
extern
std::size_t SCIPorbitprobingGetNFixedZero(
   SCIP*                 scip,               /**< SCIP data structure */
   SCIP_PRESOL*          presol              /**< orbit probing presolver */
   )
{
   assert( presol != 0 );
   assert( strcmp(SCIPpresolGetName(presol), PRESOL_NAME) == 0 );

   SCIP_PRESOLDATA* presoldata = SCIPpresolGetData(presol);
   assert( presoldata != 0 );

   return presoldata->varsfixedzero.size();
}

/** get array of variables fixed to 0 by orbit probing */
extern
uintvec& SCIPorbitprobingGetFixedZero(
   SCIP*                 scip,               /**< SCIP data structure */
   SCIP_PRESOL*          presol              /**< orbit probing presolver */
   )
{
   assert( presol != 0 );
   assert( strcmp(SCIPpresolGetName(presol), PRESOL_NAME) == 0 );

   SCIP_PRESOLDATA* presoldata = SCIPpresolGetData(presol);
   assert( presoldata != 0 );

   return presoldata->varsfixedzero;
}

/** get number of variables fixed to 1 by orbit probing */
extern
std::size_t SCIPorbitprobingGetNFixedOne(
   SCIP*                 scip,               /**< SCIP data structure */
   SCIP_PRESOL*          presol              /**< orbit probing presolver */
   )
{
   assert( presol != 0 );
   assert( strcmp(SCIPpresolGetName(presol), PRESOL_NAME) == 0 );

   SCIP_PRESOLDATA* presoldata = SCIPpresolGetData(presol);
   assert( presoldata != 0 );

   return presoldata->varsfixedone.size();
}

/** get array of variables fixed to 1 by orbit probing */
extern
const uintvec& SCIPorbitprobingGetFixedOne(
   SCIP*                 scip,               /**< SCIP data structure */
   SCIP_PRESOL*          presol              /**< orbit probing presolver */
   )
{
   assert( presol != 0 );
   assert( strcmp(SCIPpresolGetName(presol), PRESOL_NAME) == 0 );

   SCIP_PRESOLDATA* presoldata = SCIPpresolGetData(presol);
   assert( presoldata != 0 );

   return presoldata->varsfixedone;
}

/* vim: set tabstop=3 expandtab: */
// kate: space-indent on; indent-width 3; mixedindent off; indent-mode cstyle;
