/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*    This file is part of "mipsymmetries" a collection of                   */
/*    symmetry handling methods.                                             */
/*                                                                           */
/*    Copyright (C) 2010-2015  Marc Pfetsch, Thomas Rehn                     */
/*                                                                           */
/*                                                                           */
/*    Based on SCIP  --- Solving Constraint Integer Programs                 */
/*       SCIP is distributed under the terms of the SCIP Academic Licence,   */
/*       see file COPYING in the SCIP distribution.                          */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
#pragma ident "@(#) $Id: presol_symbreak.cpp 124 2011-10-19 18:19:40Z pfetsch $"

/**@file   presol_symbreak.cpp
 * @ingroup PRESOLVERS
 * @brief  presolver for adding symmetry breaking constraints
 * @author Marc Pfetsch
 * @author Thomas Rehn
 *
 * This presolver heuristically adds several symmetry breaking constraints:
 *
 * - It tries to detect orbitope structures and possibly adds the corresponding constraints.
 *
 * - It tries to detect symmetric groups and adds ordering constraints: If a symmetric group is acting on
 *   variables \f$x_1, \ldots, x_k\f$, it adds
 *   \f[
 *       x_1 \geq x_2,\; x_2 \geq x_3, \ldots, x_{k-1} \geq x_k.
 *   \f]
 *
 * - It tries to detect cyclic groups and adds the following ordering constraints: If a cyclic group is acting on
 *   variables \f$x_1, \ldots, x_k\f$, it adds
 *   \f[
 *       x_1 \geq x_2,\; x_1 \geq x_3, \ldots, x_1 \geq x_k.
 *   \f]
 *
 * - Otherwise is tries to add symmetry breaking constraints for the fundamental region. If a permutation \f$\pi\f$
 *   moves variables \f$x_1, \ldots, x_k\f$, it adds
 *   \f[
 *       x_1 + 2 x_2 + \ldots + k x_k \geq x_{\phi(1)} + 2 x_{\phi(2)} + \ldots + k x_{\phi(k)}.
 *   \f]
 *
 * - Moreover it can add inequalities for each orbit that correspond to the cyclic group inequalies.
 *
 * @note It is important to control the order of presolvers within SCIP in order to avoid contraditions. First on needs
 * to take care of presolvers that have an effect on symmetry, e.g., presol_domcol. If symmetry is computed on the
 * original formulation, we perform this presolver at the very beginning. Otherwise, we try to compute symmetry as late
 * as possible and then add constraints based on this information.
 *
 * @todo Try to combine option @p orbitminonly with the other inequalities, e.g., first generate inequalities for full
 * symmetry groups and then for the remaining orbits.
 *
 * @todo Find better inequalities for cyclic groups (i.e., for the @p orbitminonly variant).
 */

/*---+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8----+----9----+----0----+----1----+----2*/

#include "presol_symbreak.h"
#include "cons_symmetries.h"

#include <scip/cons_linear.h>
#include <scip/cons_orbitope.h>
#include <scip/cons_setppc.h>
#include <scip/prop_probing.h>

#include <cassert>
#include <cstring>
#include <vector>
#include <list>
#include <set>

#include <permlib/test/group_type.h>
#include <boost/dynamic_bitset.hpp>

/* presolver properties */
#define PRESOL_NAME            "symbreak"
#define PRESOL_DESC            "presolver for adding symmetry breaking constraints"
#define PRESOL_PRIORITY         -10000000    /**< priority of the presolver (>= 0: before, < 0: after constraint handlers) */
#define PRESOL_MAXROUNDS               -1    /**< maximal number of presolving rounds the presolver participates in (-1: no limit) */
#define PRESOL_DELAY                 TRUE    /**< should presolver be delayed, if other presolvers found reductions? */
#define PRESOL_TIMING   SCIP_PRESOLTIMING_EXHAUSTIVE   /**< timing for presolving */

/* default parameters */
#define DEFAULT_CONSSADDLP           TRUE    /**< Should the symmetry breaking constraints be added to the LP? */
#define DEFAULT_USEPROBING          FALSE    /**< Apply probing to binary variables appearing in symmetry breaking constraints? */
#define DEFAULT_ADDFUNDAMENTAL      FALSE    /**< Add inequalities from the fundamental domain for each generator? */
#define DEFAULT_ORBITMINONLY        FALSE    /**< Add only those symmetry breaking constraints which enforce minimality of a fixed, but arbitrary orbit element */
#define DEFAULT_MINORDER               -1    /**< minimal order of the symmetry group for which the presolver runs */



/* other parameters */
#define MAXPROBINGPROPROUNDS          100    /**< maximal number of propagation rounds in probing */

using namespace permlib;


/* useful typedefs */
typedef boost::shared_ptr<permlib::PermutationGroup> PermGroupPtr;
typedef std::list<SCIP_CONS*>                        conslist;
typedef std::vector<unsigned int>                    uintvector;
typedef std::set<unsigned int>                       uintset;


typedef permlib::AbstractPermutationGroup::OrbitList OrbitList;
typedef boost::scoped_ptr<OrbitList> OrbitListPtr;
typedef std::set<permlib::dom_int>::const_iterator OrbitSetIt;
typedef OrbitList::const_iterator OrbIt;


/*
 * Data structures
 */

/** presolver data */
struct SCIP_PresolData
{
   PermutationGroupPtr   group;              /**< symmetry group */
   unsigned int          nvars;              /**< number of variables for symmetry computations */
   SCIP_VAR**            vars;               /**< variables used for symmetry computations */
   SCIP_HASHMAP*         varmap;             /**< map of variables to indices in vars array */
   SCIP_Bool             addedconss;         /**< whether we already added symmetry breaking constraints */
   SCIP_Bool             performedprobing;   /**< whether we already performed probing */
   SCIP_Bool             computedsymmetry;   /**< whether symmetry has been computed already */
   int                   nprobsets;          /**< number of sets of possible probing variables */
   int                   nmaxprobsets;       /**< maximal number of sets of possible probing variables */
   int*                  nprobvars;          /**< number of binary variables in each set */
   SCIP_VAR***           probvars;           /**< sets of binary variables to apply probing on */
   SCIP_Bool             conssaddlp;         /**< Should the symmetry breaking constraints be added to the LP? */
   SCIP_Bool             useprobing;         /**< Apply probing to binary variables appearing in symmetry breaking constraints? */
   SCIP_Bool             addfundamental;     /**< Add inequalities from the fundamental domain for each generator? */
   SCIP_Bool             orbitminonly;       /**< Add only those symmetry breaking constraints which enforce minimality of a fixed, but arbitrary orbit element */
   int                   minorder;           /**< minimal order of the symmetry group for which the presolver runs */
   SCIP_CONSHDLR*        symmetriesconshdlr; /**< pointer to cons_symmetries constraint handler */
   uintvecPtr            binaryvars;         /**< indices of binary variables to perform orbital branching on */
   SCIP_Bool             enabled;            /**< run presolver? */
   SCIP_Bool             early;              /**< run presolver as early as possible if symmetry has been detected in initpre() */
   unsigned int          norbitopes;         /**< number of orbitope constraints added */
   unsigned int          nsymgroupbreak;     /**< number of full symmetric groups handled */
   unsigned int          ncycgroupbreak;     /**< number of cyclic groups handled */
   unsigned int          nfundbreak;         /**< number of fundamental domain constraints */
   unsigned int          ngenconss;          /**< number of generated constraints */
   conslist*             genconss;           /**< list of generated constraints */
};


/*
 * Local methods
 */

/** add set of possible probing variables */
static
SCIP_RETCODE addProbingSet(
   SCIP*                 scip,               /**< SCIP instance */
   SCIP_PRESOLDATA*      presoldata,         /**< presolver data */
   int                   nprobvars,          /**< number of variables in set */
   SCIP_VAR**            probvars            /**< sets of variables */
   )
{
   int j;

   assert( scip != 0 );
   assert( presoldata != 0 );
   assert( probvars != 0 );

   /* check whether all variables are binary */
   for (j = 0; j < nprobvars; ++j)
   {
      if ( SCIPvarGetType(probvars[j]) != SCIP_VARTYPE_BINARY )
         return SCIP_OKAY;
   }

   /* check if this is the first call */
   if ( presoldata->nmaxprobsets == 0 )
   {
      SCIP_CALL( SCIPallocBlockMemoryArray(scip, &presoldata->nprobvars, 10) );
      SCIP_CALL( SCIPallocBlockMemoryArray(scip, &presoldata->probvars, 10) );
      presoldata->nprobsets = 1;
      presoldata->nprobvars[0] = nprobvars;
      presoldata->probvars[0] = probvars;
      presoldata->nmaxprobsets = 10;
      return SCIP_OKAY;
   }

   /* enlarge array if necessary */
   if ( presoldata->nprobsets >= presoldata->nmaxprobsets )
   {
      int newsize = SCIPcalcMemGrowSize(scip, presoldata->nprobsets+1);
      SCIP_CALL( SCIPreallocBlockMemoryArray(scip, &presoldata->nprobvars, presoldata->nmaxprobsets, newsize) );
      SCIP_CALL( SCIPreallocBlockMemoryArray(scip, &presoldata->probvars, presoldata->nmaxprobsets, newsize) );
      presoldata->nmaxprobsets = newsize;
   }
   presoldata->nprobvars[presoldata->nprobsets] = nprobvars;
   presoldata->probvars[presoldata->nprobsets] = probvars;
   ++presoldata->nprobsets;
   assert( presoldata->nprobsets <= presoldata->nmaxprobsets );

   return SCIP_OKAY;
}


/** add ordering inequalities for a full symmetric subgroup */
template<class Iterator>
static
SCIP_RETCODE addSymmetricGroupBreakingConstraints(
   SCIP*                 scip,               /**< SCIP instance */
   SCIP_PRESOL*          presol,             /**< presolver */
   std::size_t           orbsize,            /**< size of orbit */
   const Iterator&       orbbegin,           /**< iterator over indices of variables to add */
   const Iterator&       orbend,             /**< end of iterator */
   bool                  coresetineq         /**< true if inequality based on core set should be added */
   )
{
   SCIP_CONS* cons;
   SCIP_VAR* vars[2];
   SCIP_Real vals[2] = {1,-1};
   char name[SCIP_MAXSTRLEN];
   SCIP_VAR** probvars = 0;
   int nprobvars = 0;

   assert( scip != 0 );
   assert( presol != 0 );

   SCIP_PRESOLDATA* presoldata = SCIPpresolGetData(presol);
   assert( presoldata != 0 );
   ++presoldata->nsymgroupbreak;

   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &probvars, (int) orbsize) );

   Iterator orbit = orbbegin;
   Iterator prev = orbit;
   probvars[nprobvars++] = presoldata->vars[*orbit];

   /* look through orbit - skipping the first element */
   for (++orbit; orbit != orbend; ++orbit)
   {
      assert( *orbit < presoldata->nvars );
      assert( *prev < presoldata->nvars );

      probvars[nprobvars++] = presoldata->vars[*orbit];

      vars[0] = presoldata->vars[*prev];
      vars[1] = presoldata->vars[*orbit];
      assert( vars[0] != 0 );
      assert( vars[1] != 0 );

      (void) SCIPsnprintf(name, SCIP_MAXSTRLEN-1, "fullsymorder%u", presoldata->ngenconss);
      SCIP_CALL( SCIPcreateConsLinear(scip, &cons, name, 2, vars, vals, 0.0, SCIPinfinity(scip),
            presoldata->conssaddlp, presoldata->conssaddlp, FALSE, FALSE, TRUE, FALSE, FALSE, FALSE, TRUE, FALSE) );
      SCIP_CALL( SCIPaddCons(scip, cons) );
#ifdef SCIP_DEBUG
      SCIP_CALL( SCIPprintCons(scip, cons, 0) );
      SCIPinfoMessage(scip, 0, "\n");
#endif

      /* turn off addition of coreset inequality if some variable is not integeral */
      if ( SCIPvarGetType(vars[1]) == SCIP_VARTYPE_CONTINUOUS )
         coresetineq = false;

      /* constraint will be released at end */
      presoldata->genconss->push_back(cons);
      ++presoldata->ngenconss;
      prev = orbit;
   }

   /* store probing variables */
   SCIP_CALL( addProbingSet(scip, presoldata, nprobvars, probvars) );

   /* add range constraint */
   if ( coresetineq )
   {
      /* prev should point to the last element! */
#ifndef NDEBUG
      Iterator tmp = prev;
      assert( ++tmp == orbend );
#endif
      const SCIP_Real varUb = SCIPvarGetUbLocal(presoldata->vars[*orbbegin]);
      const SCIP_Real varLb = SCIPvarGetLbLocal(presoldata->vars[*orbbegin]);

      if ( SCIPisInfinity(scip, varUb) || SCIPisInfinity(scip, -varLb) || (varUb - varLb > 1.5) )
      {
         vars[0] = presoldata->vars[*prev];
         vars[1] = presoldata->vars[*orbbegin];
         SCIP_CALL( SCIPcreateConsLinear(scip, &cons, "fullsymdelta", 2, vars, vals, -1.0, SCIPinfinity(scip),
               presoldata->conssaddlp, presoldata->conssaddlp, FALSE, FALSE, TRUE, FALSE, FALSE, FALSE, TRUE, FALSE) );
         SCIP_CALL( SCIPaddCons(scip, cons) );
#ifdef SCIP_DEBUG
         SCIP_CALL( SCIPprintCons(scip, cons, NULL) );
         SCIPinfoMessage(scip, NULL, "\n");
#endif

         /* constraint will be released at end */
         presoldata->genconss->push_back(cons);
         ++presoldata->ngenconss;
      }
   }

   return SCIP_OKAY;
}


/** add ordering inequalities for a cyclic subgroup */
template<class Iterator>
static
SCIP_RETCODE addCyclicGroupBreakingConstraints(
   SCIP*                 scip,               /**< SCIP instance */
   SCIP_PRESOL*          presol,             /**< presolver */
   const Iterator&       orbbegin,           /**< iterator over indices of variables to add */
   const Iterator&       orbend              /**< end of iterator */
   )
{
   SCIP_CONS* cons;
   SCIP_VAR* vars[2];
   SCIP_Real vals[2] = {1,-1};
   SCIP_VAR** probvars = 0;

   assert( scip != 0 );
   assert( presol != 0 );
   SCIP_PRESOLDATA* presoldata = SCIPpresolGetData(presol);
   assert( presoldata != 0 );
   ++presoldata->ncycgroupbreak;

   vars[0] = presoldata->vars[*orbbegin];
   Iterator orbit = orbbegin;
   assert( orbit != orbend );

   /* only store first variable - the others are unlikely to yield reductions */
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &probvars, 1) );
   probvars[0] = presoldata->vars[*orbbegin];
   SCIP_CALL( addProbingSet(scip, presoldata, 1, probvars) );

   for (++orbit; orbit != orbend; ++orbit)
   {
      assert( *orbit < presoldata->nvars );
      vars[1] = presoldata->vars[*orbit];

      SCIP_CALL( SCIPcreateConsLinear(scip, &cons, "cycbreak", 2, vars, vals, 0.0, SCIPinfinity(scip),
            presoldata->conssaddlp, presoldata->conssaddlp, FALSE, FALSE, TRUE, FALSE, FALSE, FALSE, TRUE, FALSE) );
      SCIP_CALL( SCIPaddCons(scip, cons) );
#ifdef SCIP_DEBUG
      SCIP_CALL( SCIPprintCons(scip, cons, NULL) );
      SCIPinfoMessage(scip, NULL, "\n");
#endif
      /* constraint will be released at end */
      presoldata->genconss->push_back(cons);
      ++presoldata->ngenconss;
   }

   return SCIP_OKAY;
}


/** sort columns of matrix for orbitope constraints */
struct OrbitopeColumnSorter
{
   OrbitopeColumnSorter(const std::vector<dom_int>& orbitCharacteristicTransponed)
      : m_orbitCharacteristicTransponed(orbitCharacteristicTransponed) { }

   bool operator()(const unsigned int& a, const unsigned int& b) const
   {
      assert( a < m_orbitCharacteristicTransponed.size() );
      assert( b < m_orbitCharacteristicTransponed.size() );
      return m_orbitCharacteristicTransponed[a] < m_orbitCharacteristicTransponed[b];
   }

private:
   const std::vector<dom_int>& m_orbitCharacteristicTransponed;
};


/** check whether all rows in potential orbitope structure are covered by set partitioning or
 *  packing constraints and add orbitope constraint if possible */
SCIP_RETCODE addCoveredOrbitope(
   SCIP*                 scip,               /**< SCIP */
   SCIP_PRESOLDATA*      presoldata,         /**< data for presolver */
   SCIP_CONS**           setppcconss,        /**< all setppc constraints */
   int                   nsetppcconss,       /**< number of setppc constraints */
   SCIP_SETPPCTYPE       type,               /**< type of setppc constraints to check */
   int                   nvars,              /**< total number of variables */
   SCIP_VAR***           vars,               /**< candidate matrix of variables */
   int&                  nrows,              /**< number of rows of variable matrix (possibly reduced) */
   int                   ncols,              /**< number of columns of variable matrix */
   const int*            varsrow,            /**< row index for variables (-1 if not present) */
   const int*            varscol,            /**< column index for variables (-1 if no present) */
   SCIP_Bool*            covered,            /**< marks the rows that are covered */
   SCIP_CONS*&           orbitopecons,       /**< pointer to orbitope constraint (or 0) */
   bool&                 success             /**< wether orbitope constraint has been added */
   )
{
   assert( vars != 0 );
   assert( nrows > 0 );
   assert( ncols > 0 );
   assert( covered != 0 );

   orbitopecons = 0;
   success = false;

   /* prepare data */
   for (int i = 0; i < nrows; ++i)
      covered[i] = FALSE;

   /* loop through set partitioning constraints */
   int ncovered = 0;
   for (int c = 0; c < nsetppcconss; ++c)
   {
      /* check type */
      if ( SCIPgetTypeSetppc(scip, setppcconss[c]) != type )
         continue;

      /* get set partitioning variables */
      int nsetppcvars = SCIPgetNVarsSetppc(scip, setppcconss[c]);
      if ( nsetppcvars != ncols )
         continue;
      assert( nsetppcvars > 1 );
      SCIP_VAR** setppcvars = SCIPgetVarsSetppc(scip, setppcconss[c]);
      assert( setppcvars != 0 );

      /* determine possible row: check first variable */
      SCIP_VAR* var = setppcvars[0];
      assert( var != 0 );
      if ( SCIPvarIsNegated(var) )
         var = SCIPvarGetNegatedVar(var);
      assert( var != 0 );
      int varidx = SCIPvarGetProbindex(var);
      assert( 0 <= varidx && varidx < nvars );

      /* skip constraints that contain variables that are not in candidate matrix */
      if ( varsrow[varidx] < 0 )
         continue;
      assert( 0 <= varsrow[varidx] && varsrow[varidx] < nrows );
      assert( 0 <= varscol[varidx] && varscol[varidx] < ncols );

      /* check whether all variables are in the same row */
      int rowidx = varsrow[varidx];
      for (int v = 1; v < nsetppcvars; ++v)
      {
         var = setppcvars[v];
         assert( var != 0 );
         if ( SCIPvarIsNegated(var) )
            var = SCIPvarGetNegatedVar(var);
         assert( var != NULL );
         varidx = SCIPvarGetProbindex(var);
         assert( 0 <= varidx && varidx < nvars );

         /* variable in settppc is not contained in potential orbitope */
         if ( varsrow[varidx] < 0 || varscol[varidx] < 0 )
         {
            /* exit loop for current setppc constraint */
            rowidx = -1;
            break;
         }

         assert( 0 <= varsrow[varidx] && varsrow[varidx] < nrows );
         assert( 0 <= varscol[varidx] && varscol[varidx] < ncols );

         /* have different row? */
         if ( varsrow[varidx] != rowidx )
            break;
      }

      if ( rowidx < 0 )
         continue;

      /* constraint covers row with index rowidx */
      if ( varsrow[varidx] == rowidx )
      {
         if ( ! covered[rowidx] )
         {
            ++ncovered;
            covered[rowidx] = TRUE;
         }
      }
   }
   SCIPdebugMessage("%d rows of %d are covered by set %s constraints.\n", ncovered, nrows,
      type == SCIP_SETPPCTYPE_PARTITIONING ? "partitioning" : "packing");

   /* if not enough rows are covered */
   if ( ncovered <= 1 )
   {
      SCIPdebugMessage("Not enough rows are covered: no orbitope constraint found.\n");
      return SCIP_OKAY;
   }

   /* remove rows that are not covered */
   if ( ncovered < nrows )
   {
      int r = nrows-1;
      while ( r >= 0 )
      {
         if ( ! covered[r] )
         {
            /* move variables - do not need to move "covered", since we work backwards */
            SCIPfreeBufferArray(scip, &(vars[r]));
            for (int k = r; k < nrows-1; ++k)
               vars[k] = vars[k+1];
            --nrows;
         }
         --r;
      }
   }

   /* add diagonal variables to probing */
   for (int i = 0; i < MIN(ncols, nrows); ++i)
   {
      SCIP_VAR** probvars = 0;
      SCIP_CALL( SCIPallocBlockMemoryArray(scip, &probvars, 1) );
      probvars[0] = vars[i][i];
      SCIP_CALL( addProbingSet(scip, presoldata, 1, probvars) );
   }

   /* create orbitope constraint */
   SCIPdebugMessage("Create orbitope constraint with %d rows and %d columns ...\n", nrows, ncols);
   SCIP_CONS* cons;
   if ( type == SCIP_SETPPCTYPE_PARTITIONING )
   {
      SCIP_CALL( SCIPcreateConsOrbitope(scip, &cons, "orbitope", vars, TRUE, nrows, ncols, TRUE,
            TRUE, FALSE, FALSE, FALSE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE) );
   }
   else
   {
      SCIP_CALL( SCIPcreateConsOrbitope(scip, &cons, "orbitope", vars, FALSE, nrows, ncols, TRUE,
            TRUE, FALSE, FALSE, FALSE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE) );
   }
   SCIP_CALL( SCIPaddCons(scip, cons) );

#ifdef SCIP_DEBUG
   SCIP_CALL( SCIPprintCons(scip, cons, NULL) );
   SCIPinfoMessage(scip, NULL, "\n");
#endif
#ifdef SCIP_OUTPUT
   for (int i = 0; i < nrows; ++i)
   {
      for (int j = 0; j < ncols; ++j)
      {
         SCIPinfoMessage(scip, NULL, "<%s> [%f,%f,%f]", SCIPvarGetName(vars[i][j]), SCIPvarGetObj(vars[i][j]),
            SCIPvarGetLbGlobal(vars[i][j]), SCIPvarGetUbGlobal(vars[i][j]));
      }
      SCIPinfoMessage(scip, NULL, "\n");
   }
#endif
   /* do not release constraint here - need it for possible presolving */

   orbitopecons = cons;
   success = true;

   return SCIP_OKAY;
}


/** try to add orbitope constraint */
static
SCIP_RETCODE tryOrbitopeConstraint(
   SCIP*                 scip,               /**< SCIP instance */
   SCIP_PRESOL*          presol,             /**< presolver */
   const PermGroupPtr&   group,              /**< permutation group */
   uintvector&           componentIndices,   /**< variable indices of the group component to be considered */
   uintvector&           orbitTransponed,    /**< array that marks for every group component index to which orbit it belongs */
   unsigned int          expectedColumnSize, /**< size of a supposed orbitope column in componentIndices */
   SCIP_CONS*&           orbitopecons,       /**< pointer to orbitope constraint (or 0) */
   bool&                 success             /**< whether orbitope constraint has been added */
   )
{
   assert( scip != 0 );
   assert( presol != 0 );

   SCIP_PRESOLDATA* presoldata = SCIPpresolGetData(presol);
   assert( presoldata != 0 );

   orbitopecons = 0;
   success = false;

   SCIPdebugMessage("Checking symmetric group acting on blocks of variables for orbitope structure ...\n");

   /* get copy of group and prepare computations */
   PermutationGroup groupCopy(*group);
   ConjugatingBaseChange<PERMUTATION,TRANSVERSAL, RandomBaseTranspose<PERMUTATION,TRANSVERSAL> > baseChange(groupCopy);

   /* determine which column a variable belongs to; n+1 means that no column has been identified yet */
   assert( group->n == presoldata->nvars );
   std::vector<dom_int> orbitopeColsMap(group->n, group->n+1);

   /* check all variables */
   int nrows = -1;
   int ncols = 0;
   for (uintvector::const_iterator dit = componentIndices.begin(); dit != componentIndices.end(); ++dit)
   {
      const dom_int alpha = *dit;
      assert( alpha < presoldata->nvars );

      /* only use binary variables */
      if ( ! SCIPvarIsBinary(presoldata->vars[alpha]) )
         continue;

      /* only look at elements that belong to the first orbit (with identifier 0) */
      if ( orbitTransponed[alpha] != 0 )
         continue;

      /* only consider entries not yet assigned */
      if ( orbitopeColsMap[alpha] != group->n+1 )
         continue;

      /* change base to start with alpha, see below ... */
      baseChange.change(groupCopy, &alpha, (&alpha)+1);

      int columnSize = 0;
      for (uintvector::const_iterator bit = componentIndices.begin(); bit != componentIndices.end(); ++bit)
      {
         const dom_int b = *bit;

         /* only use binary variables */
         if ( ! SCIPvarIsBinary(presoldata->vars[b]) )
            continue;

         /* only consider entries not yet assigned */
         if ( orbitopeColsMap[b] != group->n+1 )
            continue;

         /* Implicitly iterate over the stabilizer of alpha.  Because we changed the BSGS such that the base starts with
          * alpha, the set S contains generators for the stabilizer of alpha. */
         bool fixedForAllGenerators = true;
         for (std::list<PERMUTATION::ptr>::const_iterator pit = groupCopy.S.begin(); pit != groupCopy.S.end(); ++pit)
         {
            const PERMUTATION::ptr& p = *pit;

            /* only look at generators of the stabilizer of alpha */
            if ( p->at(alpha) != alpha )
               continue;

            if ( p->at(b) != b )
            {
               fixedForAllGenerators = false;
               break;
            }
         }

         if ( fixedForAllGenerators )
         {
            /* b is in the same column as alpha if it is also fixed by the stabilizer of alpha */
            assert( orbitopeColsMap[b] == group->n+1 );
            orbitopeColsMap[b] = (unsigned) ncols;
            ++columnSize;
         }
      }
      ++ncols;

      /* check whether column size agrees */
      if ( nrows >= 0 )
      {
         if ( columnSize != nrows )
         {
            SCIPdebugMessage("Sizes of columns do not match.\n");
            return SCIP_OKAY;
         }
      }
      else
         nrows = columnSize;
   }

   if ( nrows > (int) expectedColumnSize )
   {
      SCIPdebugMessage("Number of rows (%d) larger than expected (%u) -> stop orbitope dectection.\n", nrows, expectedColumnSize);
      return SCIP_OKAY;
   }
   assert( nrows * ncols <= (int) componentIndices.size() );
   assert( nrows == 0 || ! orbitopeColsMap.empty() );

   if ( nrows <= 0 || ncols <= 1 )
   {
      SCIPdebugMessage("Do not have enough rows (%d) or columns (%d) -> stop orbitope detection.\n", nrows, ncols);
      return SCIP_OKAY;
   }
   SCIPdebugMessage("Detected %d rows and %d columns.\n", nrows, ncols);

#if 0
   for (unsigned int d = 0; d < orbitopeCols.size() / expectedColumnSize; ++d)
   {
      std::sort(orbitopeCols.begin() + d * expectedColumnSize, orbitopeCols.begin() + (d+1) * expectedColumnSize, OrbitopeColumnSorter(orbitCharacteristicTransponed));
      print_iterable(orbitopeCols.begin() + d * expectedColumnSize, orbitopeCols.begin() + (d+1) * expectedColumnSize, 0, "orbitope col");
   }
#endif

   /* prepare table of all binary variables that are moved by action */
   SCIP_VAR*** vars;
   SCIP_CALL( SCIPallocBufferArray(scip, &vars, nrows) );
   for (int i = 0; i < nrows; ++i)
   {
      SCIP_CALL( SCIPallocBufferArray(scip, &(vars[i]), ncols) );
      BMSclearMemoryArray(vars[i], ncols);
   }

   /* store row and column indices of variables */
   int* varsrow;
   int* varscol;
   int nvars = SCIPgetNVars(scip);
   SCIP_CALL( SCIPallocBufferArray(scip, &varsrow, nvars) );
   SCIP_CALL( SCIPallocBufferArray(scip, &varscol, nvars) );
   for (int v = 0; v < nvars; ++v)
   {
      varsrow[v] = -1;
      varscol[v] = -1;
   }

   /* collect all variables and corresponding data */
   for (uintvector::const_iterator dit = componentIndices.begin(); dit != componentIndices.end(); ++dit)
   {
      const dom_int alpha = *dit;
      assert( alpha < presoldata->nvars );

      SCIP_VAR* var = presoldata->vars[alpha];
      assert( var != 0 );

      /* only use binary variables */
      if ( ! SCIPvarIsBinary(var) )
         continue;
      assert( SCIPvarIsBinary(var) );

      /* store variable */
      const int row = orbitTransponed[alpha];
      const int col = orbitopeColsMap[alpha];
      if ( row > nrows || col < 0 )
         continue;
      assert( 0 <= row && row < nrows );
      assert( 0 <= col && col < ncols );
      assert( vars[row][col] == NULL );
      vars[row][col] = var;

      /* check for negated variable */
      if ( SCIPvarIsNegated(var) )
         var = SCIPvarGetNegatedVar(var);
      assert( var != 0 );

      int varidx = SCIPvarGetProbindex(var);
      if ( varidx >= 0 )
      {
         assert( varidx < nvars );
         varsrow[varidx] = row;
         varscol[varidx] = col;
      }
   }

   /* check whether all variables are contained in either set packing or set partitioning constraints */
   SCIP_CONSHDLR* setppcconshdlr = SCIPfindConshdlr(scip, "setppc");
   assert( setppcconshdlr != 0 );
   SCIP_CONS** setppcconss = SCIPconshdlrGetConss(setppcconshdlr);
   int nsetppcconss = SCIPconshdlrGetNConss(setppcconshdlr);

   /* get space */
   SCIP_Bool* covered;
   SCIP_CALL( SCIPallocBufferArray(scip, &covered, nrows) );

   /* check set partitioning constraints */
   SCIP_CALL( addCoveredOrbitope(scip, presoldata, setppcconss, nsetppcconss, SCIP_SETPPCTYPE_PARTITIONING, nvars,
         vars, nrows, ncols, varsrow, varscol, covered, orbitopecons, success) );

   if ( ! success )
   {
      assert( orbitopecons == 0 );
      /* check set packing constraints */
      SCIP_CALL( addCoveredOrbitope(scip, presoldata, setppcconss, nsetppcconss, SCIP_SETPPCTYPE_PACKING, nvars,
            vars, nrows, ncols, varsrow, varscol, covered, orbitopecons, success) );

      if ( success )
         ++presoldata->norbitopes;
   }
   else
      ++presoldata->norbitopes;

   SCIPfreeBufferArray(scip, &covered);
   SCIPfreeBufferArray(scip, &varscol);
   SCIPfreeBufferArray(scip, &varsrow);

   for (int r = 0; r < nrows; ++r)
      SCIPfreeBufferArray(scip, &(vars[r]));
   SCIPfreeBufferArray(scip, &vars);

   return SCIP_OKAY;
}


/** add fundamental domain inequalities */
static
SCIP_RETCODE addFundamentalDomainConstraints(
   SCIP*                 scip,               /**< SCIP instance */
   SCIP_PRESOL*          presol,             /**< presolver */
   const std::list< PERMUTATION::ptr >& groupPermutations, /**< generators of permutations */
   const uintset&        brokenIndices       /**< array of indices that have been broken */
   )
{
   SCIP_CONS* cons;
   SCIP_VAR** vars = 0;
   SCIP_Real* vals = 0;
   SCIP_Real* coefs = 0;
   unsigned int nfund = 0;
   char name[SCIP_MAXSTRLEN];

   assert( scip != 0 );
   assert( presol != 0 );

   SCIP_PRESOLDATA* presoldata = SCIPpresolGetData(presol);
   assert( presoldata != 0 );

   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &vars, presoldata->nvars) );
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &vals, presoldata->nvars) );
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &coefs, presoldata->nvars) );

   /* loop through generators of permutation group */
   BOOST_FOREACH(const PERMUTATION::ptr& p, groupPermutations)
   {
      /* init coefficients */
      for (unsigned int j = 0; j < presoldata->nvars; ++j)
         coefs[j] = 0.0;

      /* count pairs of variables that are moved */
      int cnt = 0;
      for (unsigned int j = 0; j < presoldata->nvars; ++j)
      {
         /* get image */
         const unsigned int img = p->at(j);
         assert( img < presoldata->nvars );
         
         /* skip permutations that move variables for which we already have added
          * symmetry breaking constraints */
         if ( img != j && brokenIndices.count(j) )
         {
            cnt = 0;
            break;
         }
         
         if ( img != j )
         {
            coefs[j] += (SCIP_Real) (cnt+1);
            coefs[img] -= (SCIP_Real) (cnt+1);
            ++cnt;
         }
      }

      if ( cnt == 0 )
         continue;

      /* fill dense vector into sparse vector */
      int nvars = 0;
      for (unsigned int j = 0; j < presoldata->nvars; ++j)
      {
         if ( REALABS(coefs[j]) > 0.5 )
         {
            vars[nvars] = presoldata->vars[j];
            vals[nvars++] = coefs[j];
            assert( nvars <= (int) presoldata->nvars );
         }
      }

      /* sort values for probing variables */
      SCIPsortDownRealPtr(vals, (void**) vars, nvars);

      /* set up probing variables */
      SCIP_VAR** probvars = 0;
      SCIP_CALL( SCIPduplicateBlockMemoryArray(scip, &probvars, vars, nvars) );

      /* add probing variables */
      SCIP_CALL( addProbingSet(scip, presoldata, nvars, probvars) );

      /* create linear constraint */
      /* note that the inequalities have to be >= in order not to conflict with ordering inequalities above */
      (void) SCIPsnprintf(name, SCIP_MAXSTRLEN, "fund#%u", presoldata->ngenconss);
      SCIP_CALL( SCIPcreateConsLinear(scip, &cons, name, nvars, vars, vals, 0.0, SCIPinfinity(scip),
            presoldata->conssaddlp, presoldata->conssaddlp, FALSE, FALSE, TRUE, FALSE, FALSE, FALSE, TRUE, FALSE) );
      SCIP_CALL( SCIPaddCons(scip, cons) );
#ifdef SCIP_DEBUG
      SCIP_CALL( SCIPprintCons(scip, cons, NULL) );
      SCIPinfoMessage(scip, NULL, "\n");
#endif
      /* do not release constraint here - will be done later */
      presoldata->genconss->push_back(cons);
      ++presoldata->ngenconss;
      ++presoldata->nfundbreak;
      ++nfund;
   }
   SCIPdebugMessage("Added fundamental domain inequalities: %u.\n", nfund);

   SCIPfreeBlockMemoryArray(scip, &vars, presoldata->nvars);
   SCIPfreeBlockMemoryArray(scip, &vals, presoldata->nvars);

   return SCIP_OKAY;
}


/** analyze symmetries for symmetric and cyclic subgroups and adds symmetry breaking constraints */
static
SCIP_RETCODE addSymmetryBreakingConstraintsBSGS(
   SCIP*                 scip,               /**< SCIP instance */
   SCIP_PRESOL*          presol              /**< presolver */
   )
{
   SCIP_PRESOLDATA* presoldata;

   assert( scip != 0 );
   assert( presol != 0 );

   presoldata = SCIPpresolGetData(presol);
   assert( presoldata != 0 );

   /* exit if no or only trivial symmetry group is available */
   if ( ! presoldata->group || presoldata->group->order() == 1 || presoldata->nvars == 0 )
      return SCIP_OKAY;

   if ( SCIPisStopped(scip) )
      return SCIP_OKAY;

   /* get set-partining constraints for orbitopes (needed below) */
   SCIP_CONSHDLR* setppcconshdlr = SCIPfindConshdlr(scip, "setppc");
   assert( setppcconshdlr != 0 );
   int nsetppcconss = SCIPconshdlrGetNConss(setppcconshdlr);

   const PermGroupPtr group = dynamic_cast<AbstractBSGS<TRANSVERSAL>*>(presoldata->group.get())->bsgs();
   const PermutationGroup& bsgs = *group;

   std::list<ComponentPtr> components;
   SCIP_CALL( determineComponents(group, components) );

   if ( SCIPisStopped(scip) )
      return SCIP_OKAY;

   /* if we want to add inequalities for the minimal element in an orbit only */
   if ( presoldata->orbitminonly )
   {
      BOOST_FOREACH(const ComponentPtr& comp, components)
      {
         /* We are allowed to force an arbitrary element of an arbitary orbit of this component
          * to be minimal. For this we choose the first orbit that we encounter. */
         const size_t firstOrbitElement = comp->bitset.find_first();
         assert( firstOrbitElement != boost::dynamic_bitset<>::npos );

         OrbitAsSet orbit;
         orbit.orbit(firstOrbitElement, bsgs.S, Transversal<PERMUTATION>::TrivialAction());

         /* enforce minimality */
         SCIP_CALL( addCyclicGroupBreakingConstraints(scip, presol, orbit.begin(), orbit.end()) );
      }

      /* do not use other symmetry breaking constraints */
      return SCIP_OKAY;
   }

   /*
    *   STEP 2: Analyze each component
    */

   /* variable indices for which we add complete symmetry breaking constraints */
   uintset brokenIndices;
   int cc = 0;
   BOOST_FOREACH(const ComponentPtr& comp, components)
   {
      if ( SCIPisStopped(scip) )
         return SCIP_OKAY;

      permlib::GroupType* compType;
      std::vector<unsigned int> componentIndices;
      SCIP_CALL( analyzeComponent(comp, compType, componentIndices) );
      assert(compType);

      if ( compType->type() == GroupType::Named )
      {
         NamedGroupType* namedType = reinterpret_cast<NamedGroupType*>(compType);

         if ( strcmp(namedType->name(), "S") == 0 )
         {
            /* symmetric group of degree namedType->typeDegree() */
            if ( namedType->isNaturalAction() )
            {
               /* This is a ``real'' symmetric group. We may use all variables in componentIndices. */
               SCIPdebugMessage("Adding ordering constraints for symmetric group of degree %u ...\n", namedType->typeDegree());
               SCIP_CALL( addSymmetricGroupBreakingConstraints(scip, presol, componentIndices.size(), componentIndices.begin(), componentIndices.end(), true) );
            }
            else
            {
               /* This is a sub-direct product that is only isomorphic to a symmetric group. */

               /* check whether we have an orbitope structure: compute expected number of rows (= expectedColumnSize) */
               const unsigned int expectedColumnSize = namedType->realDegree() / namedType->typeDegree();

               /* collect indices in transposed vector by looping over the orbits (= rows) */
               std::vector<dom_int> orbitTransponed(group->n, expectedColumnSize);
               std::list<boost::shared_ptr<OrbitAsSet> > allorbits =
                  orbits<unsigned long, Transversal<PERMUTATION>::TrivialAction>(bsgs, componentIndices.begin(), componentIndices.end());
               unsigned int orbitIndex = 0;
               for (std::list<boost::shared_ptr<OrbitAsSet> >::const_iterator aoit = allorbits.begin(); aoit != allorbits.end(); ++aoit)
               {
                  OrbitAsSet::const_iterator oit = (*aoit)->begin();
                  const OrbitAsSet::const_iterator oend = (*aoit)->end();
                  while (oit != oend)
                  {
                     orbitTransponed[*oit] = orbitIndex;
                     ++oit;
                  }
                  ++orbitIndex;
               }
               assert( orbitIndex == expectedColumnSize);

               /* try to detect orbitope structure */
               SCIP_CONS* orbitopecons;
               bool success = false;
               SCIP_CALL( tryOrbitopeConstraint(scip, presol, group, componentIndices, orbitTransponed, expectedColumnSize, orbitopecons, success) );

               if ( success )
               {
                  assert( orbitopecons != 0 );
                  presoldata->genconss->push_back(orbitopecons);
                  ++presoldata->ngenconss;
               }
               else
               {
                  /* If no orbitope constraint could be found, we can only use variables from one orbit
                   * of componentIndices (e.g. from the first orbit). We must not add coreset-type 
                   * inequalities in this case. */
                  OrbitAsSet orbit;
                  orbit.orbit(componentIndices.front(), bsgs.S, Transversal<PERMUTATION>::TrivialAction());

                  SCIPdebugMessage("Adding ordering constraints for orbit of symmetric group of size %lu ...\n", (unsigned long) orbit.size());
                  SCIP_CALL( addSymmetricGroupBreakingConstraints(scip, presol, orbit.size(), orbit.begin(), orbit.end(), false) );
               }
            }

            /* mark variable indices as used */
            brokenIndices.insert(componentIndices.begin(), componentIndices.end());
         }
         else if ( strcmp(namedType->name(), "C") == 0 )
         {
            /* cyclic group of degree namedType->typeDegree() */
            if (namedType->isNaturalAction())
            {
               /* This is a ``real'' cyclic group. We may use all variables in componentIndices. */
               SCIP_CALL( addCyclicGroupBreakingConstraints(scip, presol, componentIndices.begin(), componentIndices.end()) );
               SCIPdebugMessage("Added ordering constraints for cyclic group of degree %u.\n", namedType->typeDegree());
            }
            else
            {
               OrbitAsSet orbit;

               /* This is a sub-direct product that is only isomorphic to a cyclic group. We may
                * only use variables from one orbit of componentIndices (e.g. from the first orbit). */
               orbit.orbit(componentIndices.front(), bsgs.S, Transversal<PERMUTATION>::TrivialAction());

               SCIP_CALL( addCyclicGroupBreakingConstraints(scip, presol, orbit.begin(), orbit.end()) );
               SCIPdebugMessage("Added ordering constraints for orbit of cyclic group of size %lu.\n", (unsigned long) orbit.size());
            }

            /* mark variable indices as used */
            brokenIndices.insert(componentIndices.begin(), componentIndices.end());
         }
         else
         {
            SCIPdebugMessage("Found subgroup type that is not a symmetric or cyclic group.\n");
         }
      }
      else if ( compType->type() == GroupType::Anonymous )
      {
         std::vector<dom_int> orbitCharacteristic;
         GroupType* subtype = largeSymmetricDiagonalSubgroup(comp, orbitCharacteristic);
         if ( subtype && subtype->type() == GroupType::Named && orbitCharacteristic.size() > 0 && nsetppcconss > 3 )
         {
            NamedGroupType* namedSubtype = reinterpret_cast<NamedGroupType*>(subtype);

            /* check whether we have an orbitope structure: compute expected number of rows (= expectedColumnSize) */
            const unsigned int expectedColumnSize = namedSubtype->realDegree() / namedSubtype->typeDegree();

            /* collect indices in transposed vector by looping over the orbits (= rows) */
            std::vector<dom_int> orbitTransponed(bsgs.n, expectedColumnSize - 1);
            for (unsigned int i = 0; i < orbitCharacteristic.size(); ++i)
            {
               orbitTransponed[componentIndices[i]] = orbitCharacteristic[i];
            }
            PermGroupPtr diagonal = vectorStabilizer(bsgs, orbitTransponed.begin(), orbitTransponed.end(), expectedColumnSize - 1);
            assert( diagonal );

            /* try to detect orbitope structure */
            SCIP_CONS* orbitopecons;
            bool success = false;
            SCIP_CALL( tryOrbitopeConstraint(scip, presol, diagonal, componentIndices, orbitTransponed, expectedColumnSize, orbitopecons, success) );
            if ( success )
            {
               assert( orbitopecons != 0 );
               presoldata->genconss->push_back(orbitopecons);
               ++presoldata->ngenconss;

               /* mark variable indices as used */
               brokenIndices.insert(componentIndices.begin(), componentIndices.end());
            }
         }
         delete subtype;
      }
      else
      {
         SCIPdebugMessage("Could not determine symmetric or cyclic subgroup.\n");
      }

      delete compType;

      ++cc;
   }

   /* add linear inequalities arising from the fundamental domain (take all ones vector) */
   if ( presoldata->addfundamental )
   {
      SCIP_CALL( addFundamentalDomainConstraints(scip, presol, bsgs.S, brokenIndices) );
   }

   return SCIP_OKAY;
}


/** analyze symmetries for symmetric and cyclic subgroups and adds symmetry breaking constraints */
static
SCIP_RETCODE addSymmetryBreakingConstraintsHeur(
   SCIP*                 scip,               /**< SCIP instance */
   SCIP_PRESOL*          presol              /**< presolver */
   )
{
   SCIP_PRESOLDATA* presoldata;

   assert( scip != 0 );
   assert( presol != 0 );

   presoldata = SCIPpresolGetData(presol);
   assert( presoldata != 0 );

   /* exit if no or only trivial symmetry group is available */
   if ( ! presoldata->group || presoldata->group->order() == 1 || presoldata->nvars == 0 )
      return SCIP_OKAY;

   OrbitListPtr orb( presoldata->group->orbits() );
   for (OrbIt it = orb->begin(); it != orb->end(); ++it)
   {
      SCIP_CALL( addSymmetricGroupBreakingConstraints(scip, presol, it->size(), it->begin(), it->end(), true) );
      SCIPdebugMessage("Added ordering constraints for symmetric group of degree %lu.\n", (unsigned long) (*it).size());
   }
   return SCIP_OKAY;
}


/** analyze symmetries for symmetric and cyclic subgroups and adds symmetry breaking constraints */
static
SCIP_RETCODE addSymmetryBreakingConstraints(
   SCIP*                 scip,               /**< SCIP instance */
   SCIP_PRESOL*          presol              /**< presolver */
   )
{
   SCIP_PRESOLDATA* presoldata;

   assert( scip != 0 );
   assert( presol != 0 );

   presoldata = SCIPpresolGetData(presol);
   assert( presoldata != 0 );

   /* exit if no or only trivial symmetry group is available */
   if ( ! presoldata->group || presoldata->group->order() == 1 || presoldata->nvars == 0 )
      return SCIP_OKAY;

   switch ( presoldata->group->type() )
   {
   case permlib::AGT_BSGS:
      SCIP_CALL( addSymmetryBreakingConstraintsBSGS(scip, presol) );
      break;

   case permlib::AGT_SymmetricProduct:
      SCIP_CALL( addSymmetryBreakingConstraintsHeur(scip, presol) );
      break;

   default:
      SCIPerrorMessage("Unknown abstract group type.\n");
      return SCIP_ERROR;
   }

   return SCIP_OKAY;
}


/** applies and evaluates probing of a single binary variable in the given direction */
static
SCIP_RETCODE applyProbingVar(
   SCIP*                 scip,               /**< SCIP data structure */
   int                   proprounds,         /**< number of propagation rounds */
   SCIP_VAR**            vars,               /**< problem variables */
   int                   nvars,              /**< number of problem variables */
   SCIP_VAR*             probvar,            /**< variable to apply probing on */
   SCIP_Bool             probingdir,         /**< value to fix probing variable to */
   SCIP_Real*            impllbs,            /**< array to store lower bounds after applying implications and cliques */
   SCIP_Real*            implubs,            /**< array to store upper bounds after applying implications and cliques */
   SCIP_Real*            proplbs,            /**< array to store lower bounds after full propagation */
   SCIP_Real*            propubs,            /**< array to store upper bounds after full propagation */
   SCIP_Bool*            cutoff              /**< pointer to store whether the probing direction is infeasible */
   )
{  /*lint --e{818}*/
   assert( probvar != NULL );
   assert( impllbs != NULL );
   assert( implubs != NULL );
   assert( proplbs != NULL );
   assert( propubs != NULL );
   assert( cutoff != NULL );
   assert( SCIPvarIsBinary(probvar) );
   assert( SCIPvarGetLbLocal(probvar) < 0.5 );
   assert( SCIPvarGetUbLocal(probvar) > 0.5 );

   /* start probing mode */
   SCIP_CALL( SCIPstartProbing(scip) );

   /* fix variable */
   if ( probingdir == FALSE )
   {
      SCIP_CALL( SCIPchgVarUbProbing(scip, probvar, 0.0) );
   }
   else
   {
      SCIP_CALL( SCIPchgVarLbProbing(scip, probvar, 1.0) );
   }

   /* apply propagation of implication graph and clique table */
   SCIP_CALL( SCIPpropagateProbingImplications(scip, cutoff) );
   if ( !(*cutoff) )
   {
      for (int i = 0; i < nvars; ++i)
      {
         impllbs[i] = SCIPvarGetLbLocal(vars[i]);
         implubs[i] = SCIPvarGetUbLocal(vars[i]);
      }
   }

   /* apply propagation */
   if ( !(*cutoff) )
   {
      SCIP_CALL( SCIPpropagateProbing(scip, proprounds, cutoff, NULL) );
   }

   /* evaluate propagation */
   if ( !(*cutoff) )
   {
      for (int i = 0; i < nvars; ++i)
      {
         proplbs[i] = SCIPvarGetLbLocal(vars[i]);
         propubs[i] = SCIPvarGetUbLocal(vars[i]);
      }
   }

   /* exit probing mode */
   SCIP_CALL( SCIPendProbing(scip) );

   return SCIP_OKAY;
}


/** apply probing on list of variables */
static
SCIP_RETCODE applyProbing(
   SCIP*                 scip,               /**< SCIP data structure */
   SCIP_PRESOLDATA*      presoldata,         /**< data of presolver */
   int*                  nfixedvars,         /**< pointer to store number of fixed variables */
   int*                  naggrvars,          /**< pointer to store number of aggregated variables */
   int*                  nchgbds,            /**< pointer to store number of changed bounds */
   SCIP_Bool&            cutoff              /**< pointer to store whether cutoff occured */
   )
{
   SCIP_Real* zeroimpllbs;
   SCIP_Real* zeroimplubs;
   SCIP_Real* zeroproplbs;
   SCIP_Real* zeropropubs;
   SCIP_Real* oneimpllbs;
   SCIP_Real* oneimplubs;
   SCIP_Real* oneproplbs;
   SCIP_Real* onepropubs;

   assert( scip != 0 );
   assert( presoldata != 0 );

   cutoff = FALSE;

   /* get number of variables */
   int nvars = presoldata->nvars;
   assert( nvars > 0 );
   SCIP_VAR** vars = presoldata->vars;
   assert( vars != 0 );

   int nprobsets = presoldata->nprobsets;
   int* nprobvars = presoldata->nprobvars;
   SCIP_VAR*** probvars = presoldata->probvars;

   /* get temporary memory for storing probing results */
   SCIP_CALL( SCIPallocBufferArray(scip, &zeroimpllbs, nvars) );
   SCIP_CALL( SCIPallocBufferArray(scip, &zeroimplubs, nvars) );
   SCIP_CALL( SCIPallocBufferArray(scip, &zeroproplbs, nvars) );
   SCIP_CALL( SCIPallocBufferArray(scip, &zeropropubs, nvars) );
   SCIP_CALL( SCIPallocBufferArray(scip, &oneimpllbs, nvars) );
   SCIP_CALL( SCIPallocBufferArray(scip, &oneimplubs, nvars) );
   SCIP_CALL( SCIPallocBufferArray(scip, &oneproplbs, nvars) );
   SCIP_CALL( SCIPallocBufferArray(scip, &onepropubs, nvars) );

   /* loop through sets of variables */
   for (int i = 0; i < nprobsets && ! cutoff && ! SCIPisStopped(scip); ++i)
   {
      /* loop through probing set */
      assert( nprobvars[i] > 0 );
      for (int j = 0; j < nprobvars[i] && ! cutoff && ! SCIPisStopped(scip); ++j)
      {
         SCIP_Bool localcutoff = FALSE;
         SCIP_Bool fixed = FALSE;

         SCIP_VAR* probvar = probvars[i][j];
         assert( probvar != 0 );
         assert( SCIPvarGetType(probvar) == SCIP_VARTYPE_BINARY );

         /* ignore variables, that were fixed, aggregated, or deleted in prior probings */
         if ( !SCIPvarIsActive(probvar) || SCIPvarIsDeleted(probvar)
            || SCIPvarGetLbLocal(probvar) > 0.5 || SCIPvarGetUbLocal(probvar) < 0.5 )
            continue;

         SCIPdebugMessage("set: %d - applying probing on variable <%s> == 0 (nlocks=%d/%d, impls=%d/%d, clqs=%d/%d)\n",
            i, SCIPvarGetName(probvar), SCIPvarGetNLocksDown(probvar), SCIPvarGetNLocksUp(probvar),
            SCIPvarGetNImpls(probvar, FALSE), SCIPvarGetNImpls(probvar, TRUE),
            SCIPvarGetNCliques(probvar, FALSE), SCIPvarGetNCliques(probvar, TRUE));

         /* apply probing for fixing the variable to zero */
         SCIP_CALL( applyProbingVar(scip, MAXPROBINGPROPROUNDS, vars, nvars, probvar, FALSE,
               zeroimpllbs, zeroimplubs, zeroproplbs, zeropropubs, &localcutoff) );

         if ( localcutoff )
         {
            SCIPdebugMessage("Fixing variable <%s> to 1.\n", SCIPvarGetName(probvar));
            if ( SCIPgetStage(scip) != SCIP_STAGE_SOLVING || SCIPnodeGetDepth(SCIPgetCurrentNode(scip)) == 0 )
            {
               /* the variable can be fixed to TRUE */
               SCIP_CALL( SCIPfixVar(scip, probvar, 1.0, &cutoff, &fixed) );
            }
            else
            {
               SCIP_CALL( SCIPtightenVarLb(scip, probvar, 1.0, TRUE, &cutoff, &fixed) );
            }
            if ( cutoff )
            {
               SCIPdebugMessage("Fixing variable <%s> to 1 detected cutoff.\n", SCIPvarGetName(probvar));
               break;
            }

            if ( fixed )
            {
               SCIPdebugMessage("fixed probing variable <%s> to 1.0, nlocks=(%d/%d)\n",
                  SCIPvarGetName(probvar), SCIPvarGetNLocksDown(probvar), SCIPvarGetNLocksUp(probvar));
               (*nfixedvars)++;
            }
            continue; /* don't try upwards direction, because the variable is already fixed */
         }

         /* ignore variables, that were fixed, aggregated, or deleted in prior probings
          * (propagators in zero-probe might have found global fixings but did not trigger the localcutoff)
          */
         if ( !SCIPvarIsActive(probvar) || SCIPvarIsDeleted(probvar)
            || SCIPvarGetLbLocal(probvar) > 0.5 || SCIPvarGetUbLocal(probvar) < 0.5 )
            continue;

         SCIPdebugMessage("set: %d - applying probing on variable <%s> == 1 (nlocks=%d/%d, impls=%d/%d, clqs=%d/%d)\n",
            i, SCIPvarGetName(probvar), SCIPvarGetNLocksDown(probvar), SCIPvarGetNLocksUp(probvar),
            SCIPvarGetNImpls(probvar, FALSE), SCIPvarGetNImpls(probvar, TRUE),
            SCIPvarGetNCliques(probvar, FALSE), SCIPvarGetNCliques(probvar, TRUE));

         /* apply probing for fixing the variable to one */
         SCIP_CALL( applyProbingVar(scip, MAXPROBINGPROPROUNDS, vars, nvars, probvar, TRUE,
               oneimpllbs, oneimplubs, oneproplbs, onepropubs, &localcutoff) );

         if ( localcutoff )
         {
            SCIPdebugMessage("Fixing variable <%s> to 0.\n", SCIPvarGetName(probvar));
            if ( SCIPgetStage(scip) != SCIP_STAGE_SOLVING || SCIPnodeGetDepth(SCIPgetCurrentNode(scip)) == 0 )
            {
               /* the variable can be fixed to FALSE */
               SCIP_CALL( SCIPfixVar(scip, probvar, 0.0, &cutoff, &fixed) );
               assert( fixed );
            }
            else
            {
               SCIP_CALL( SCIPtightenVarUb(scip, probvar, 0.0, TRUE, &cutoff, &fixed) );
            }
            if ( cutoff )
            {
               SCIPdebugMessage("Fixing variable <%s> to 1 detected cutoff.\n", SCIPvarGetName(probvar));
               break;
            }

            if ( fixed )
            {
               SCIPdebugMessage("fixed probing variable <%s> to 0.0, nlocks=(%d/%d)\n",
                  SCIPvarGetName(probvar), SCIPvarGetNLocksDown(probvar), SCIPvarGetNLocksUp(probvar));
               (*nfixedvars)++;
            }
            continue; /* don't analyze probing deductions, because the variable is already fixed */
         }
         assert( ! fixed );

         /* analyze probing deductions */
         int localnfixedvars = 0;
         int localnaggrvars = 0;
         int localnimplications = 0;
         int localnchgbds = 0;
         SCIP_CALL( SCIPanalyzeDeductionsProbing(scip, probvar, 0.0, 1.0,
               nvars, vars, zeroimpllbs, zeroimplubs, zeroproplbs, zeropropubs, oneimpllbs, oneimplubs, oneproplbs, onepropubs,
               &localnfixedvars, &localnaggrvars, &localnimplications, &localnchgbds, &cutoff) );

         *nfixedvars += localnfixedvars;
         *naggrvars += localnaggrvars;
         *nchgbds += localnchgbds;

         if ( localnfixedvars > 0 || localnaggrvars > 0 || localnchgbds > 0 )
         {
            SCIPdebugMessage("fixed vars: %d, aggregated vars: %d, bound changes: %d\n", localnfixedvars, localnaggrvars, localnchgbds);
         }
#ifdef SCIP_DEBUG
         if ( cutoff )
         {
            SCIPdebugMessage("Detected cutoff by probing on variable <%s>.\n", SCIPvarGetName(probvar));
         }
#endif

         /* if no fixing has been found take next set */
         if ( localnfixedvars + localnaggrvars + localnchgbds == 0 )
            break;
      }
   }

   /* free temporary memory */
   SCIPfreeBufferArray(scip, &onepropubs);
   SCIPfreeBufferArray(scip, &oneproplbs);
   SCIPfreeBufferArray(scip, &oneimplubs);
   SCIPfreeBufferArray(scip, &oneimpllbs);
   SCIPfreeBufferArray(scip, &zeropropubs);
   SCIPfreeBufferArray(scip, &zeroproplbs);
   SCIPfreeBufferArray(scip, &zeroimplubs);
   SCIPfreeBufferArray(scip, &zeroimpllbs);

   return SCIP_OKAY;
}


/*
 * Callback methods of presolver
 */


/** destructor of presolver to free user data (called when SCIP is exiting) */
static
SCIP_DECL_PRESOLFREE(presolFreeSymbreak)
{  /*lint --e{715}*/
   SCIP_PRESOLDATA* presoldata;

   assert( scip != 0 );
   assert( presol != 0 );

   presoldata = SCIPpresolGetData(presol);
   assert( presoldata != 0 );

   /* free list of probing sets and variables */
   for (int i = 0; i < presoldata->nprobsets; ++i)
   {
      SCIPfreeBlockMemoryArray(scip, &(presoldata->probvars[i]), presoldata->nprobvars[i]);
   }
   SCIPfreeBlockMemoryArrayNull(scip, &presoldata->nprobvars, presoldata->nmaxprobsets);
   SCIPfreeBlockMemoryArrayNull(scip, &presoldata->probvars, presoldata->nmaxprobsets);
   presoldata->nmaxprobsets = 0;
   presoldata->nprobsets = 0;

   /* free pointers to symmetry group and binary variables */
   presoldata->group.reset();
   presoldata->binaryvars.reset();
   delete presoldata->genconss;

   SCIPfreeMemory(scip, &presoldata);

   return SCIP_OKAY;
}


/** initialization method of presolver (called after problem was transformed) */
static
SCIP_DECL_PRESOLINIT(presolInitSymbreak)
{
   SCIP_PRESOLDATA* presoldata;

   assert( scip != 0 );
   assert( presol != 0 );

   SCIPdebugMessage("Initpre method of symmetry breaking presolver ...\n");

   presoldata = SCIPpresolGetData(presol);
   assert( presoldata != 0 );

   /* check whether we should run */
   SCIP_CALL( SCIPgetBoolParam(scip, "usesymbreak", &presoldata->enabled) );

   if ( presoldata->enabled )
   {
      /* allow all problem specifications, since we handle them in the code above */
      SYMsetProbSpecRequirement(presoldata->symmetriesconshdlr, SCIP_PROBSPEC_BINARY);
      SYMsetProbSpecRequirement(presoldata->symmetriesconshdlr, SCIP_PROBSPEC_INTEGER);
      SYMsetProbSpecRequirement(presoldata->symmetriesconshdlr, SCIP_PROBSPEC_REAL);
   }

   return SCIP_OKAY;
}


/** deinitialization method of presolver (called before transformed problem is freed) */
static
SCIP_DECL_PRESOLEXIT(presolExitSymbreak)
{
   SCIP_PRESOLDATA* presoldata;

   assert( scip != 0 );
   assert( presol != 0 );

   SCIPdebugMessage("Exit method of symmetry breaking presolver ...\n");

   presoldata = SCIPpresolGetData(presol);
   assert( presoldata != 0 );

   if ( presoldata->ngenconss > 0 )
   {
      SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "\nNumber of orbitope constraints added:\t\t%d\n", presoldata->norbitopes);
      SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Number of full symmetric groups handled:\t%u\n", presoldata->nsymgroupbreak);
      SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Number of cyclic groups handled:\t\t%u\n", presoldata->ncycgroupbreak);
      SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Number of fundamental domains handled:\t\t%u\n", presoldata->nfundbreak);
      SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Total number of constraints added:\t\t%u\n\n", presoldata->ngenconss);
   }

   /* release constraints */
   conslist::const_iterator it, iend = presoldata->genconss->end();
   assert( presoldata->genconss->size() == presoldata->ngenconss );
   for (it = presoldata->genconss->begin(); it != iend; ++it)
   {
      SCIP_CONS* cons = *it;
      assert( cons != 0 );
      SCIP_CALL( SCIPreleaseCons(scip, &cons) );
   }
   presoldata->genconss->clear();

   return SCIP_OKAY;
}


/** presolving initialization method of presolver (called when presolving is about to begin) */
static
SCIP_DECL_PRESOLINITPRE(presolInitpreSymbreak)
{  /*lint --e{715}*/
   SCIP_PRESOLDATA* presoldata;

   assert( scip != 0 );
   assert( presol != 0 );

   presoldata = SCIPpresolGetData(presol);
   assert( presoldata != 0 );

   /* perform probing in every run */
   presoldata->performedprobing = FALSE;

   /* check whether we have to run the presolver at the beginning of presolving */
   presoldata->early = ! SYMdetectSymmetriesPresolved(presoldata->symmetriesconshdlr);
   if ( presoldata->early && presoldata->enabled )
   {
      SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Executing presolver <%s> early, since symmetries are computed early.\n\n", SCIPpresolGetName(presol));

      SCIP_CALL( SCIPsetIntParam(scip, "presolving/" PRESOL_NAME"/priority", 90000000) );
#if SCIP_VERSION < 320
      SCIP_CALL( SCIPsetBoolParam(scip, "presolving/" PRESOL_NAME"/delay", FALSE) );
#endif
   }

   return SCIP_OKAY;
}


/** execution method of presolver */
static
SCIP_DECL_PRESOLEXEC(presolExecSymbreak)
{  /*lint --e{715}*/
   SCIP_PRESOLDATA* presoldata;

   assert( scip != 0 );
   assert( presol != 0 );
   assert( result != 0 );
   assert( SCIPgetStage(scip) == SCIP_STAGE_PRESOLVING );

   *result = SCIP_DIDNOTRUN;

   presoldata = SCIPpresolGetData(presol);
   assert( presoldata != 0 );

   /* possibly skip presolver */
   if ( ! presoldata->enabled )
      return SCIP_OKAY;

   /* skip presolving if we are not at the end */
   if ( ! presoldata->early && ! SCIPisPresolveFinished(scip) )
      return SCIP_OKAY;

   /* possibly stop */
   if ( SCIPisStopped(scip) )
      return SCIP_OKAY;

   SCIPdebugMessage("Presolving method of symmetry breaking presolver ...\n");

   /* get symmetry information, if not already computed */
   if ( ! presoldata->computedsymmetry )
   {
      SCIPdebugMessage("Symmetry breaking presolver: computing symmetry ...\n");
      assert( ! presoldata->group );
      SCIP_CALL( SYMgetSymmetryData(scip, presoldata->symmetriesconshdlr, presoldata->group, presoldata->nvars, presoldata->vars, presoldata->binaryvars, presoldata->varmap) );
      presoldata->computedsymmetry = TRUE;

      if ( ! presoldata->group )
      {
         SCIPdebugMessage("Symmetry breaking presolver: no symmetry has been found, turning presolver off.\n");
         presoldata->enabled = FALSE;
         return SCIP_OKAY;
      }
      else if ( presoldata->group->order<double>() < (double) presoldata->minorder )
      {
         SCIPdebugMessage("Symmetry breaking presolver: order of symmetry group too small, turning presolver off.\n");
         presoldata->enabled = FALSE;
         return SCIP_OKAY;
      }
   }

   /* at this point, the symmetry group should be computed and nontrivial */
   assert( presoldata->group );
   *result = SCIP_DIDNOTFIND;

   /* possibly stop */
   if ( SCIPisStopped(scip) )
      return SCIP_OKAY;

   /* store old values */
   int noldfixedvars = *nfixedvars;
   int noldaggrvars = *naggrvars;
   int noldbdchgs = *nchgbds;
   int noldaddconss = *naddconss;

   /* if not already done add symmetry breaking constraints */
   if ( ! presoldata->addedconss )
   {
      /* add symmetry breaking constraints */
      unsigned int noldngenconns = presoldata->ngenconss;

      try
      {
         SCIP_CALL( addSymmetryBreakingConstraints(scip, presol) );
      }
      catch (const std::bad_alloc& ba)
      {
         /* try to catch bad memory allocation */
         SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Aborted adding symmetry breaking constraints because of lack of memory.\n");
         return SCIP_OKAY;
      }

      assert( presoldata->genconss->size() == presoldata->ngenconss );
      presoldata->addedconss = TRUE;
      *naddconss += presoldata->ngenconss - noldngenconns;
      SCIPdebugMessage("Added symmetry breaking constraints: %u.\n", presoldata->ngenconss - noldngenconns);

      /* if constraints have been added, loop through generated constraints and presolve each */
      conslist::const_iterator it, iend = presoldata->genconss->end();
      for (it = presoldata->genconss->begin(); it != iend; ++it)
      {
#if ( SCIP_VERSION >= 320 )
         SCIP_CALL( SCIPpresolCons(scip, *it, nrounds, SCIP_PRESOLTIMING_ALWAYS, nnewfixedvars, nnewaggrvars, nnewchgvartypes,
               nnewchgbds, nnewholes, nnewdelconss, nnewaddconss, nnewupgdconss, nnewchgcoefs, nnewchgsides, nfixedvars, naggrvars,
               nchgvartypes, nchgbds, naddholes, ndelconss, naddconss, nupgdconss, nchgcoefs, nchgsides, result) );
#else
         SCIP_CALL( SCIPpresolCons(scip, *it, nrounds, nnewfixedvars, nnewaggrvars, nnewchgvartypes, nnewchgbds, nnewholes,
               nnewdelconss, nnewaddconss, nnewupgdconss, nnewchgcoefs, nnewchgsides, nfixedvars, naggrvars, nchgvartypes,
               nchgbds, naddholes, ndelconss, naddconss, nupgdconss, nchgcoefs, nchgsides, result) );
#endif

         /* exit if cutoff has been detected */
         if ( *result == SCIP_CUTOFF || *result == SCIP_UNBOUNDED )
         {
            SCIPdebugMessage("Presolving constraint <%s> detected cutoff or unboundedness.\n", SCIPconsGetName(*it));
            return SCIP_OKAY;
         }
      }
      SCIPdebugMessage("Presolved %u constraints generated by symbreak.\n", presoldata->ngenconss);

      /* possibly stop */
      if ( SCIPisStopped(scip) )
         return SCIP_OKAY;
   }

   /* probing should come at the end of presolving, should be turned on, should not been performed already, and there should be probing sets */
   if ( SCIPisPresolveFinished(scip) && presoldata->useprobing && ! presoldata->performedprobing && presoldata->nprobsets > 0 )
   {
      assert( presoldata->addedconss );
      assert( presoldata->nprobvars != 0 );
      assert( presoldata->probvars != 0 );

      SCIP_Bool cutoff = FALSE;

      SCIPdebugMessage("Applying probing to %d probing sets.\n", presoldata->nprobsets);
      SCIP_CALL( applyProbing(scip, presoldata, nfixedvars, naggrvars, nchgbds, cutoff) );
      presoldata->performedprobing = TRUE;

      /* treat cutoff */
      if ( cutoff )
      {
         SCIPdebugMessage("Probing detected cutoff.\n");
         *result = SCIP_CUTOFF;
         return SCIP_OKAY;
      }

      /* again presolve constraints after probing */
      conslist::const_iterator it, iend = presoldata->genconss->end();
      for (it = presoldata->genconss->begin(); it != iend; ++it)
      {
         if ( SCIPconsIsActive(*it) )
         {
#if ( SCIP_VERSION >= 320 )
            SCIP_CALL( SCIPpresolCons(scip, *it, nrounds, SCIP_PRESOLTIMING_ALWAYS, nnewfixedvars, nnewaggrvars, nnewchgvartypes,
                  nnewchgbds, nnewholes, nnewdelconss, nnewaddconss, nnewupgdconss, nnewchgcoefs, nnewchgsides, nfixedvars, naggrvars,
                  nchgvartypes, nchgbds, naddholes, ndelconss, naddconss, nupgdconss, nchgcoefs, nchgsides, result) );
#else
            SCIP_CALL( SCIPpresolCons(scip, *it, nrounds, nnewfixedvars, nnewaggrvars, nnewchgvartypes, nnewchgbds, nnewholes,
                  nnewdelconss, nnewaddconss, nnewupgdconss, nnewchgcoefs, nnewchgsides, nfixedvars, naggrvars, nchgvartypes,
                  nchgbds, naddholes, ndelconss, naddconss, nupgdconss, nchgcoefs, nchgsides, result) );
#endif

            /* exit if cutoff has been detected */
            if ( *result == SCIP_CUTOFF || *result == SCIP_UNBOUNDED )
            {
               SCIPdebugMessage("Presolving constraint <%s> detected cutoff or unboundedness.\n", SCIPconsGetName(*it));
               return SCIP_OKAY;
            }
         }
      }
      SCIPdebugMessage("Presolved %u constraints generated by symbreak.\n", presoldata->ngenconss);
   }

   /* determine success */
   if ( noldaddconss + noldbdchgs + noldaggrvars + noldfixedvars > 0 )
      *result = SCIP_SUCCESS;

   return SCIP_OKAY;
}





/*
 * presolver specific interface methods
 */

/** creates the symbreak presolver and includes it in SCIP */
SCIP_RETCODE SCIPincludePresolSymbreak(
   SCIP*                 scip                /**< SCIP data structure */
   )
{
   SCIP_PRESOLDATA* presoldata = 0;

   /* create presolver data */
   SCIP_CALL( SCIPallocMemory(scip, &presoldata) );

   /* we must call the constructor explictly, because memory was m'alloced and not new'ed */
   new (&presoldata->group) PermutationGroupPtr();
   new (&presoldata->binaryvars) uintvecPtr();
   presoldata->vars = 0;
   presoldata->nvars = 0;
   presoldata->varmap = 0;
   presoldata->addedconss = FALSE;
   presoldata->performedprobing = FALSE;
   presoldata->computedsymmetry = FALSE;
   presoldata->nprobsets = 0;
   presoldata->nmaxprobsets = 0;
   presoldata->nprobvars = 0;
   presoldata->probvars = 0;
   presoldata->enabled = TRUE;
   presoldata->early = FALSE;
   presoldata->norbitopes = 0;
   presoldata->nsymgroupbreak = 0;
   presoldata->ncycgroupbreak = 0;
   presoldata->nfundbreak = 0;
   presoldata->ngenconss = 0;
   presoldata->genconss = new conslist();

   /* determine cons_symmetries constraint handler */
   SCIP_CONSHDLR* conshdlr = SCIPfindConshdlr(scip, "symmetries");
   if ( conshdlr == 0 )
   {
      SCIPerrorMessage("Could not find constraint handler <symmetries>.\n");
      return SCIP_ERROR;
   }
   presoldata->symmetriesconshdlr = conshdlr;

   /* include presolver */
   SCIP_PRESOL* presol = 0;
#if SCIP_VERSION >= 320
   SCIP_CALL( SCIPincludePresolBasic(scip, &presol, PRESOL_NAME, PRESOL_DESC, PRESOL_PRIORITY, PRESOL_MAXROUNDS, PRESOL_TIMING,
         presolExecSymbreak, presoldata) );
#else
   SCIP_CALL( SCIPincludePresolBasic(scip, &presol, PRESOL_NAME, PRESOL_DESC, PRESOL_PRIORITY, PRESOL_MAXROUNDS, PRESOL_DELAY, presolExecSymbreak, presoldata) );
#endif

   /* set additional callbacks */
   SCIP_CALL( SCIPsetPresolFree(scip, presol, presolFreeSymbreak) );
   SCIP_CALL( SCIPsetPresolInit(scip, presol, presolInitSymbreak) );
   SCIP_CALL( SCIPsetPresolExit(scip, presol, presolExitSymbreak) );
   SCIP_CALL( SCIPsetPresolInitpre(scip, presol, presolInitpreSymbreak) );

   /* add symmetry breaking presolver parameters */
   SCIP_CALL( SCIPaddBoolParam(scip,
         "presolving/" PRESOL_NAME"/conssaddlp",
         "Should the symmetry breaking constraints be added to the LP?",
         &presoldata->conssaddlp, TRUE, DEFAULT_CONSSADDLP, NULL, NULL) );

   SCIP_CALL( SCIPaddBoolParam(scip,
         "presolving/" PRESOL_NAME"/useprobing",
         "Should probing be applied to binary variables appearing in symmetry breaking constraints?",
         &presoldata->useprobing, TRUE, DEFAULT_USEPROBING, NULL, NULL) );

   SCIP_CALL( SCIPaddBoolParam(scip,
         "presolving/" PRESOL_NAME"/addfundamental",
         "Add inequalities from the fundamental domain for each generator?",
         &presoldata->addfundamental, TRUE, DEFAULT_ADDFUNDAMENTAL, NULL, NULL) );

   SCIP_CALL( SCIPaddBoolParam(scip,
         "presolving/" PRESOL_NAME"/orbitminonly",
         "Add only those symmetry breaking constraints which enforce minimality of a fixed, but arbitrary orbit element?",
         &presoldata->orbitminonly, TRUE, DEFAULT_ORBITMINONLY, NULL, NULL) );

   SCIP_CALL( SCIPaddIntParam(scip,
         "presolving/" PRESOL_NAME"/minorder",
         "minimal order of the symmetry group for which the presolver runs",
         &presoldata->minorder, TRUE, DEFAULT_MINORDER, -1, INT_MAX, NULL, NULL) );

   return SCIP_OKAY;
}

/* vim: set tabstop=3 expandtab: */
// kate: space-indent on; indent-width 3; mixedindent off; indent-mode cstyle;
