/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*    This file is part of "mipsymmetries" a collection of                   */
/*    symmetry handling methods.                                             */
/*                                                                           */
/*    Copyright (C) 2010-2015  Marc Pfetsch, Thomas Rehn                     */
/*                                                                           */
/*                                                                           */
/*    Based on SCIP  --- Solving Constraint Integer Programs                 */
/*       SCIP is distributed under the terms of the SCIP Academic Licence,   */
/*       see file COPYING in the SCIP distribution.                          */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
#pragma ident "@(#) $Id: prop_orbitalfixing.cpp 124 2011-10-19 18:19:40Z pfetsch $"
// #define PERMLIB_DEBUGMODE 1
// #define SCIP_OUTPUT
/**@file   prop_orbitalfixing.cpp
 * @ingroup PROPAGATORS
 * @brief  propagator for orbital fixing
 * @author Marc Pfetsch
 * @author Thomas Rehn
 */

/*---+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8----+----9----+----0----+----1----+----2*/

#include "prop_orbitalfixing.h"

#include <cassert>

#include <scip/pub_tree.h>

#include "orbital_fixing.h"
#include "cons_symmetries.h"
#include "cons_subtreebool.h"
#include "presol_orbitprobing.h"


/* propagator properties */
#define PROP_NAME              "orbitalfixing"
#define PROP_DESC              "propagator for orbital fixing"
#define PROP_PRIORITY                 0 /**< propagator priority */
#define PROP_FREQ                     1 /**< propagator frequency */
#define PROP_DELAY                FALSE /**< should propagation method be delayed, if other propagators found reductions? */
#define PROP_TIMING                SCIP_PROPTIMING_BEFORELP  /**< propagation timing mask */

/* default parameters */
#define SCIP_DEFAULT_AVOIDSUBTREENODES   FALSE    /**< if symmetry group becomes trivial avoid orbital fixing on sub-nodes? */
#define SCIP_DEFAULT_LOCALSYMMETRY       FALSE    /**< Compute local symmetry group in each node? */



using namespace permlib;




/*
 * Data structures
 */

/** propagator data for orbtial branching */
struct SCIP_PropData
{
   PermutationGroupPtr   group;              /**< symmetry group */
   unsigned int          nvars;              /**< number of variables */
   SCIP_VAR**            vars;               /**< variables in constraint */
   SCIP_HASHMAP*         varmap;             /**< map of variables to indices in vars array */
   SCIP_Longint          nodenumber;         /**< number of node where propagation has been last applied */
   SCIP_CONSHDLR*        symmetriesconshdlr; /**< pointer to cons_symmetries constraint handler */
   uintvecPtr            binaryvars;         /**< indices of binary variables to perform orbital branching on */
   SCIP_Bool             enabled;            /**< run orbital branching? */
   unsigned int          nfixedzero;         /**< number of variables fixed to 0 */
   unsigned int          nfixedone;          /**< number of variables fixed to 1 */
   SCIP_Bool             avoidsubtreenodes;  /**< if symmetry group becomes trivial, avoid orbital fixing on sub-nodes? */
   SCIP_Bool             localsymmetry;      /**< Compute local symmetry group in each node? */
   SCIP_CONSHDLR*        subtreeboolconshdlr;/**< pointer to cons_subtreebool constraint handler (needed for avoidsubtreenodes) */
};




/*
 * Local methods
 */


/** Get branching variables on the path to root */
static
SCIP_RETCODE computeBranchingVariables(
   SCIP*                 scip,               /**< SCIP pointer */
   unsigned int          nvars,              /**< number of variables */
   SCIP_HASHMAP*         varmap,             /**< map of variables to indices in vars array */
   uintvec&              B1                  /**< indices of all variables with branched value 1 */
   )
{
   /* get curent node */
   SCIP_NODE* node = SCIPgetCurrentNode(scip);

#ifdef SCIP_OUTPUT
   SCIP_CALL( SCIPprintNodeRootPath(scip, node, NULL) );
#endif

   /* prepare data */
   B1.clear();

   /* follow path to the root (in the root no domains were changed due to branching) */
   while ( SCIPnodeGetDepth(node) != 0 )
   {
      SCIP_BOUNDCHG* boundchg;
      SCIP_DOMCHG* domchg;
      int nboundchgs;

      /* get domain changes of current node */
      domchg = SCIPnodeGetDomchg(node);
      assert( domchg != 0 );

      /* loop through all bound changes */
      nboundchgs = SCIPdomchgGetNBoundchgs(domchg);
      for (int i = 0; i < nboundchgs; ++i)
      {
         /* get bound change info */
         boundchg = SCIPdomchgGetBoundchg(domchg, i);
         assert( boundchg != 0 );

         /* branching decisions have to be in the beginning of the bound change array */
         if ( SCIPboundchgGetBoundchgtype(boundchg) != SCIP_BOUNDCHGTYPE_BRANCHING )
            break;

         /* get corresponding branching variable */
         SCIP_VAR* branchvar = SCIPboundchgGetVar(boundchg);

         /* we only consider binary variables */
         if ( SCIPvarIsBinary(branchvar) )
         {
            /* make sure that branching variable is known */
            assert( SCIPhashmapExists(varmap, (void*) branchvar) );

            if ( SCIPvarGetLbLocal(branchvar) > 0.5 )
            {
               unsigned int branchvaridx = (unsigned int) (size_t) SCIPhashmapGetImage(varmap, (void*) branchvar);
               assert( branchvaridx < nvars );
               B1.push_back(branchvaridx);
            }
         }
      }

      node = SCIPnodeGetParent(node);
   }

   return SCIP_OKAY;
}


/** propagate orbital fixing */
static
SCIP_RETCODE propagate(
   SCIP*                 scip,               /**< SCIP pointer */
   SCIP_PROPDATA*        propdata,           /**< propagator data */
   SCIP_Bool&            infeasible,         /**< whether the node is detected to be infeasible */
   unsigned int&         ngen                /**< number of propagations */
   )
{
   assert( scip != NULL );
   assert( propdata != NULL );
   assert( propdata->vars != NULL );
   assert( propdata->nvars > 0 );
   assert( propdata->varmap != NULL );
   assert( propdata->group );

   infeasible = FALSE;
   ngen = 0;

   /* check whether we want to avoid using orbital fixing */
   if ( propdata->avoidsubtreenodes )
   {
      assert( propdata->subtreeboolconshdlr != 0 );
      if ( SUBTREEgetBool(propdata->subtreeboolconshdlr) )
      {
         SCIPdebugMessage("Avoid applying orbital fixing, because a trivial subgroup appeared earlier.\n");
         return SCIP_OKAY;
      }
   }

   /* check whether we want to compute stabilizers */
   PermutationGroupPtr nodeSymmetry;
   try
   {
      if ( propdata->localsymmetry )
      {
         /* compute local symmetry information */
         SCIP_CALL( SYMgetLocalSymmetry(scip, propdata->symmetriesconshdlr, nodeSymmetry) );
      }
      else
      {
         /* get branching variables */
         uintvec B1;
         SCIP_CALL( computeBranchingVariables(scip, propdata->nvars, propdata->varmap, B1) );

         /* prepare stabilizer w.r.t. all variables branched to 1 */
         if ( B1.size() )
            nodeSymmetry = PermutationGroupPtr(propdata->group->setStabilizer(B1));
         else
            nodeSymmetry = propdata->group;
      }
   }
   catch (std::bad_alloc& ba)
   {
      SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Aborted orbital fixing because we went out of memory.\n");
      return SCIP_OKAY;
   }

   if ( nodeSymmetry != 0 && nodeSymmetry->order() > 1 )
   {
      /* we may use orbital fixing */
      unsigned int nfixedone = 0;
      SCIP_CALL( orbitalFixing(scip, propdata->vars, propdata->nvars, nodeSymmetry, infeasible, nfixedone, ngen) );
      propdata->nfixedzero += ngen - nfixedone;
      propdata->nfixedone += nfixedone;

      SCIPdebugMessage("Orbital fixings: %u total, %u one.\n", ngen, nfixedone);
   }
   else
   {
      SCIPdebugMessage("Symmetry group is trivial. Skip orbital fixing.\n");

      /* turn off orbital fixing if this has been demanded */
      if ( propdata->avoidsubtreenodes )
      {
         /* store that we want to avoid orbital branching */
         assert( propdata->subtreeboolconshdlr != 0 );
         SCIP_CALL( SUBTREEsetTrue(scip, propdata->subtreeboolconshdlr) );
      }
   }

   return SCIP_OKAY;
}




/*
 * Callback methods of presolver
 */


/** destructor of propagator to free user data (called when SCIP is exiting) */
static
SCIP_DECL_PROPFREE(propFreeOrbitalfixing)
{  /*lint --e{715,818}*/
   assert( prop != NULL );

   SCIPdebugMessage("Freeing propagator <%s> ...\n", SCIPpropGetName(prop));

   SCIP_PROPDATA* propdata = SCIPpropGetData(prop);
   assert( propdata != NULL );

   propdata->group.reset();
   propdata->binaryvars.reset();
   SCIPfreeMemory(scip, &propdata);

   return SCIP_OKAY;
}


/** initialization method of propagator (called after problem was transformed) */
static
SCIP_DECL_PROPINIT(propInitOrbitalfixing)
{  /*lint --e{715}*/
   assert( prop != NULL );

   SCIPdebugMessage("Init propagator <%s> ...\n", SCIPpropGetName(prop));

   SCIP_PROPDATA* propdata = SCIPpropGetData(prop);
   assert( propdata != NULL );

   SCIP_CALL( SCIPgetBoolParam(scip, "useorbitalfixing", &propdata->enabled) );

   if ( propdata->enabled )
   {
      /* work on binary variables while fixing integer variables */
      SYMsetProbSpecRequirement(propdata->symmetriesconshdlr, SCIP_PROBSPEC_BINARY);
      SYMsetProbSpecRequirementFixed(propdata->symmetriesconshdlr, SCIP_PROBSPEC_INTEGER);
   }

   return SCIP_OKAY;
}


/** deinitialization method of propagator (called before transformed problem is freed) */
static
SCIP_DECL_PROPEXIT(propExitOrbitalfixing)
{  /*lint --e{715}*/
   assert( prop != NULL );

   SCIP_PROPDATA* propdata = SCIPpropGetData(prop);
   assert( propdata != NULL );

   if ( propdata->enabled )
   {
      SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "\nNumber of variables fixed to 0 by orbital fixing:\t%6u\n", propdata->nfixedzero);
      SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Number of variables fixed to 1 by orbital fixing:\t%6u\n", propdata->nfixedone);
   }

   return SCIP_OKAY;
}


/** solving process initialization method of propagator (called when branch and bound process is about to begin) */
static
SCIP_DECL_PROPINITSOL(propInitsolOrbitalfixing)
{  /*lint --e{715}*/
   assert( scip != 0 );
   assert( prop != 0 );

   /* get data */
   SCIP_PROPDATA* propdata = SCIPpropGetData(prop);
   assert( propdata != NULL );

   /* possibly skip orbital fixing */
   if ( ! propdata->enabled )
      return SCIP_OKAY;

   assert( SCIPisTransformed(scip) );

   /* get symmetry information */
   SCIP_CALL( SYMgetSymmetryData(scip, propdata->symmetriesconshdlr, propdata->group, propdata->nvars, propdata->vars,
         propdata->binaryvars, propdata->varmap) );

   if ( ! propdata->group )
   {
      SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Skip orbital fixing, since no symmetries were found.\n");
      propdata->enabled = FALSE;
   }

   return SCIP_OKAY;
}


/** execution method of propagator */
static
SCIP_DECL_PROPEXEC(propExecOrbitalfixing)
{  /*lint --e{715}*/
   assert( scip != NULL );
   assert( result != NULL );

   *result = SCIP_DIDNOTRUN;

   /* do not run if we are in the root or not yet solving */
   if (SCIPgetDepth(scip) <= 0 || SCIPgetStage(scip) < SCIP_STAGE_SOLVING)
      return SCIP_OKAY;

   /* do nothing if we are in a probing node */
   if ( SCIPinProbing(scip) )
      return SCIP_OKAY;

   /* get data */
   SCIP_PROPDATA* propdata = SCIPpropGetData(prop);
   assert( propdata != NULL );

   /* return if there is no symmetry available */
   if ( propdata->nvars == 0 || propdata->vars == NULL )
      return SCIP_OKAY;

   /* exit if no or only trivial symmetry group is available */
   if ( propdata->group == 0 || propdata->group->order() == 1)
      return SCIP_OKAY;

   /* return if we already ran in this node */
   SCIP_Longint nn = SCIPnodeGetNumber(SCIPgetCurrentNode(scip));
   if ( nn == propdata->nodenumber )
      return SCIP_OKAY;
   propdata->nodenumber = nn;

   /* propagate  */
   *result = SCIP_DIDNOTFIND;

   unsigned int ngen = 0;
   SCIP_Bool infeasible = FALSE;
   SCIPdebugMessage("Propagating <%s>.\n", SCIPpropGetName(prop));
   SCIP_CALL( propagate(scip, propdata, infeasible, ngen) );
   if ( infeasible )
      *result = SCIP_CUTOFF;
   else if ( ngen > 0 )
      *result = SCIP_REDUCEDDOM;

   return SCIP_OKAY;
}


/** propagation conflict resolving method of propagator
 *
 *  @todo Implement reverse propagation.
 */
static
SCIP_DECL_PROPRESPROP(propRespropOrbitalfixing)
{  /*lint --e{715,818}*/
   assert( result != 0 );

   *result = SCIP_DIDNOTFIND;
   return SCIP_OKAY;
}


/** creates the orbitalfixing propagator and includes it in SCIP */
SCIP_RETCODE SCIPincludePropOrbitalfixing(
   SCIP*                 scip                /**< SCIP data structure */
   )
{
   SCIP_PROPDATA* propdata = 0;

   /* create orbitalfixing propagator data */
   SCIP_CALL( SCIPallocMemory(scip, &propdata) );

   /* we must call the proptructor explictly, because memory was m'alloced and not new'ed */
   new (&propdata->group) PermutationGroupPtr();
   new (&propdata->binaryvars) uintvecPtr();
   propdata->nodenumber = -1;
   propdata->vars = NULL;
   propdata->nvars = 0;
   propdata->varmap = NULL;
   propdata->nfixedzero = 0;
   propdata->nfixedone = 0;
   propdata->avoidsubtreenodes = FALSE;

   /* determine cons_subtreebool constraint handler */
   SCIP_CONSHDLR* conshdlr = SCIPfindConshdlr(scip, "subtreebool");
   if ( conshdlr == 0 )
   {
      SCIPerrorMessage("Could not find constraint handler <subtreebool>.\n");
      return SCIP_ERROR;
   }
   propdata->subtreeboolconshdlr = conshdlr;

   /* determine cons_symmetries constraint handler */
   conshdlr = SCIPfindConshdlr(scip, "symmetries");
   if ( conshdlr == 0 )
   {
      SCIPerrorMessage("Could not find constraint handler <symmetries>.\n");
      return SCIP_ERROR;
   }
   propdata->symmetriesconshdlr = conshdlr;

   /* include propagator */
   SCIP_PROP* prop = 0;
   SCIP_CALL( SCIPincludePropBasic(scip, &prop, PROP_NAME, PROP_DESC, PROP_PRIORITY, PROP_FREQ, PROP_DELAY, PROP_TIMING, propExecOrbitalfixing, propdata) );

   /* set callbacks */
   SCIP_CALL( SCIPsetPropFree(scip, prop, propFreeOrbitalfixing) );
   SCIP_CALL( SCIPsetPropInit(scip, prop, propInitOrbitalfixing) );
   SCIP_CALL( SCIPsetPropExit(scip, prop, propExitOrbitalfixing) );
   SCIP_CALL( SCIPsetPropInitsol(scip, prop, propInitsolOrbitalfixing) );
   SCIP_CALL( SCIPsetPropResprop(scip, prop, propRespropOrbitalfixing) );

   /* add orbital fixing parameters */
   SCIP_CALL( SCIPaddBoolParam(scip, "propagating/orbitalfixing/avoidsubtreenodes",
         "if symmetry group becomes trivial avoid orbital fixing on sub-nodes?",
         &propdata->avoidsubtreenodes, TRUE, SCIP_DEFAULT_AVOIDSUBTREENODES, NULL, NULL) );

   SCIP_CALL( SCIPaddBoolParam(scip, "propagating/orbitalfixing/localsymmetry",
         "Compute local symmetry group in each node?",
         &propdata->localsymmetry, TRUE, SCIP_DEFAULT_LOCALSYMMETRY, NULL, NULL) );

   return SCIP_OKAY;
}

/* vim: set tabstop=3 expandtab: */
