/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*    This file is part of "mipsymmetries" a collection of                   */
/*    symmetry handling methods.                                             */
/*                                                                           */
/*    Copyright (C) 2010-2015  Marc Pfetsch, Thomas Rehn                     */
/*                                                                           */
/*                                                                           */
/*    Based on SCIP  --- Solving Constraint Integer Programs                 */
/*       SCIP is distributed under the terms of the SCIP Academic Licence,   */
/*       see file COPYING in the SCIP distribution.                          */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
#pragma ident "@(#) $Id: prop_isoprune.cpp 124 2011-10-19 18:19:40Z pfetsch $"
// #define SCIP_DEBUG
// #define PERMLIB_DEBUGMODE 1
// #define SCIP_OUTPUT
/**@file   prop_isoprune.cpp
 * @ingroup PROPAGATORS
 * @brief  propagator for isomorphism pruning
 * @author Marc Pfetsch
 * @author Thomas Rehn
 *
 * This propagator implements isomorphism pruning as developed by Margot in a series of papers:@par
 *  F. Margot: Pruning by isomorphism in branch-and-cut@n
 *  Math. Progrm. 94, No. 1, 71-90 (2002)
 * @par
 *  F. Margot: Exploiting orbits in symmetric ILP@n
 *  Math. Program. 98, 3-21 (2003).@par
 * @par
 *  F. Margot: Symmetry in integer linear programming@n
 *  in: M. Juenger, T. Liebling, D. Naddef, G.L. Nemhauser, W. Pulleyblank, G. Reinelt, G. Rinaldi, L. Wolsey (Eds.),@n
 *  50 Years of Integer Programming 1958-2008, Springer-Verlag, Berlin, Heidelberg, 2010, pp. 647-681. (Chapter 17).
 *
 * So far we only implemented the handling of binary branching variables. Integer variables are ignored, but we make
 * sure that they are not moved by symmetries.
 *
 * @todo Implement version for generals MIPs (currently only mixed binary programs are supported).
 */

/*---+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8----+----9----+----0----+----1----+----2*/

#include "prop_isoprune.h"
#include "cons_symmetries.h"

#include <scip/pub_tree.h>

#include "orbital_fixing.h"
#include "cons_subtreebool.h"
#include "presol_orbitprobing.h"

#include <cassert>


/* propagator properties */
#define PROP_NAME              "isoprune"
#define PROP_DESC              "propagator for isomorphism pruning"
#define PROP_PRIORITY                 0 /**< propagator priority */
#define PROP_FREQ                     1 /**< propagator frequency */
#define PROP_DELAY                FALSE /**< should propagation method be delayed, if other propagators found reductions? */
#define PROP_TIMING                SCIP_PROPTIMING_BEFORELP  /**< propagation timing mask */

/* default parameters */
#define SCIP_DEFAULT_AVOIDSUBTREENODES   FALSE    /**< if symmetry group becomes trivial avoid isomorphism pruning on sub-nodes? */


using namespace permlib;




/*
 * Data structures
 */

/** propagator data for isomorphism pruning */
struct SCIP_PropData
{
   PermutationGroupPtr   group;              /**< symmetry group */
   unsigned int          nvars;              /**< number of variables */
   SCIP_VAR**            vars;               /**< variables in constraint */
   SCIP_HASHMAP*         varmap;             /**< map of variables to indices in vars array */
   SCIP_Longint          nodenumber;         /**< number of node where propagation has been last applied */
   SCIP_CONSHDLR*        symmetriesconshdlr; /**< pointer to cons_symmetries constraint handler */
   uintvecPtr            binaryvars;         /**< indices of binary variables to perform isomorphism pruning on */
   SCIP_Bool             enabled;            /**< run isomorphism pruning? */
   unsigned int          nfixedzero;         /**< number of variables fixed to 0 */
   unsigned int          nfixedone;          /**< number of variables fixed to 1 */
   SCIP_Bool             avoidsubtreenodes;  /**< if symmetry group becomes trivial, avoid isomorphism pruning on sub-nodes? */
   SCIP_CONSHDLR*        subtreeboolconshdlr;/**< pointer to cons_subtreebool constraint handler (needed for avoidsubtreenodes) */
};




/*
 * Local methods
 */


/** Get branching variables on the path to root */
static
SCIP_RETCODE computeBranchingVariables(
   SCIP*                 scip,               /**< SCIP pointer */
   unsigned int          nvars,              /**< number of variables */
   SCIP_HASHMAP*         varmap,             /**< map of variables to indices in vars array */
   uintvec&              B,                  /**< indices of all binary branching variables */
   uintvec&              B1                  /**< indices of all variables with branched value 1 */
   )
{  /*lint --e{818}*/
   /* get curent node */
   SCIP_NODE* node = SCIPgetCurrentNode(scip);

#ifdef SCIP_OUTPUT
   SCIP_CALL( SCIPprintNodeRootPath(scip, node, NULL) );
#endif

   /* prepare data */
   B.clear();
   B1.clear();

   /* follow path to the root (in the root no domains were changed due to branching) */
   while ( SCIPnodeGetDepth(node) != 0 )
   {
      SCIP_BOUNDCHG* boundchg;
      SCIP_DOMCHG* domchg;
      int nboundchgs;

      /* get domain changes of current node */
      domchg = SCIPnodeGetDomchg(node);
      assert( domchg != 0 );

      /* loop through all bound changes */
      nboundchgs = SCIPdomchgGetNBoundchgs(domchg);
      for (int i = 0; i < nboundchgs; ++i)
      {
         /* get bound change info */
         boundchg = SCIPdomchgGetBoundchg(domchg, i);
         assert( boundchg != 0 );

         /* branching decisions have to be in the beginning of the bound change array */
         if ( SCIPboundchgGetBoundchgtype(boundchg) != SCIP_BOUNDCHGTYPE_BRANCHING )
            break;

         /* get corresponding branching variable */
         SCIP_VAR* branchvar = SCIPboundchgGetVar(boundchg);

         /* we only consider binary variables */
         if ( SCIPvarIsBinary(branchvar) )
         {
            /* make sure that branching variable is known */
            assert( SCIPhashmapExists(varmap, (void*) branchvar) );

            unsigned int branchvaridx = (unsigned int) (size_t) SCIPhashmapGetImage(varmap, (void*) branchvar);
            assert( branchvaridx < nvars );
            B.push_back(branchvaridx);

            if ( SCIPvarGetLbLocal(branchvar) > 0.5 )
               B1.push_back(branchvaridx);
         }
      }
      node = SCIPnodeGetParent(node);
   }

   return SCIP_OKAY;
}


/** propagate isomorphism pruning */
static
SCIP_RETCODE propagate(
   SCIP*                 scip,               /**< SCIP pointer */
   SCIP_PROPDATA*        propdata,           /**< propagator data */
   SCIP_Bool&            infeasible,         /**< whether the node is detected to be infeasible */
   unsigned int&         ngen                /**< number of propagations */
   )
{
   assert( scip != NULL );
   assert( propdata != NULL );
   assert( propdata->vars != NULL );
   assert( propdata->nvars > 0 );
   assert( propdata->varmap != NULL );

   infeasible = FALSE;
   ngen = 0;

   /* check whether we want to avoid using isomorphism pruning */
   if ( propdata->avoidsubtreenodes )
   {
      assert( propdata->subtreeboolconshdlr != 0 );
      if ( SUBTREEgetBool(propdata->subtreeboolconshdlr) )
      {
         SCIPdebugMessage("Avoid applying isomorphism pruning, because a trivial subgroup appeared earlier.\n");
         return SCIP_OKAY;
      }
   }

   /* get branching variables */
   uintvec B;
   uintvec B1;
   SCIP_CALL( computeBranchingVariables(scip, propdata->nvars, propdata->varmap, B, B1) );

   /* do nothing, if we do not have binary branching variables */
   if ( B.empty() )
      return SCIP_OKAY;

   /* index of the last branching variable */
   unsigned int k = B.front();
   SCIP_VAR* branchvar = propdata->vars[k];
   assert( branchvar != NULL );
   SCIPdebugMessage("Last branching variable <%s>.\n", SCIPvarGetName(branchvar));

   /* no pruning if no variable is branched to 1 */
   if ( B1.empty() )
   {
      /* we may use orbital fixing */
      unsigned int nfixedone = 0;
      SCIP_CALL( orbitalFixing(scip, propdata->vars, propdata->nvars, propdata->group, infeasible, nfixedone, ngen) );
      propdata->nfixedzero += ngen - nfixedone;
      propdata->nfixedone += nfixedone;
      return SCIP_OKAY;
   }

   /* compute representative in orbit of current node */
   uintvec Breverse(B.rbegin(), B.rend());
   bool isNoRepresentative = ! propdata->group->isLexMinSet(B1, Breverse);

   /* check whether representative equals the current node -> if equal we have a lexmin node */
   if (isNoRepresentative)
   {
      /* do not call conflict analysis, because it does not work together with isomorphism pruning */

      SCIPdebugMessage(" -> Infeasible node: does not correspond to the lexmin representative.\n");

      /* tree node does not correspond to a lexmin node -> prune node */
      infeasible = TRUE;
   }
   /* orbital fixing if we branched to zero and have a one-branched variable */
   else if (!B1.empty())
   {
      SCIPdebugMessage(" -> lexmin representative, compute stabilizer and orbits.\n");

      try
      {
         /* compute stabilizer of variables with fixed values */
         PermutationGroupPtr rankStabilizer(propdata->group->setStabilizer(B1));
         assert( rankStabilizer );
         assert( SCIPvarGetType(branchvar) == SCIP_VARTYPE_BINARY );

         /* run orbital fixing if group is non-trivial */
         if ( rankStabilizer->order() > 1 )
         {
            unsigned int nfixedone = 0;
            SCIP_CALL( orbitalFixing(scip, propdata->vars, propdata->nvars, rankStabilizer, infeasible, nfixedone, ngen) );
            propdata->nfixedzero += ngen - nfixedone;
            propdata->nfixedone += nfixedone;
         }
         else
         {
            SCIPdebugMessage("Symmetry group is trivial. Skip orbital fixing (from ismorphism pruning).\n");

            /* turn off isomorphism pruning (and orbital fixing) if stabilizer is trivial and this has been demanded */
            if ( propdata->avoidsubtreenodes )
            {
               /* store that we want to avoid isomorphism pruning */
               assert( propdata->subtreeboolconshdlr != 0 );
               SCIP_CALL( SUBTREEsetTrue(scip, propdata->subtreeboolconshdlr) );
            }
         }
      }
      catch (std::bad_alloc& ba)
      {
         SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Aborted isomorphism pruning because we went out of memory.\n");
         return SCIP_OKAY;
      }
   } // end if !isNoRepresentative

   return SCIP_OKAY;
}




/*
 * Callback methods of presolver
 */


/** destructor of propagator to free user data (called when SCIP is exiting) */
static
SCIP_DECL_PROPFREE(propFreeIsoprune)
{  /*lint --e{715,818}*/
   assert( prop != NULL );

   SCIPdebugMessage("Freeing propagator <%s> ...\n", SCIPpropGetName(prop));

   SCIP_PROPDATA* propdata = SCIPpropGetData(prop);
   assert( propdata != NULL );

   propdata->group.reset();
   propdata->binaryvars.reset();
   SCIPfreeMemory(scip, &propdata);

   return SCIP_OKAY;
}


/** initialization method of propagator (called after problem was transformed) */
static
SCIP_DECL_PROPINIT(propInitIsoprune)
{  /*lint --e{715}*/
   assert( prop != NULL );

   SCIPdebugMessage("Init propagator <%s> ...\n", SCIPpropGetName(prop));

   SCIP_PROPDATA* propdata = SCIPpropGetData(prop);
   assert( propdata != NULL );

   SCIP_CALL( SCIPgetBoolParam(scip, "useisoprune", &propdata->enabled) );

   if ( propdata->enabled )
   {
      /* work on binary variables while fixing integer variables */
      SYMsetProbSpecRequirement(propdata->symmetriesconshdlr, SCIP_PROBSPEC_BINARY);
      SYMsetProbSpecRequirementFixed(propdata->symmetriesconshdlr, SCIP_PROBSPEC_INTEGER);
   }

   return SCIP_OKAY;
}


/** deinitialization method of propagator (called before transformed problem is freed) */
static
SCIP_DECL_PROPEXIT(propExitIsoprune)
{  /*lint --e{715}*/
   assert( prop != NULL );

   SCIP_PROPDATA* propdata = SCIPpropGetData(prop);
   assert( propdata != NULL );

   if ( propdata->enabled )
   {
      SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "\nNumber of variables fixed to 0 by isomorphism pruning:\t%6u\n", propdata->nfixedzero);
      SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Number of variables fixed to 1 by isomorphism pruning:\t%6u\n", propdata->nfixedone);
   }

   return SCIP_OKAY;
}


/** solving process initialization method of propagator (called when branch and bound process is about to begin) */
static
SCIP_DECL_PROPINITSOL(propInitsolIsoprune)
{  /*lint --e{715}*/
   assert( scip != 0 );
   assert( prop != 0 );

   /* get data */
   SCIP_PROPDATA* propdata = SCIPpropGetData(prop);
   assert( propdata != NULL );

   /* possibly skip isomorphism pruning */
   if ( ! propdata->enabled )
      return SCIP_OKAY;

   assert( SCIPisTransformed(scip) );

   /* get symmetry information */
   SCIP_CALL( SYMgetSymmetryData(scip, propdata->symmetriesconshdlr, propdata->group, propdata->nvars, propdata->vars,
         propdata->binaryvars, propdata->varmap) );

   if ( ! propdata->group )
   {
      SCIPverbMessage(scip, SCIP_VERBLEVEL_MINIMAL, 0, "Skip isomorphism pruning, since no symmetries were found.\n");
      propdata->enabled = FALSE;
   }

   return SCIP_OKAY;
}


/** execution method of propagator */
static
SCIP_DECL_PROPEXEC(propExecIsoprune)
{  /*lint --e{715}*/
   assert( scip != NULL );
   assert( result != NULL );

   *result = SCIP_DIDNOTRUN;

   /* do not run if we are in the root or not yet solving */
   if (SCIPgetDepth(scip) <= 0 || SCIPgetStage(scip) < SCIP_STAGE_SOLVING)
      return SCIP_OKAY;

   /* do nothing if we are in a probing node */
   if ( SCIPinProbing(scip) )
      return SCIP_OKAY;

   /* get data */
   SCIP_PROPDATA* propdata = SCIPpropGetData(prop);
   assert( propdata != NULL );

   /* return if there is no symmetry available */
   if ( propdata->nvars == 0 || propdata->vars == NULL )
      return SCIP_OKAY;

   /* exit if no or only trivial symmetry group is available */
   if ( propdata->group == 0 || propdata->group->order() == 1)
      return SCIP_OKAY;

   /* return if we already ran in this node */
   SCIP_Longint nn = SCIPnodeGetNumber(SCIPgetCurrentNode(scip));
   if ( nn == propdata->nodenumber )
      return SCIP_OKAY;
   propdata->nodenumber = nn;

   /* propagate  */
   *result = SCIP_DIDNOTFIND;

   unsigned int ngen = 0;
   SCIP_Bool infeasible = FALSE;
   SCIPdebugMessage("Propagating <%s>.\n", SCIPpropGetName(prop));
   SCIP_CALL( propagate(scip, propdata, infeasible, ngen) );
   if ( infeasible )
      *result = SCIP_CUTOFF;
   else if ( ngen > 0 )
      *result = SCIP_REDUCEDDOM;

   return SCIP_OKAY;
}


/** propagation conflict resolving method of propagator
 *
 *  @todo Implement reverse propagation.
 */
static
SCIP_DECL_PROPRESPROP(propRespropIsoprune)
{  /*lint --e{715,818}*/
   assert( result != 0 );

   *result = SCIP_DIDNOTFIND;
   return SCIP_OKAY;
}


/** creates the isoprune propagator and includes it in SCIP */
SCIP_RETCODE SCIPincludePropIsoprune(
   SCIP*                 scip                /**< SCIP data structure */
   )
{
   SCIP_PROPDATA* propdata = 0;

   /* create isoprune propagator data */
   SCIP_CALL( SCIPallocMemory(scip, &propdata) );

   /* we must call the constructor explictly, because memory was m'alloced and not new'ed */
   new (&propdata->group) PermutationGroupPtr();
   new (&propdata->binaryvars) uintvecPtr();
   propdata->nodenumber = -1;
   propdata->vars = NULL;
   propdata->nvars = 0;
   propdata->varmap = NULL;
   propdata->nfixedzero = 0;
   propdata->nfixedone = 0;
   propdata->enabled = FALSE;
   propdata->avoidsubtreenodes = FALSE;

   /* determine cons_subtreebool constraint handler */
   SCIP_CONSHDLR* conshdlr = SCIPfindConshdlr(scip, "subtreebool");
   if ( conshdlr == 0 )
   {
      SCIPerrorMessage("Could not find constraint handler <subtreebool>.\n");
      return SCIP_ERROR;
   }
   propdata->subtreeboolconshdlr = conshdlr;

   /* determine cons_symmetries constraint handler */
   conshdlr = SCIPfindConshdlr(scip, "symmetries");
   if ( conshdlr == 0 )
   {
      SCIPerrorMessage("Could not find constraint handler <symmetries>.\n");
      return SCIP_ERROR;
   }
   propdata->symmetriesconshdlr = conshdlr;

   /* include propagator */
   SCIP_PROP* prop = 0;
   SCIP_CALL( SCIPincludePropBasic(scip, &prop, PROP_NAME, PROP_DESC, PROP_PRIORITY, PROP_FREQ, PROP_DELAY, PROP_TIMING, propExecIsoprune, propdata) );

   /* set callbacks */
   SCIP_CALL( SCIPsetPropFree(scip, prop, propFreeIsoprune) );
   SCIP_CALL( SCIPsetPropInit(scip, prop, propInitIsoprune) );
   SCIP_CALL( SCIPsetPropExit(scip, prop, propExitIsoprune) );
   SCIP_CALL( SCIPsetPropInitsol(scip, prop, propInitsolIsoprune) );
   SCIP_CALL( SCIPsetPropResprop(scip, prop, propRespropIsoprune) );

   /* add isomorphism pruning parameters */
   SCIP_CALL( SCIPaddBoolParam(scip, "propagating/isoprune/avoidsubtreenodes",
         "if symmetry group becomes trivial avoid isomorphism pruning on sub-nodes?",
         &propdata->avoidsubtreenodes, TRUE, SCIP_DEFAULT_AVOIDSUBTREENODES, NULL, NULL) );

   return SCIP_OKAY;
}

/* vim: set tabstop=3 expandtab: */
// kate: space-indent on; indent-width 3; mixedindent off; indent-mode cstyle;
